//****************************************************
//
//	project	PAFEE
//!	@file:	LockFreeRingBuffer.cpp
//!	@author: 
//!	@brief:	
//
//****************************************************
//
// P A F E E   - Pragmatic Application Framework for Embedded Environment- 
//
//
// Copyright (c) 2009 KN Information Systems Corporation,
//                    Hirocon Corporation,
//                    Technocrats Corporation
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//

#include "pafee_string.h"
#include "PafeeLockFreeRingBuffer.h"

using namespace Pafee;

//----------------------------------------------------------------------------
//!	Constructor
//! @return: 
//----------------------------------------------------------------------------
LockFreeRingBuffer::LockFreeRingBuffer(int capacity/*=0*/, int elementSize/*=0*/, Allocator* pAllocator/*=NULL*/)
{
	m_isInitialized = false;

	Initialize(capacity, elementSize, pAllocator);
}

//----------------------------------------------------------------------------
//!	Destructor
//! @return: 
//----------------------------------------------------------------------------
LockFreeRingBuffer::~LockFreeRingBuffer()
{
	if(m_pElementArray)
	{
		Deallocate(m_pElementArray);
	}
	
	if(m_pBuffer)
	{
		Deallocate(m_pBuffer);
	}
}

//----------------------------------------------------------------------------
//!	Initialize RockFreeRingBuffer
//! @return: Result(true:success false: failure)
//----------------------------------------------------------------------------
bool LockFreeRingBuffer::Initialize(int capacity, int elementSize, Allocator* pAllocator/*=NULL*/)
{
	if(!m_isInitialized)
	{
		int i;
		char* pBufferPointer;
		
		m_bufferCapacity = capacity + 1;
		m_maxElementSize = elementSize;
		SetAllocator(pAllocator);
		
		m_pBuffer = (char*) Allocate(m_bufferCapacity * m_maxElementSize);
		if(!m_pBuffer)
		{
			return false;
		}
		m_pElementArray = (BUFFER_ELEMENT*) Allocate(sizeof(BUFFER_ELEMENT) * m_bufferCapacity);
		if(!m_pElementArray)
		{
			return false;
		}

		pBufferPointer = m_pBuffer;
		for(i=0; i<m_bufferCapacity; i++)
		{
			m_pElementArray[i].pData = pBufferPointer;
			m_pElementArray[i].length = elementSize;
			pBufferPointer += elementSize;
		}

		m_readPoint = 0;
		m_writePoint = 0;

#ifdef PAFEE_STATISTICS_ENABLE
		pafee_memset(&m_statistics, 0, sizeof(m_statistics));
#endif

		m_isInitialized = true;
	}
	return true;
}

//----------------------------------------------------------------------------
//!	Get current number of elements
//! @return: Result(true:success false: failure)
//----------------------------------------------------------------------------
int LockFreeRingBuffer::GetCount(void)
{
	if(m_totalWriteCount > m_totalReadCount)
	{
		return  (m_totalWriteCount - m_totalReadCount);
	}
	return  (int)((0xffffffff - m_totalReadCount) + m_totalWriteCount + 1);
}

//----------------------------------------------------------------------------
//!	return if ring buffer is empty.
//! @return: Result(true:empty false: is not empty)
//----------------------------------------------------------------------------
bool LockFreeRingBuffer::IsEmpty(void)
{
	if(GetCount() == 0)
	{
		return true;
	}
	return false;
}

//----------------------------------------------------------------------------
//!	Get buffer write pointer.
//! @return: write pointer
//----------------------------------------------------------------------------
void* LockFreeRingBuffer::GetWriteBuffer(int* pLength)
{
	*pLength = m_maxElementSize;
	return m_pElementArray[m_writePoint].pData;
}

//----------------------------------------------------------------------------
//!	Enqueue
//! @return: Result(true: success false: failure)
//----------------------------------------------------------------------------
bool LockFreeRingBuffer::Enqueue(void* pData, int length)
{
	if(GetCount() < m_bufferCapacity)
	{
		if(length > m_maxElementSize)
		{
			return false;
		}

		{
			BUFFER_ELEMENT* pElement = &m_pElementArray[m_writePoint];
			pafee_memcpy(pElement->pData, pData, length);

			if(++m_writePoint >= m_bufferCapacity)
			{
				m_writePoint = 0;
			}
		}
		m_totalWriteCount++;
		STATISTICS_MAX(m_statistics.peakElementCount, GetCount());
	}
	else
	{
		STATISTICS_INC(m_statistics.overflowCount);
		return false;
	}
	return true;
}

//----------------------------------------------------------------------------
//!	Enqueue
//! @return: Result(true: success false: failure)
//----------------------------------------------------------------------------
bool LockFreeRingBuffer::Enqueue(int length)
{
	if(GetCount() < m_bufferCapacity)
	{
		if(length >m_maxElementSize)
		{
			return false;
		}
		
		m_pElementArray[m_writePoint].length = length;

		if(++m_writePoint >= m_bufferCapacity)
		{
			m_writePoint = 0;
		}
		m_totalWriteCount++;
		STATISTICS_MAX(m_statistics.peakElementCount, GetCount());
	}
	else
	{
		STATISTICS_INC(m_statistics.overflowCount);
		return false;
	}
	return true;
}

//----------------------------------------------------------------------------
//!	Dequeue
//! @return: Result(true: success false: failure)
//----------------------------------------------------------------------------
bool LockFreeRingBuffer::Dequeue(void* pBuffer, int* pLength)
{
	if(GetCount() > 0)
	{
		BUFFER_ELEMENT* pElement = &m_pElementArray[m_readPoint];
		*pLength = pElement->length;
		pafee_memcpy(pBuffer, pElement->pData, pElement->length);
		if(++m_readPoint >= m_bufferCapacity)
		{
			m_readPoint = 0;
		}
		m_totalReadCount++;
	}
	else
	{
		return false;
	}
		
	return true;
}

//----------------------------------------------------------------------------
//!	Dequeue
//! @return: buffer current read pointer
//----------------------------------------------------------------------------
void* LockFreeRingBuffer::Dequeue(int* pLength)
{
	if(GetCount() > 0)
	{
		BUFFER_ELEMENT* pElement = &m_pElementArray[m_readPoint];
		*pLength = pElement->length;
		
		if(++m_readPoint >= m_bufferCapacity)
		{
			m_readPoint = 0;
		}
		m_totalReadCount++;
		return pElement->pData;
	}
	return NULL;
}

//----------------------------------------------------------------------------
//!	Get buffer read pointer
//! @return: Result(true: success false: failure)
//----------------------------------------------------------------------------
bool LockFreeRingBuffer::Peek(void* pBuffer, int* pLength)
{
	if(GetCount() > 0)
	{
		BUFFER_ELEMENT* pElement = &m_pElementArray[m_readPoint];
		int length = pElement->length;

		*pLength = length;
		pafee_memcpy(pBuffer, pElement->pData, pElement->length);
		m_totalReadCount++;
		return true;
	}
	return false;
}

//----------------------------------------------------------------------------
//!	Get buffer read pointer
//! @return: buffer current read pointer
//----------------------------------------------------------------------------
void* LockFreeRingBuffer::Peek(int* pLength/*=NULL*/)
{
	if(GetCount() > 0)
	{
		BUFFER_ELEMENT* pElement = &m_pElementArray[m_readPoint];
		if(pLength)
		{
			*pLength = pElement->length;
		}
		return pElement->pData;
	}
	return NULL;
}

//----------------------------------------------------------------------------
//!	Get element size that is top of buffer
//! @return: Element size
//----------------------------------------------------------------------------
int LockFreeRingBuffer::PeekLength(void)
{
	if(m_writePoint != m_readPoint)
	{
		return m_pElementArray[m_readPoint].length;
	}
	return 0;
}

//----------------------------------------------------------------------------
//!	Clear buffer
//! @return: void
//----------------------------------------------------------------------------
void LockFreeRingBuffer::ClearAll(void)
{
	m_readPoint = 0;
	m_writePoint = 0;
}

//----------------------------------------------------------------------------
//!	Get statistics info
//! @return: void
//----------------------------------------------------------------------------
#ifdef PAFEE_STATISTICS_ENABLE
void LockFreeRingBuffer::GetStatistics(LockFreeRingBuffer::STATISTICS* pStatistics)
{
	m_statistics.eneqeueCount = m_totalWriteCount;
	if(m_totalWriteCount > m_totalReadCount)
	{
		m_statistics.currentElementCount = (m_totalWriteCount - m_totalReadCount);
	}
	else
	{
		m_statistics.currentElementCount = (0xffffffff - m_totalReadCount) + m_totalWriteCount + 1;
	}
	pafee_memcpy(pStatistics, &m_statistics, sizeof(m_statistics));
}
#endif // PAFEE_STATISTICS_ENABLE
#ifdef PAFEE_UNITTEST

//----------------------------------------------------------------------------
//!	Unit test
//! @return: void
//----------------------------------------------------------------------------
void LockFreeRingBuffer::UnitTest(void)
{
	// @@@
}
#endif
