//****************************************************
//
//    project   Pafee
//!   @file:    PafeeMultiMap.cpp
//!   @author:  M.Tanaka
//!   @brief:   MultiMap Class
//
//*****************************************************

//
// P A F E E   - Pragmatic Application Framework for Embedded Environment- 
//
//
// Copyright (c) 2009 KN Information Systems Corporation,
//                    Hirocon Corporation,
//                    Technocrats Corporation
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//

#include "PafeeMultiMap.h"

namespace Pafee
{

//----------------------------------------
//! MultiMap
//----------------------------------------
MultiMap::MultiMap(
	int hashSize, ///< [in] hash size
	Allocator* pAllocator ///< [in] Memory Allocator
)
{
	m_size =0;
	m_pHashTable = NULL;
	m_indexCount =0;
	m_writeIndex=0;
	m_pWriteBuffer=NULL;
	m_writeBufferSize=0;
	Initialize( hashSize, pAllocator);
#ifdef PAFEE_STATISTICS_ENABLE
	m_Statistics.addCount = 0;
	m_Statistics.acqCount = 0;
	m_Statistics.deleteCount = 0;
	m_Statistics.maxElementCount = 0;
	m_Statistics.collisionCount = 0;
	m_Statistics.minListCount = 0;
	m_Statistics.maxListCount = 0;
	m_Statistics.presentElementCount = 0;
#endif
}

//----------------------------------------
//! MultiMap
//----------------------------------------
MultiMap::MultiMap(
	const MultiMap& mmpc, ///< [in] initial value of MultiMap
	Allocator* pAllocator ///< [in] Memory Allocator
)
{
	m_size =0;
	m_pHashTable = NULL;
	m_indexCount =0;
	m_writeIndex=0;
	m_pWriteBuffer=NULL;
	m_writeBufferSize=0;
	Initialize( mmpc.m_hashSize, pAllocator);
	for(int i=0; i<m_hashSize; i++)
	{
		m_pHashTable[i] = mmpc.m_pHashTable[i];
	}
	m_size =mmpc.m_size;
#ifdef PAFEE_STATISTICS_ENABLE
	m_Statistics.addCount = mmpc.m_Statistics.addCount;
	m_Statistics.acqCount = mmpc.m_Statistics.acqCount;
	m_Statistics.deleteCount = mmpc.m_Statistics.deleteCount;
	m_Statistics.maxElementCount = mmpc.m_Statistics.maxElementCount;
	m_Statistics.collisionCount = mmpc.m_Statistics.collisionCount;
	m_Statistics.minListCount = mmpc.m_Statistics.minListCount;
	m_Statistics.maxListCount = mmpc.m_Statistics.maxListCount;
	m_Statistics.presentElementCount = mmpc.m_Statistics.presentElementCount;
  #endif
}

//----------------------------------------
//! ~MultiMap
//----------------------------------------
MultiMap::~MultiMap()
{
	if ( m_pHashTable != NULL )
	{
		for(int i=0; i<m_hashSize; i++)
		{
			(&(m_pHashTable[i]))->~List();
		}
		Deallocate(m_pHashTable);
	}
}

//----------------------------------------
//! Initialize
//! @return: Result(ture:success false:not success)
//----------------------------------------
bool MultiMap::Initialize(
	int hashSize, ///< [in] hash size
	Allocator* pAllocator ///< [in] Memory Allocator
)
{
	SetAllocator(pAllocator);
	if(hashSize == 0)
	{
		return false;
	}
	if(m_pHashTable != NULL)
	{
		return false;
	}
	m_hashSize = hashSize;
	m_pHashTable = (List*)Allocate(sizeof( List) * m_hashSize);
	if(!m_pHashTable)
	{
		return false;
	}
	for(int i=0; i<m_hashSize; i++)
	{
		new (&(m_pHashTable[i]))List(pAllocator);
	}
	return true;
}

//-------------------------------------------------------------------------------------
//! operator =
//! @return Result(the address of MultiMap)
//-------------------------------------------------------------------------------------
MultiMap& MultiMap::operator =(
	const MultiMap& mmpc ///< [in] initial value of MultiMap
)
{
	//this->RemoveAll();
	if ( m_pHashTable != NULL )
	{
		for(int i=0; i<m_hashSize; i++)
		{
			(&(m_pHashTable[i]))->~List();
		}
		Deallocate(m_pHashTable);
		m_pHashTable = NULL;
	}
	m_size =0;
	m_pHashTable = NULL;
	m_indexCount =0;
	m_writeIndex=0;
	m_pWriteBuffer=NULL;
	m_writeBufferSize=0;
	Initialize( mmpc.m_hashSize, m_pAllocator);
	for(int i=0; i<m_hashSize; i++)
	{
		m_pHashTable[i] = mmpc.m_pHashTable[i];
	}
	m_size =mmpc.m_size;
#ifdef PAFEE_STATISTICS_ENABLE
	m_Statistics.addCount = mmpc.m_Statistics.addCount;
	m_Statistics.acqCount = mmpc.m_Statistics.acqCount;
	m_Statistics.deleteCount = mmpc.m_Statistics.deleteCount;
	m_Statistics.maxElementCount = mmpc.m_Statistics.maxElementCount;
	m_Statistics.collisionCount = mmpc.m_Statistics.collisionCount;
	m_Statistics.minListCount = mmpc.m_Statistics.minListCount;
	m_Statistics.maxListCount = mmpc.m_Statistics.maxListCount;
	m_Statistics.presentElementCount = mmpc.m_Statistics.presentElementCount;
#endif
	return *this;
}

//----------------------------------------
//! GetWriteBuffer
//! @return: Result(Write Buffer Pointer)
//----------------------------------------
void* MultiMap::GetWriteBuffer(
	const char* szKey, ///< [in] Character Key
	int length ///< [in] Data Length
)
{
	m_writeIndex = Hash( szKey );
	m_writeBufferSize = GetBufferSize(szKey, length);
	m_pWriteBuffer = (char*)m_pHashTable[m_writeIndex].GetWriteBuffer( AligmentSize(m_writeBufferSize) );
	return (void*)SetBuffer(m_pWriteBuffer,szKey,length);
}

//----------------------------------------
//! GetWriteBuffer
//! @return: Result(Write Buffer Pointer)
//----------------------------------------
void* MultiMap::GetWriteBuffer(
	const int key, ///< [in] Number Key
	int length ///< [in] Data Length
)
{
	m_writeIndex = Hash( key );
	m_writeBufferSize = GetBufferSize(key, length);
	m_pWriteBuffer = (char*)m_pHashTable[m_writeIndex].GetWriteBuffer(AligmentSize(m_writeBufferSize));
	return (void*)SetBuffer(m_pWriteBuffer,key,length);
}

//----------------------------------------
//! Set
//! @return: Result(ture:success false:not success)
//----------------------------------------
bool MultiMap::Set(
	const char* szKey, ///< [in] Character Key
	void* pData, ///< [in] Data
	int length ///< [in] Data Length
)
{
	unsigned int index = Hash( szKey );
	char* pSetBuffer;
	int bufferSize;

#ifdef PAFEE_STATISTICS_ENABLE
	if(m_pHashTable[index].GetCount() > 0)
	{
		m_Statistics.collisionCount++;
	}
#endif

	pSetBuffer = this->SetData(szKey, (char *)pData, length, &bufferSize);
	m_pHashTable[index].AddBack( pSetBuffer, bufferSize);
	m_size++;
#ifdef PAFEE_STATISTICS_ENABLE
	m_Statistics.addCount++;
	m_Statistics.presentElementCount++;
	if(m_Statistics.presentElementCount >= m_Statistics.maxElementCount)
	{
		m_Statistics.maxElementCount = m_Statistics.presentElementCount;
	}
#endif
	return true;
}

//----------------------------------------
//! Set
//! @return: Result(ture:success false:not success)
//----------------------------------------
bool MultiMap::Set(
	int key, ///< [in] Number Key
	void* pData, ///< [in] Data
	int length ///< [in] Data Length
)
{
	unsigned int index = Hash( key );
	char* pSetBuffer;
	int bufferSize;
#ifdef PAFEE_STATISTICS_ENABLE
	if(m_pHashTable[index].GetCount() > 0)
	{
		m_Statistics.collisionCount++;
	}
#endif
	pSetBuffer = this->SetData(key, (char *)pData, length, &bufferSize);
	m_pHashTable[index].AddBack( pSetBuffer, bufferSize);
	m_size++;
#ifdef PAFEE_STATISTICS_ENABLE
	m_Statistics.addCount++;
	m_Statistics.presentElementCount++;
	if(m_Statistics.presentElementCount >= m_Statistics.maxElementCount)
	{
		m_Statistics.maxElementCount = m_Statistics.presentElementCount;
	}
#endif
	return true;
 }

//----------------------------------------
//! Set
//! @return: Result(ture:success false:not success)
//----------------------------------------
bool MultiMap::Set(
	int length ///< [in] Data Length
)
{
	if(m_pWriteBuffer == NULL)
	{
		return false;
	}
#ifdef PAFEE_STATISTICS_ENABLE
	if(m_pHashTable[m_writeIndex].GetCount() > 0)
	{
		m_Statistics.collisionCount++;
	}
#endif
	m_pHashTable[m_writeIndex].AddBack(m_pWriteBuffer, m_writeBufferSize);
	m_size++;
#ifdef PAFEE_STATISTICS_ENABLE
	m_Statistics.addCount++;
	m_Statistics.presentElementCount++;
	if(m_Statistics.presentElementCount >= m_Statistics.maxElementCount)
	{
		m_Statistics.maxElementCount = m_Statistics.presentElementCount;
	}
#endif
	return true;
}

//----------------------------------------
//! GetStartPosition
//! @return: Result(Start Position)
//----------------------------------------
MultiMap::POSITION MultiMap::GetStartPosition(void)
{
	MultiMap::POSITION myPosition;
	for(int index=0;index<m_hashSize;index++)
	{
		List::POSITION pos = m_pHashTable[index].GetFrontPosition();
		if ( pos != NULL )
		{
			myPosition.pos = pos;
			myPosition.m_positionIndex = index;
			myPosition.m_pCharkey = NULL;
			myPosition.m_numberKey = -1;
			return myPosition;
		}
	}
	myPosition.pos = NULL;
	myPosition.m_positionIndex = 0;
	myPosition.m_pCharkey = NULL;
	myPosition.m_numberKey = -1;
	return myPosition;
}

//----------------------------------------
//! GetStartPosition
//! @return: Result(Start Position)
//----------------------------------------
MultiMap::POSITION MultiMap::GetStartPosition(
	const char* szKey ///< [in] Character key
)
{
	MultiMap::POSITION myPosition;
	char* keyChar;
	int keyNum;
	char* pData;
	int length;
	unsigned int index = Hash( szKey );
	List::POSITION pos = m_pHashTable[index].GetFrontPosition();
	List::POSITION prepos;
	while(pos)
	{
		prepos = pos;
		pData = (char*)m_pHashTable[index].GetNext(pos,&length);
		keyChar = this->GetKey(pData,&keyNum);
		if(keyChar != NULL)
		{
			if ( pafee_strcmp(keyChar, szKey ) == 0 )
			{
				myPosition.pos = prepos;
				myPosition.m_positionIndex = index;
				myPosition.m_pCharkey = keyChar;
				myPosition.m_numberKey = -1;
				return myPosition;
			}
		}
	}
	myPosition.pos = NULL;
	myPosition.m_positionIndex = 0;
	myPosition.m_pCharkey = szKey;
	myPosition.m_numberKey = -1;
	return myPosition;
}

//----------------------------------------
//! GetStartPosition
//! @return: Result(Start Position)
//----------------------------------------
MultiMap::POSITION MultiMap::GetStartPosition(
	int key ///< [in] Number key
)
{
	MultiMap::POSITION myPosition;
	char* keyChar;
	int keyNum;
	char* pData;
	int length;
	unsigned int index = Hash( key );
	List::POSITION pos = m_pHashTable[index].GetFrontPosition();
	List::POSITION prepos;
	while(pos)
	{
		prepos = pos;
		pData = (char*)m_pHashTable[index].GetNext(pos,&length);
		keyChar = this->GetKey(pData,&keyNum);
		if(keyChar == NULL)
		{
			if (  keyNum == key  )
			{
				myPosition.pos = prepos;
				myPosition.m_positionIndex = index;
				myPosition.m_pCharkey = NULL;
				myPosition.m_numberKey = keyNum;
				return myPosition;
			}
		}
	}
	myPosition.pos = NULL;
	myPosition.m_positionIndex = 0;
	myPosition.m_pCharkey = NULL;
	myPosition.m_numberKey = key;
	return myPosition;
}

//----------------------------------------
//! GetNext
//! @return: Result(ture:success false:not success)
//----------------------------------------
bool MultiMap::GetNext(
	MultiMap::POSITION& pos, ///< [in/out] specified position
	void* pBuffer, ///< [out] Prepared Buffer
	int* pLength, ///< [out] Data Length
	int bufSize ///< [in] Prepared Buffer Length
)
{
	char* pData;
	char* keyChar;
	int keyNum;
	//
	if(pos.pos == NULL)
	{
		pos.m_positionIndex++;
		pos.pos = m_pHashTable[pos.m_positionIndex].GetFrontPosition();
	}
	for( int j=pos.m_positionIndex;j<m_hashSize;j++)
	{
		while(pos.pos)
		{
			pData = (char*)m_pHashTable[j].GetNext(pos.pos,pLength);
			if ( pData != NULL )
			{
				if((pos.m_pCharkey != NULL) || (pos.m_numberKey != -1))
				{
					keyChar = this->GetKey(pData,&keyNum);
					if(keyChar == NULL)
					{
						if (  (pos.m_numberKey != -1) && (keyNum == pos.m_numberKey)  )
						{
							if(bufSize < this->GetDataLength((char*)pData))
							{
								return false;
							}
							*pLength = this->GetData( (char*)pBuffer, (char*)pData );
							return true;
						}
					}
					else
					{
						if ( (pos.m_pCharkey != NULL) && (pafee_strcmp(keyChar, pos.m_pCharkey ) == 0) )
						{
							if(bufSize < this->GetDataLength((char*)pData))
							{
								return false;
							}
							*pLength = this->GetData( (char*)pBuffer, (char*)pData );
							return true;
						}
					}
				}
				else
				{
					if(bufSize < this->GetDataLength((char*)pData))
					{
						return false;
					}
					*pLength = this->GetData( (char*)pBuffer, (char*)pData );
					return true;
				}
			}
		}
		pos.pos = m_pHashTable[j+1].GetFrontPosition();
		pos.m_positionIndex++;
	}
	return false;
}

//----------------------------------------
//! GetNext
//! @return: Result(Data Pointer)
//----------------------------------------
void* MultiMap::GetNext(
	MultiMap::POSITION& pos, ///< [in/out] specified position
	int* pLength ///< [out] Data Length
)
{
	void* pBuffer=NULL;
	char* pData;
	char* keyChar;
	int keyNum;
	//
	if(pos.pos == NULL)
	{
		pos.m_positionIndex++;
		pos.pos = m_pHashTable[pos.m_positionIndex].GetFrontPosition();
	}
	for( int j=pos.m_positionIndex;j<m_hashSize;j++)
	{
		while(pos.pos)
		{
			pData = (char*)m_pHashTable[j].GetNext(pos.pos,pLength);
			if ( pData != NULL )
			{
				if((pos.m_pCharkey != NULL) || (pos.m_numberKey != -1))
				{
					keyChar = this->GetKey(pData,&keyNum);
					if(keyChar == NULL)
					{
						if ( (pos.m_numberKey != -1) && (keyNum == pos.m_numberKey) )
						{
							pBuffer = this->GetData(pData, pLength);
							return pBuffer;
						}
					}
					else
					{
						if ( (pos.m_pCharkey != NULL) && (pafee_strcmp(keyChar, pos.m_pCharkey ) == 0) )
						{
							pBuffer = this->GetData(pData, pLength);
							return pBuffer;
						}
					}
				}
				else
				{
					pBuffer = this->GetData(pData, pLength);
					return pBuffer;
				}
			}
		}
		pos.pos = m_pHashTable[j+1].GetFrontPosition();
		pos.m_positionIndex++;
	}
	return NULL;
}

//----------------------------------------
//! Remove
//! @return: void
//----------------------------------------
void MultiMap::Remove(
	const char* szKey ///< [in] Character Key
)
{
	char* keyChar;
	int keyNum;
	unsigned int index = Hash( szKey );
	char* pData;
	int nextLength;
	//
	List::POSITION prepos;
	List::POSITION pos = m_pHashTable[index].GetFrontPosition();
	while(pos)
	{
		prepos = pos;
		pData = (char*)m_pHashTable[index].GetNext(pos,&nextLength);
		keyChar = this->GetKey(pData,&keyNum);
		if(keyChar != NULL)
		{
			if ( pafee_strcmp(keyChar, szKey ) == 0 )
			{
#ifdef PAFEE_STATISTICS_ENABLE
				if ( m_pHashTable[index].GetCount() >=2 )
				{
					m_Statistics.collisionCount--;
				}
#endif
				m_pHashTable[index].RemoveAt(prepos);
				m_size--;
#ifdef PAFEE_STATISTICS_ENABLE
				m_Statistics.deleteCount++;
				m_Statistics.presentElementCount--;
#endif
			}
		}
	}
	return;
}

//----------------------------------------
//! Remove
//! @return: void
//----------------------------------------
void MultiMap::Remove(
	int key ///< [in] Number Key
)
{
	char* keyChar;
	int keyNum;
	unsigned int index = Hash( key );
	char* pData;
	int nextLength;
	List::POSITION prepos;
	List::POSITION pos = m_pHashTable[index].GetFrontPosition();
	while(pos)
	{
		prepos = pos;
		pData = (char*)m_pHashTable[index].GetNext(pos,&nextLength);
		keyChar = this->GetKey(pData,&keyNum);
		if(keyChar == NULL)
		{
			if (  keyNum == key  )
			{
#ifdef PAFEE_STATISTICS_ENABLE
				if ( m_pHashTable[index].GetCount() >=2 )
				{
					m_Statistics.collisionCount--;
				}
#endif
				m_pHashTable[index].RemoveAt(prepos);
				m_size--;
#ifdef PAFEE_STATISTICS_ENABLE
				m_Statistics.deleteCount++;
				m_Statistics.presentElementCount--;
#endif
			}
		}
	}
	return;
}

//----------------------------------------
//! RemoveAll
//! @return: void
//----------------------------------------
void MultiMap::RemoveAll(void)
{
	for(int i=0;i<m_hashSize;i++)
	{
#ifdef PAFEE_STATISTICS_ENABLE
		m_Statistics.deleteCount+=m_pHashTable[i].GetCount();
		m_Statistics.presentElementCount-=m_pHashTable[i].GetCount();
#endif
		m_pHashTable[i].RemoveAll();
	}
#ifdef PAFEE_STATISTICS_ENABLE
	m_Statistics.collisionCount=0;
#endif
	m_size=0;
	return;
}

//----------------------------------------------------------------------------
//!	GetBufferSize
//! @return: Result(Buffer Size)
//----------------------------------------------------------------------------
int MultiMap::GetBufferSize(
	const int key, ///< [in] Number Key
	int length ///< [in] Data Length
)
{
	char usedCharNum1=0;
	char usedCharNum2=0;
	GetUsedCharSize(key, length, &usedCharNum1, &usedCharNum2);
	return 2+usedCharNum1+usedCharNum2+length;
}

//----------------------------------------------------------------------------
//!	GetBufferSize
//! @return: Result(ture:success false:not success)
//----------------------------------------------------------------------------
int MultiMap::GetBufferSize(
	const char* szKey, ///< [in] Character Key
	int length ///< [in] Data Length
)
{
	char usedCharNum=0;
	char usedCharNum1=0;
	unsigned int usedCharNum2=0;
	GetUsedCharSize(szKey,length,&usedCharNum,&usedCharNum1,&usedCharNum2);
	return 2+usedCharNum+usedCharNum1+usedCharNum2+length;
}

//----------------------------------------------------------------------------
//!	SetBuffer
//! @return: Result(Data Start Pointer)
//----------------------------------------------------------------------------
char* MultiMap::SetBuffer(
	char* pBuffer, ///< [in/out] Set Buffer
	const int key, ///< [in] Number Key
	int length ///< [in] Data Length
)
{
	char* p;
	char usedCharNum1=0;
	char usedCharNum2=0;
	int size;
	//
	GetUsedCharSize(key, length, &usedCharNum1, &usedCharNum2);
	pafee_memset(pBuffer,'\0',2+usedCharNum1+usedCharNum2+length);
	size =  SetDataSub(pBuffer,key,length,usedCharNum1,usedCharNum2);
	//data start pointer
	p = pBuffer;
	p+=size;
	return p;
}

//----------------------------------------------------------------------------
//!	SetBuffer
//! @return: Result(Data Start Pointer)
//----------------------------------------------------------------------------
char* MultiMap::SetBuffer(
	char* pBuffer, ///< [in/out] Set Buffer
	const char* key, ///< [in] Character Key
	int length ///< [in] Data Length
)
{
	char* p;
	char usedCharNum=0;
	char usedCharNum1=0;
	unsigned int usedCharNum2=0;
	int size;
	GetUsedCharSize(key,length,&usedCharNum,&usedCharNum1,&usedCharNum2);
	//buffer
	pafee_memset(pBuffer,'\0',2+usedCharNum+usedCharNum1+usedCharNum2+length);
	size = SetDataSub(pBuffer, key, length, usedCharNum, usedCharNum1, usedCharNum2 );
	//data set
	p = pBuffer;
	p+=size;
	return p;
}

//----------------------------------------------------------------------------
//!	GetKeyType
//! @return: Result(Key Type)
//----------------------------------------------------------------------------
int MultiMap::GetKeyType(
	const char* pElement ///< [in] Target Buffer
)
{
	if ( ( *pElement & 0x80 ) == 0x80 )
	{
		return 1;
	}
	return 0;
}

//----------------------------------------------------------------------------
//!	GetKey
//! @return: Result(Character Key)
//----------------------------------------------------------------------------
char* MultiMap::GetKey( 
	char* pData, ///< [in] Target Buffer
	int* pKeyNum ///< [in/out] Number Key
)
{
	short start1,start3;
	char usedCharNum=0;
	char usedCharNum1=0;
	char usedCharNum2=0;
	int keyLen;
	char* p;
	//
	start1 = ( *(unsigned short*)pData & 0x7C00 ) >>10;
	start3 = *(short*)pData & 0x001F;
	//Character Key
	if ( ( *pData & 0x80 ) == 0x80 )
	{
		//key length
		usedCharNum = start1 - 2;
		p = pData;
		p+=2;
		if(usedCharNum == 1)
		{
			keyLen = *p & 0xFF;
		}
		else if(usedCharNum == 2)
		{
			keyLen = *(short* )p & 0xFFFF;
		}
		else
		{
			keyLen = *(int* )p & 0xFFFFFFFF;
		}
		//data length
		usedCharNum1 = start3 - start1;
		p = pData;
		p+=2+usedCharNum+usedCharNum1;
		*pKeyNum = -1;
		return p;
	}
	else
	{
		//data length
		usedCharNum1 = start1 - 2;
		//key
		usedCharNum2 = start3 - start1;
		p = pData;
		p+=2+usedCharNum1;
		if(usedCharNum2 == 1)
		{
			keyLen = *p & 0xFF;
		}
		else if(usedCharNum2 == 2)
		{
			keyLen = *(short* )p & 0xFFFF;
		}
		else
		{
			keyLen = *(int* )p & 0xFFFFFFFF;
		}
		*pKeyNum = keyLen;
		return NULL;
	}
}

//----------------------------------------------------------------------------
//!	GetData
//! @return: Result(Data Pointer)
//----------------------------------------------------------------------------
char* MultiMap::GetData( 
	char* pIn, ///< [in]  Element Buffer
	int* pLength ///< [in/out] Data Length
)
{
	char* p;
	int lengthToData;
	int dataLength;
	//
	GetDataSub(pIn, &lengthToData, &dataLength);
	//data
	p = pIn;
	p+=lengthToData;
	*pLength = dataLength;
	return p;
}

//----------------------------------------------------------------------------
//!	GetData
//! @return: Result(Data Length)
//----------------------------------------------------------------------------
int MultiMap::GetData(
	char* pOut, ///< [in/out]  Prepared Buffer
	const char* pIn ///< [in] Element Buffer
)
{
	const char* p;
	int lengthToData;
	int dataLength;
	GetDataSub(pIn, &lengthToData, &dataLength);
	//data
	p = pIn;
	p+=lengthToData;
	pafee_memcpy(pOut,p,dataLength);
	return dataLength;
}

//----------------------------------------------------------------------------
//!	GetDataLength
//! @return: Result(Data Length)
//----------------------------------------------------------------------------
int MultiMap::GetDataLength(
	const char* pIn ///< [in]  Element Buffer
)
{
	int lengthToData;
	int dataLength;
	//
	GetDataSub(pIn, &lengthToData, &dataLength);
	return dataLength;
}

//----------------------------------------------------------------------------
//!	SetData
//! @return: Result(Data Pointer)
//----------------------------------------------------------------------------
char* MultiMap::SetData(
	const int key, ///< [in] Number Key
	const char* pData, ///< [in] Data
	int length, ///< [in] Data Length
	int* pBufferSize ///< [in/out] new buffer size
)
{
	char* p;
	char usedCharNum1=0;
	char usedCharNum2=0;
	char* pNewBuffer;
	int size;
	//
	GetUsedCharSize(key, length, &usedCharNum1, &usedCharNum2);
	//create buffer
	*pBufferSize = 2+usedCharNum1+usedCharNum2+length;
	pNewBuffer = BufferAllocate( *pBufferSize );
	pafee_memset(pNewBuffer,'\0',2+usedCharNum1+usedCharNum2+length);
	size =  SetDataSub(pNewBuffer,key,length,usedCharNum1,usedCharNum2);
	//data set
	p = pNewBuffer;
	p+=size;
	pafee_memcpy(p,pData,length);
	return pNewBuffer;
}

//----------------------------------------------------------------------------
//!	SetData
//! @return: Result(new data buffer pointer)
//----------------------------------------------------------------------------
char* MultiMap::SetData(
	const char* szKey, ///< [in] Character Key
	const char* pData, ///< [in] Data
	int length, ///< [in] Data Length
	int* pBufferSize ///< [in/out] New Buffer Size
)
{
	char* p;
	char usedCharNum=0;
	char usedCharNum1=0;
	unsigned int usedCharNum2=0;
	char* pNewBuffer;
	int size;
	//
	GetUsedCharSize(szKey,length,&usedCharNum,&usedCharNum1,&usedCharNum2);
	//create buffer
	*pBufferSize = 2+usedCharNum+usedCharNum1+usedCharNum2+length;
	pNewBuffer = BufferAllocate( *pBufferSize );
	pafee_memset(pNewBuffer,'\0',2+usedCharNum+usedCharNum1+usedCharNum2+length);
	size = SetDataSub(pNewBuffer, szKey, length, usedCharNum, usedCharNum1, usedCharNum2 );
	//data set
	p = pNewBuffer;
	p+=size;
	pafee_memcpy(p,pData,length);
	return pNewBuffer;
}

//----------------------------------------------------------------------------
//!	GetUsedCharSize
//! @return: Result(void)
//----------------------------------------------------------------------------
void MultiMap::GetUsedCharSize(
	const int key, ///< [in] Number Key
	int length, ///< [in] Data Length
	char* usedCharNum1, ///< [out] Data Length Used Size
	char* usedCharNum2 ///< [out] Key Used Size
)
{
	int maskBit;
	int usedBitCount1=0;
	int usedBitCount2=0;
	//
	//mask bit
	maskBit = 0xFFFFFFFF;
	//data length
	while(1)
	{
		usedBitCount1++;
		maskBit=maskBit<<1;
		if((length & maskBit) == 0)
		{
			break;
		}
	}
	if(usedBitCount1 <= 8)
	{
		*usedCharNum1 = 1;
	}
	else if(usedBitCount1 <= 16)
	{
		*usedCharNum1 = 2;
	}
	else
	{
		*usedCharNum1 = 4;
	}
	//key length
	maskBit = 0xFFFFFFFF;
	while(1)
	{
		usedBitCount2++;
		maskBit=maskBit<<1;
		if((key & maskBit) == 0)
		{
			break;
		}
	}
	if(usedBitCount2 <= 8)
	{
		*usedCharNum2 = 1;
	}
	else if(usedBitCount2 <= 16)
	{
		*usedCharNum2 = 2;
	}
	else
	{
		*usedCharNum2 = 4;
	}
}

//----------------------------------------------------------------------------
//!	GetUsedCharSize
//! @return: Result(void)
//----------------------------------------------------------------------------
void MultiMap::GetUsedCharSize(
	const char* szKey, ///< [in] Character Key
	int length, ///< [in] Data Length
	char* usedCharNum, ///< [out] Key Length Used Size
	char* usedCharNum1, ///< [out] Data Length Used Size
	unsigned int* usedCharNum2 ///< [out] Key Used Size
)
{
	int maskBit;
	int usedBitCount=0;
	int usedBitCount1=0;
	//
	//mask bit
	maskBit = 0xFFFFFFFF;
	//key length
	*usedCharNum2 = pafee_strlen(szKey)+1;
	//key used length
	while(1)
	{
		usedBitCount++;
		maskBit=maskBit<<1;
		if((*usedCharNum2 & maskBit) == 0)
		{
			break;
		}
	}
	if(usedBitCount <= 8)
	{
		*usedCharNum = 1;
	}
	else if(usedBitCount <= 16)
	{
		*usedCharNum = 2;
	}
	else
	{
		*usedCharNum = 4;
	}
	//data length
	maskBit = 0xFFFFFFFF;
	while(1)
	{
		usedBitCount1++;
		maskBit=maskBit<<1;
		if((length & maskBit) == 0)
		{
			break;
		}
	}
	if(usedBitCount1 <= 8)
	{
		*usedCharNum1 = 1;
	}
	else if(usedBitCount1 <= 16)
	{
		*usedCharNum1 = 2;
	}
	else
	{
		*usedCharNum1 = 4;
	}
}

//----------------------------------------------------------------------------
//!	GetDataSub
//! @return: Result(void)
//----------------------------------------------------------------------------
void MultiMap::GetDataSub(
	const char* pBuffer, ///< [in] Data Buffer
	int* pLengthToData, ///< [out] Length to Data
	int* pDataLength ///< [out] Data Length
)
{
	short start1,start3;
	char usedCharNum=0;
	char usedCharNum1=0;
	char usedCharNum2=0;
	const char* p;
	int keyLen;
	int dataLen;
	//
	//get data length start position
	start1 = ( *(unsigned short*)pBuffer & 0x7C00 ) >>10;
	//get data start position
	start3 = *(short*)pBuffer & 0x001F;
	//Character Key
	if ( ( *pBuffer & 0x80 ) == 0x80 )
	{
		//key length
		usedCharNum = start1 - 2;
		p = pBuffer;
		p+=2;
		if(usedCharNum == 1)
		{
			keyLen = *p & 0xFF;
		}
		else if(usedCharNum == 2)
		{
			keyLen = *(short* )p & 0xFFFF;
		}
		else
		{
			keyLen = *(int* )p & 0xFFFFFFFF;
		}
		//data length
		usedCharNum1 = start3 - start1;
		p = pBuffer;
		p+=2+usedCharNum;
		if(usedCharNum1 == 1)
		{
			dataLen = *p & 0xFF;
		}
		else if(usedCharNum1 == 2)
		{
			dataLen = *(short* )p & 0xFFFF;
		}
		else
		{
			dataLen = *(int* )p & 0xFFFFFFFF;
		}
		//data
		*pLengthToData = 2+usedCharNum+usedCharNum1+keyLen;
		*pDataLength = dataLen;
	}
	else
	{
		//data length
		usedCharNum1 = start1 - 2;
		p = pBuffer;
		p+=2;
		if(usedCharNum1 == 1)
		{
			dataLen = *p & 0xFF;
		}
		else if(usedCharNum1 == 2)
		{
			dataLen = *(short* )p & 0xFFFF;
		}
		else
		{
			dataLen = *(int* )p & 0xFFFFFFFF;
		}
		//key
		usedCharNum2 = start3 - start1;
		//data
		*pLengthToData = 2+usedCharNum1+usedCharNum2;
		*pDataLength = dataLen;
	}
}

//----------------------------------------------------------------------------
//!	SetDataSub
//! @return: Result(int: length to data)
//----------------------------------------------------------------------------
int MultiMap::SetDataSub(
	char* pBuffer, ///< [in/out] Set Data Buffer
	const int key, ///< [in] Number Key
	int length, ///< [in] Data Length
	char usedCharNum1, ///< [in] Used Char Num
	char usedCharNum2 ///< [in] Used Char Num
)
{
	short startPosShort;
	char* p;
	//key type set
	*pBuffer = *pBuffer | 0x00;
	//key start position set
	startPosShort = 2+usedCharNum1;
	startPosShort<<=10;
	*(unsigned short* )pBuffer = *(unsigned short* )pBuffer | startPosShort;
	//data start position set
	startPosShort = 2+usedCharNum1+usedCharNum2;
	*(short* )pBuffer = *(short* )pBuffer | startPosShort;
	//data length set
	p = pBuffer;
	p+=2;
	if(usedCharNum1 == 1)
	{
		*p = *p | length;
	}
	else if(usedCharNum1 == 2)
	{
		*(short* )p = *(short* )p | length;
	}
	else
	{
		*(int* )p = *(int* )p | length;
	}
	//key set
	p = pBuffer;
	p+=2+usedCharNum1;
	if(usedCharNum2 == 1)
	{
		*p = *p | key;
	}
	else if(usedCharNum2 == 2)
	{
		*(short* )p = *(short* )p | key;
	}
	else
	{
		*(int* )p = *(int* )p | key;
	}
	//data start pointer
	return 2+usedCharNum1+usedCharNum2;
}

//----------------------------------------------------------------------------
//!	SetDataSub
//! @return: Result(int: length to data)
//----------------------------------------------------------------------------
int MultiMap::SetDataSub(
	char* pBuffer, ///< [in/out] Set Data Buffer
	const char* szKey, ///< [in] Character Key
	int length, ///< [in] Data Length
	char usedCharNum, ///< [in]  Used Char Num
	char usedCharNum1, ///< [in]  Used Char Num
	unsigned int usedCharNum2 ///< [in]  Used Char Num
)
{
	short startPosShort;
	char* p;
	//
	//key type set
	*pBuffer = *pBuffer | 0x80;
	//data length start position set
	startPosShort = 2+usedCharNum;
	startPosShort<<=10;
	*(unsigned short* )pBuffer = *(unsigned short* )pBuffer | startPosShort;
	//key start position
	startPosShort = 2+usedCharNum+usedCharNum1;
	*(short* )pBuffer = *(short* )pBuffer | startPosShort;
	//key length set
	p = pBuffer;
	p+=2;
	if(usedCharNum == 1)
	{
		*p = *p | usedCharNum2;
	}
	else if(usedCharNum == 2)
	{
		*(short* )p = *(short* )p | usedCharNum2;
	}
	else
	{
		*(int* )p = *(int* )p | usedCharNum2;
	}
	//data length set
	p = pBuffer;
	p+=2+usedCharNum;
	if(usedCharNum1 == 1)
	{
		*p = *p | length;
	}
	else if(usedCharNum1 == 2)
	{
		*(short* )p = *(short* )p | length;
	}
	else
	{
		*(int* )p = *(int* )p | length;
	}
	//key set
	p = pBuffer;
	p+=2+usedCharNum+usedCharNum1;
	pafee_memcpy(p,szKey,usedCharNum2);
	//data set
	return 2+usedCharNum+usedCharNum1+usedCharNum2;
}

#ifdef PAFEE_STATISTICS_ENABLE
//----------------------------------------
//! GetStatistics
//! @return: void
//----------------------------------------
void MultiMap::GetStatistics(
	MultiMap::STATISTICS* pStatistics ///< [out] statistics Pointer
)
{
	MultiMap::POSITION pos = GetStartPosition();
	for(int j=pos.m_positionIndex; j<m_hashSize; j++)
	{
		int listElementCount = m_pHashTable[j].GetCount();
		if(j == pos.m_positionIndex)
		{
			m_Statistics.minListCount = listElementCount;
			m_Statistics.maxListCount = listElementCount;
		}
		else
		{
			//min
			if(listElementCount < m_Statistics.minListCount)
			{
				m_Statistics.minListCount = listElementCount;
			}
			//max
			if(m_Statistics.maxListCount < listElementCount)
			{
				m_Statistics.maxListCount = listElementCount;
			}
		}
	}
	pafee_memcpy ( pStatistics, &m_Statistics, sizeof (MultiMap::STATISTICS) );
}
#endif //PAFEE_STATISTICS_ENABLE

#ifdef PAFEE_UNITTEST

//----------------------------------------------------------------------------
//!	Unit test
//! @return: void
//----------------------------------------------------------------------------
void MultiMap::UnitTest(void)
{
	{
		const int HASHSIZE=4;
		int key;
		int jj;
		int size = 0;
		char* pSetData1 = "XXXXXXX";
		char* pSetData2 = "YYYYYYY";
		char* pSetData3 = "ZZZZZZZ";
		int len;
		char work[5];
		char work2[16];
		char* pGetData2;
		char* pGetData3;
		char cBuffer[2];
		//dchar *pSkey;
		//STATISTICS
		int addCount=0;
		int acqCount=0;
		int deleteCount=0;
		int maxElementCount=0;
		int collisionCount=0;
		int minListCount=0;
		int maxListCount=0;
		int presentElementCount=0;
		int collisionarray[HASHSIZE];
#ifdef PAFEE_STATISTICS_ENABLE
		MultiMap::STATISTICS statisticsRtn;
#endif
		pafee_memset(work,'\0',sizeof(work));
		pafee_memset(work2,'\0',sizeof(work2));
		//Initialize
		MultiMap mp1(5);
		assert(!mp1.Initialize(10,NULL));
		//test Initialize/GetCount/IsEmpty/GetWriteBuffer/Set(int length)
		MultiMap mp;
		mp.Initialize(HASHSIZE,NULL);
		len = 7;
		assert(mp.GetCount()==0);
		assert(mp.IsEmpty());
#ifdef PAFEE_STATISTICS_ENABLE
		mp.GetStatistics(&statisticsRtn);
		assert(statisticsRtn.addCount == addCount);
		assert(statisticsRtn.acqCount == acqCount);
		assert(statisticsRtn.deleteCount == deleteCount);
		assert(statisticsRtn.maxElementCount == maxElementCount);
		assert(statisticsRtn.collisionCount == collisionCount);
		assert(statisticsRtn.minListCount == minListCount);
		assert(statisticsRtn.maxListCount == maxListCount);
		assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		for (jj=0;jj<HASHSIZE;jj++)
		{
			collisionarray[jj]=0;
		}
		for (key=0;key<6;key++)
		{
			char* p= (char*)mp.GetWriteBuffer(key,len);
			pafee_memcpy(p,"XXXXXXX",len);
			assert(mp.Set(len));
			addCount++;
			presentElementCount++;
			maxElementCount++;
			collisionarray[mp.Hash(key)]++;
		}
	//	collisionCount = presentElementCount-4;
	//	minListCount = presentElementCount / 4;
	//	maxListCount = (presentElementCount / 4)+1;
		collisionCount=0;
		minListCount=INT_MAX;
		maxListCount=0;
		int eh;
		for (eh=0;eh<HASHSIZE;eh++)
		{
			if ( collisionarray[eh] >= 1 )
			{
				collisionCount+=collisionarray[eh]-1;
			}
			if ( minListCount > collisionarray[eh] )
			{
				minListCount = collisionarray[eh];
			}
			if ( maxListCount < collisionarray[eh] )
			{
				maxListCount = collisionarray[eh];
			}
		}
		assert(mp.GetCount() == 6);
		assert(!mp.IsEmpty());
#ifdef PAFEE_STATISTICS_ENABLE
		mp.GetStatistics(&statisticsRtn);
		assert(statisticsRtn.addCount == addCount);
		assert(statisticsRtn.acqCount == acqCount);
		assert(statisticsRtn.deleteCount == deleteCount);
		assert(statisticsRtn.maxElementCount == maxElementCount);
		assert(statisticsRtn.collisionCount == collisionCount);
		assert(statisticsRtn.minListCount == minListCount);
		assert(statisticsRtn.maxListCount == maxListCount);
		assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		//test Set(int key)
		for (key=0;key<6;key++)
		{
			assert(mp.Set(key,pSetData2,pafee_strlen(pSetData2)));
			addCount++;
			presentElementCount++;
			maxElementCount++;
			collisionarray[mp.Hash(key)]++;
		}
		for (key=0;key<6;key++)
		{
			assert(mp.Set(key,pSetData3,pafee_strlen(pSetData3)));
			addCount++;
			presentElementCount++;
			maxElementCount++;
			collisionarray[mp.Hash(key)]++;
		}
		//collisionCount = 14;
		//minListCount = 3;
		//maxListCount = 6;
		collisionCount=0;
		minListCount=INT_MAX;
		maxListCount=0;
		for (eh=0;eh<HASHSIZE;eh++)
		{
			if ( collisionarray[eh] >= 1 )
			{
				collisionCount+=collisionarray[eh]-1;
			}
			if ( minListCount > collisionarray[eh] )
			{
				minListCount = collisionarray[eh];
			}
			if ( maxListCount < collisionarray[eh] )
			{
				maxListCount = collisionarray[eh];
			}
		}
#ifdef PAFEE_STATISTICS_ENABLE
		mp.GetStatistics(&statisticsRtn);
		assert(statisticsRtn.addCount == addCount);
		assert(statisticsRtn.acqCount == acqCount);
		assert(statisticsRtn.deleteCount == deleteCount);
		assert(statisticsRtn.maxElementCount == maxElementCount);
		assert(statisticsRtn.collisionCount == collisionCount);
		assert(statisticsRtn.minListCount == minListCount);
		assert(statisticsRtn.maxListCount == maxListCount);
		assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		//GetStartPosition(int key)/GetNext
		key = 5;
		MultiMap::POSITION pos = mp.GetStartPosition(key);
		assert(mp.GetNext(pos,work2,&len,sizeof(work2)));
		assert(pafee_memcmp(pSetData1,work2,pafee_strlen(pSetData1)) == 0);
		assert(len == (int)pafee_strlen(pSetData1));
		//GetNext(No copy version)
		pGetData2 = (char* )mp.GetNext(pos,&len);
		assert(pGetData2 != NULL);
		assert(pafee_memcmp(pSetData2,pGetData2,pafee_strlen(pSetData2)) == 0);
		assert(len == (int)pafee_strlen(pSetData2));
		//GetNext(No copy version)
		pGetData3 = (char* )mp.GetNext(pos,&len);
		assert(pGetData3 != NULL);
		assert(pafee_memcmp(pSetData3,pGetData3,pafee_strlen(pSetData3)) == 0);
		assert(len == (int)pafee_strlen(pSetData3));
		//GetStartPosition()/GetNext
		size = mp.GetCount();
		assert(size == 18);
		pos = mp.GetStartPosition();
	/*
		for(jj=0;jj<size;jj++)
		{
			assert(mp.GetNext(pos,work2,&len,sizeof(work2)));
			if((jj == 0) || (jj == 1) || (jj == 6) || (jj == 7) || (jj == 12) || (jj == 15))
			{
				assert(pafee_memcmp(pSetData1,work2,pafee_strlen(pSetData1)) == 0);//NG
			}
			else if((jj == 2) || (jj == 3) || (jj == 8) || (jj == 9) || (jj == 13) || (jj == 16))
			{
				assert(pafee_memcmp(pSetData2,work2,pafee_strlen(pSetData2)) == 0);
			}
			else
			{
				assert(pafee_memcmp(pSetData3,work2,pafee_strlen(pSetData3)) == 0);
			}
		}
	*/
#ifdef PAFEE_STATISTICS_ENABLE
		mp.GetStatistics(&statisticsRtn);
		assert(statisticsRtn.addCount == addCount);
		assert(statisticsRtn.acqCount == acqCount);
		assert(statisticsRtn.deleteCount == deleteCount);
		assert(statisticsRtn.maxElementCount == maxElementCount);
		assert(statisticsRtn.collisionCount == collisionCount);
		assert(statisticsRtn.minListCount == minListCount);
		assert(statisticsRtn.maxListCount == maxListCount);
		assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		//Remove
		key=5;
		mp.Remove(key);
		collisionarray[mp.Hash(key)]--;
		collisionarray[mp.Hash(key)]--;
		collisionarray[mp.Hash(key)]--;
		assert(mp.GetCount() == 15);
		assert(!mp.IsEmpty());
		deleteCount+=3; 
		presentElementCount-=3;
		collisionCount=0;
		minListCount=INT_MAX;
		maxListCount=0;
		for (eh=0;eh<HASHSIZE;eh++)
		{
			if ( collisionarray[eh] >= 1 )
			{
				collisionCount+=collisionarray[eh]-1;
			}
			if ( minListCount > collisionarray[eh] )
			{
				minListCount = collisionarray[eh];
			}
			if ( maxListCount < collisionarray[eh] )
			{
				maxListCount = collisionarray[eh];
			}
		}
#ifdef PAFEE_STATISTICS_ENABLE
		mp.GetStatistics(&statisticsRtn);
		assert(statisticsRtn.addCount == addCount);
		assert(statisticsRtn.acqCount == acqCount);
		assert(statisticsRtn.deleteCount == deleteCount);
		assert(statisticsRtn.maxElementCount == maxElementCount);
		assert(statisticsRtn.collisionCount == collisionCount);
		assert(statisticsRtn.minListCount == minListCount);
		assert(statisticsRtn.maxListCount == maxListCount);
		assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		key=2;
		mp.Remove(key);
		collisionarray[mp.Hash(key)]--;
		collisionarray[mp.Hash(key)]--;
		collisionarray[mp.Hash(key)]--;
		assert(mp.GetCount() == 12);
		assert(!mp.IsEmpty());
		deleteCount+=3;
		presentElementCount-=3;
		minListCount -= 3;
		collisionCount=0;
		minListCount=INT_MAX;
		maxListCount=0;
		for (eh=0;eh<HASHSIZE;eh++)
		{
			if ( collisionarray[eh] >= 1 )
			{
				collisionCount+=collisionarray[eh]-1;
			}
			if ( minListCount > collisionarray[eh] )
			{
				minListCount = collisionarray[eh];
			}
			if ( maxListCount < collisionarray[eh] )
			{
				maxListCount = collisionarray[eh];
			}
		}
#ifdef PAFEE_STATISTICS_ENABLE
		mp.GetStatistics(&statisticsRtn);
		assert(statisticsRtn.addCount == addCount);
		assert(statisticsRtn.acqCount == acqCount);
		assert(statisticsRtn.deleteCount == deleteCount);
		assert(statisticsRtn.maxElementCount == maxElementCount);
		assert(statisticsRtn.collisionCount == collisionCount);
		assert(statisticsRtn.minListCount == minListCount);
		assert(statisticsRtn.maxListCount == maxListCount);
		assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		key=0;
		mp.Remove(key);
		collisionarray[mp.Hash(key)]--;
		collisionarray[mp.Hash(key)]--;
		collisionarray[mp.Hash(key)]--;
		assert(mp.GetCount() == 9);
		assert(!mp.IsEmpty());
		deleteCount+=3;
		presentElementCount-=3;
		maxListCount-=3;
		collisionCount=0;
		minListCount=INT_MAX;
		maxListCount=0;
		for (eh=0;eh<HASHSIZE;eh++)
		{
			if ( collisionarray[eh] >= 1 )
			{
				collisionCount+=collisionarray[eh]-1;
			}
			if ( minListCount > collisionarray[eh] )
			{
				minListCount = collisionarray[eh];
			}
			if ( maxListCount < collisionarray[eh] )
			{
				maxListCount = collisionarray[eh];
			}
		}
#ifdef PAFEE_STATISTICS_ENABLE
		mp.GetStatistics(&statisticsRtn);
		assert(statisticsRtn.addCount == addCount);
		assert(statisticsRtn.acqCount == acqCount);
		assert(statisticsRtn.deleteCount == deleteCount);
		assert(statisticsRtn.maxElementCount == maxElementCount);
		assert(statisticsRtn.collisionCount == collisionCount);
		assert(statisticsRtn.minListCount == minListCount);
		assert(statisticsRtn.maxListCount == maxListCount);
		assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		//RemoveAll
		mp.RemoveAll();
		for (jj=0;jj<HASHSIZE;jj++)
		{
			collisionarray[jj]=0;
		}
		assert(!mp.GetCount());
		assert(mp.IsEmpty());
		deleteCount += presentElementCount;
		minListCount = 0;
		maxListCount = 0;
		presentElementCount = 0;
		collisionCount=0;
#ifdef PAFEE_STATISTICS_ENABLE
		mp.GetStatistics(&statisticsRtn);
		assert(statisticsRtn.addCount == addCount);
		assert(statisticsRtn.acqCount == acqCount);
		assert(statisticsRtn.deleteCount == deleteCount);
		assert(statisticsRtn.maxElementCount == maxElementCount);
		assert(statisticsRtn.collisionCount == collisionCount);
		assert(statisticsRtn.minListCount == minListCount);
		assert(statisticsRtn.maxListCount == maxListCount);
		assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		//Set/GetCount
		for (key=0;key<6;key++)
		{
			assert(mp.Set(key,pSetData1,pafee_strlen(pSetData1)));
			addCount++;
			presentElementCount++;
			collisionarray[mp.Hash(key)]++;
		}
		//minListCount = presentElementCount / 4;
		//maxListCount = (presentElementCount / 4)+1;
		//collisionCount = collisionCount + (6-4);
		collisionCount=0;
		minListCount=INT_MAX;
		maxListCount=0;
		for (eh=0;eh<HASHSIZE;eh++)
		{
			if ( collisionarray[eh] >= 1 )
			{
				collisionCount+=collisionarray[eh]-1;
			}
			if ( minListCount > collisionarray[eh] )
			{
				minListCount = collisionarray[eh];
			}
			if ( maxListCount < collisionarray[eh] )
			{
				maxListCount = collisionarray[eh];
			}
		}
#ifdef PAFEE_STATISTICS_ENABLE
		mp.GetStatistics(&statisticsRtn);
		assert(statisticsRtn.addCount == addCount);
		assert(statisticsRtn.acqCount == acqCount);
		assert(statisticsRtn.deleteCount == deleteCount);
		assert(statisticsRtn.maxElementCount == maxElementCount);
		assert(statisticsRtn.collisionCount == collisionCount);
		assert(statisticsRtn.minListCount == minListCount);
		assert(statisticsRtn.maxListCount == maxListCount);
		assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		mp.RemoveAll();
		for (jj=0;jj<HASHSIZE;jj++)
		{
			collisionarray[jj]=0;
		}
		deleteCount += presentElementCount;
		minListCount = 0;
		maxListCount = 0;
		presentElementCount = 0;
		//test pattern 2
		//test GetCount/IsEmpty/GetWriteBuffer/Set(int length)
		assert(mp.GetCount()==0);
		assert(mp.IsEmpty());
		for (key=0;key<6;key++)
		{
			cBuffer[0] = (char)(key + 49);
			//pSkey = &c;
			cBuffer[1] = '\0';
			char* p= (char*)mp.GetWriteBuffer(cBuffer,len);
			pafee_memcpy(p,"XXXXXXX",len);
			assert(mp.Set(len));
			addCount++;
			presentElementCount++;
			collisionarray[mp.Hash(cBuffer)]++;
		}
		collisionCount += (6/4)+1;
		minListCount += 6/4;
		maxListCount += (6/4)+1;
		collisionCount=0;
		minListCount=INT_MAX;
		maxListCount=0;
		for (eh=0;eh<HASHSIZE;eh++)
		{
			if ( collisionarray[eh] >= 1 )
			{
				collisionCount+=collisionarray[eh]-1;
			}
			if ( minListCount > collisionarray[eh] )
			{
				minListCount = collisionarray[eh];
			}
			if ( maxListCount < collisionarray[eh] )
			{
				maxListCount = collisionarray[eh];
			}
		}
		assert(mp.GetCount() == 6);
		assert(!mp.IsEmpty());
#ifdef PAFEE_STATISTICS_ENABLE
		mp.GetStatistics(&statisticsRtn);
		assert(statisticsRtn.addCount == addCount);
		assert(statisticsRtn.acqCount == acqCount);
		assert(statisticsRtn.deleteCount == deleteCount);
		assert(statisticsRtn.maxElementCount == maxElementCount);
		assert(statisticsRtn.collisionCount == collisionCount);
		assert(statisticsRtn.minListCount == minListCount);
		assert(statisticsRtn.maxListCount == maxListCount);
		assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		//test Set(character key)
		for (key=0;key<6;key++)
		{
			cBuffer[0] = (char)(key + 49);
			//pSkey = &c;
			cBuffer[1] = '\0';
			assert(mp.Set(cBuffer,pSetData2,pafee_strlen(pSetData2)));
			addCount++;
			presentElementCount++;
			collisionarray[mp.Hash(cBuffer)]++;
		}
		for (key=0;key<6;key++)
		{
			cBuffer[0] = (char)(key + 49);
			//pSkey = &c;
			cBuffer[1] = '\0';
			assert(mp.Set(cBuffer,pSetData3,pafee_strlen(pSetData3)));
			addCount++;
			presentElementCount++;
			collisionarray[mp.Hash(cBuffer)]++;
		}
		//collisionCount += 12;
		//minListCount = 3;
		//maxListCount = 6;
		collisionCount=0;
		minListCount=INT_MAX;
		maxListCount=0;
		for (eh=0;eh<HASHSIZE;eh++)
		{
			if ( collisionarray[eh] >= 1 )
			{
				collisionCount+=collisionarray[eh]-1;
			}
			if ( minListCount > collisionarray[eh] )
			{
				minListCount = collisionarray[eh];
			}
			if ( maxListCount < collisionarray[eh] )
			{
				maxListCount = collisionarray[eh];
			}
		}
#ifdef PAFEE_STATISTICS_ENABLE
		mp.GetStatistics(&statisticsRtn);
		assert(statisticsRtn.addCount == addCount);
		assert(statisticsRtn.acqCount == acqCount);
		assert(statisticsRtn.deleteCount == deleteCount);
		assert(statisticsRtn.maxElementCount == maxElementCount);
		assert(statisticsRtn.collisionCount == collisionCount);
		assert(statisticsRtn.minListCount == minListCount);
		assert(statisticsRtn.maxListCount == maxListCount);
		assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		//GetStartPosition(character key)/GetNext
		cBuffer[0] = (char)(1 + 49);
		//pSkey = &c;
		cBuffer[1] = '\0';
		pos = mp.GetStartPosition(cBuffer);
		assert(mp.GetNext(pos,work2,&len,sizeof(work2)));
		assert(pafee_memcmp(pSetData1,work2,pafee_strlen(pSetData1)) == 0);
		assert(len == (int)pafee_strlen(pSetData1));
		//GetNext(No copy version)
		pGetData2 = (char* )mp.GetNext(pos,&len);
		assert(pGetData2 != NULL);
		assert(pafee_memcmp(pSetData2,pGetData2,pafee_strlen(pSetData2)) == 0);
		assert(len == (int)pafee_strlen(pSetData2));
		//GetNext(No copy version)
		pGetData3 = (char* )mp.GetNext(pos,&len);
		assert(pGetData3 != NULL);
		assert(pafee_memcmp(pSetData3,pGetData3,pafee_strlen(pSetData3)) == 0);
		assert(len == (int)pafee_strlen(pSetData3));
		//GetStartPosition()/GetNext
		size = mp.GetCount();
		assert(size == 18);
		pos = mp.GetStartPosition();
		/*
		for(jj=0;jj<size;jj++)
		{
			assert(mp.GetNext(pos,work2,&len,sizeof(work2)));
			if((jj == 0) || (jj == 3) || (jj == 4) || (jj == 9) || (jj == 10) || (jj == 15))
			{
				assert(pafee_memcmp(pSetData1,work2,pafee_strlen(pSetData1)) == 0);
			}
			else if((jj == 1) || (jj == 5) || (jj == 6) || (jj == 11) || (jj == 12) || (jj == 16))
			{
				assert(pafee_memcmp(pSetData2,work2,pafee_strlen(pSetData2)) == 0);
			}
			else
			{
				assert(pafee_memcmp(pSetData3,work2,pafee_strlen(pSetData3)) == 0);
			}
		}
		*/
#ifdef PAFEE_STATISTICS_ENABLE
		mp.GetStatistics(&statisticsRtn);
		assert(statisticsRtn.addCount == addCount);
		assert(statisticsRtn.acqCount == acqCount);
		assert(statisticsRtn.deleteCount == deleteCount);
		assert(statisticsRtn.maxElementCount == maxElementCount);
		assert(statisticsRtn.collisionCount == collisionCount);
		assert(statisticsRtn.minListCount == minListCount);
		assert(statisticsRtn.maxListCount == maxListCount);
		assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		//Remove
		cBuffer[0] = (char)(1 + 49);
		//pSkey = &c;
		cBuffer[1] = '\0';
		mp.Remove(cBuffer);
		collisionarray[mp.Hash(cBuffer)]--;
		collisionarray[mp.Hash(cBuffer)]--;
		collisionarray[mp.Hash(cBuffer)]--;
		assert(mp.GetCount() == 15);
		assert(!mp.IsEmpty());
		deleteCount+=3;
		presentElementCount-=3;
		collisionCount=0;
		minListCount=INT_MAX;
		maxListCount=0;
		for (eh=0;eh<HASHSIZE;eh++)
		{
			if ( collisionarray[eh] >= 1 )
			{
				collisionCount+=collisionarray[eh]-1;
			}
			if ( minListCount > collisionarray[eh] )
			{
				minListCount = collisionarray[eh];
			}
			if ( maxListCount < collisionarray[eh] )
			{
				maxListCount = collisionarray[eh];
			}
		}
#ifdef PAFEE_STATISTICS_ENABLE
		mp.GetStatistics(&statisticsRtn);
		assert(statisticsRtn.addCount == addCount);
		assert(statisticsRtn.acqCount == acqCount);
		assert(statisticsRtn.deleteCount == deleteCount);
		assert(statisticsRtn.maxElementCount == maxElementCount);
		assert(statisticsRtn.collisionCount == collisionCount);
		assert(statisticsRtn.minListCount == minListCount);
		assert(statisticsRtn.maxListCount == maxListCount);
		assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		cBuffer[0] = (char)(2 + 49);
		//pSkey = &c;
		cBuffer[1] = '\0';
		mp.Remove(cBuffer);
		collisionarray[mp.Hash(cBuffer)]--;
		collisionarray[mp.Hash(cBuffer)]--;
		collisionarray[mp.Hash(cBuffer)]--;
		assert(mp.GetCount() == 12);
		assert(!mp.IsEmpty());
		deleteCount+=3;
		minListCount-=3;
		presentElementCount-=3;
		collisionCount=0;
		minListCount=INT_MAX;
		maxListCount=0;
		for (eh=0;eh<HASHSIZE;eh++)
		{
			if ( collisionarray[eh] >= 1 )
			{
				collisionCount+=collisionarray[eh]-1;
			}
			if ( minListCount > collisionarray[eh] )
			{
				minListCount = collisionarray[eh];
			}
			if ( maxListCount < collisionarray[eh] )
			{
				maxListCount = collisionarray[eh];
			}
		}
#ifdef PAFEE_STATISTICS_ENABLE
		mp.GetStatistics(&statisticsRtn);
		assert(statisticsRtn.addCount == addCount);
		assert(statisticsRtn.acqCount == acqCount);
		assert(statisticsRtn.deleteCount == deleteCount);
		assert(statisticsRtn.maxElementCount == maxElementCount);
		assert(statisticsRtn.collisionCount == collisionCount);
		assert(statisticsRtn.minListCount == minListCount);
		assert(statisticsRtn.maxListCount == maxListCount);
		assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		//RemoveAll
		mp.RemoveAll();
		for (jj=0;jj<HASHSIZE;jj++)
		{
			collisionarray[jj]=0;
		}
		assert(!mp.GetCount());
		assert(mp.IsEmpty());
		deleteCount += presentElementCount;
		minListCount = 0;
		maxListCount = 0;
		presentElementCount = 0;
		collisionCount=0;
#ifdef PAFEE_STATISTICS_ENABLE
		mp.GetStatistics(&statisticsRtn);
		assert(statisticsRtn.addCount == addCount);
		assert(statisticsRtn.acqCount == acqCount);
		assert(statisticsRtn.deleteCount == deleteCount);
		assert(statisticsRtn.maxElementCount == maxElementCount);
		assert(statisticsRtn.collisionCount == collisionCount);
		assert(statisticsRtn.minListCount == minListCount);
		assert(statisticsRtn.maxListCount == maxListCount);
		assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		//test pattern 3
		//Set(Number Key)
		for (key=0;key<6;key++)
		{
			assert(mp.Set(key,pSetData1,pafee_strlen(pSetData1)));
			addCount++;
			presentElementCount++;
			collisionarray[mp.Hash(key)]++;
		}
		//Set(Character Key)
		for (key=0;key<6;key++)
		{
			cBuffer[0] = (char)(key + 49);
			//pSkey = &c;
			cBuffer[1] = '\0';
			assert(mp.Set(cBuffer,pSetData1,pafee_strlen(pSetData1)));
			addCount++;
			presentElementCount++;
			collisionarray[mp.Hash(cBuffer)]++;
		}
		//Set(Number Key)
		for (key=0;key<6;key++)
		{
			assert(mp.Set(key,pSetData2,pafee_strlen(pSetData2)));
			addCount++;
			presentElementCount++;
			collisionarray[mp.Hash(key)]++;
		}
		//Set(Character Key)
		for (key=0;key<6;key++)
		{
			cBuffer[0] = (char)(key + 49);
			//pSkey = &c;
			cBuffer[1] = '\0';
			assert(mp.Set(cBuffer,pSetData2,pafee_strlen(pSetData2)));
			addCount++;
			presentElementCount++;
			collisionarray[mp.Hash(cBuffer)]++;
		}
		//Set(Number Key)
		for (key=0;key<6;key++)
		{
			assert(mp.Set(key,pSetData3,pafee_strlen(pSetData3)));
			addCount++;
			presentElementCount++;
			collisionarray[mp.Hash(key)]++;
		}
		//Set(Character Key)
		for (key=0;key<6;key++)
		{
			cBuffer[0] = (char)(key + 49);
			//pSkey = &c;
			cBuffer[1] = '\0';
			assert(mp.Set(cBuffer,pSetData3,pafee_strlen(pSetData3)));
			addCount++;
			presentElementCount++;
			collisionarray[mp.Hash(cBuffer)]++;
		}
		maxElementCount = 36;
		//collisionCount += 32;
		//minListCount = 6;
		//maxListCount = 12;
		collisionCount=0;
		minListCount=INT_MAX;
		maxListCount=0;
		for (eh=0;eh<HASHSIZE;eh++)
		{
			if ( collisionarray[eh] >= 1 )
			{
				collisionCount+=collisionarray[eh]-1;
			}
			if ( minListCount > collisionarray[eh] )
			{
				minListCount = collisionarray[eh];
			}
			if ( maxListCount < collisionarray[eh] )
			{
				maxListCount = collisionarray[eh];
			}
		}
#ifdef PAFEE_STATISTICS_ENABLE
		mp.GetStatistics(&statisticsRtn);
		assert(statisticsRtn.addCount == addCount);
		assert(statisticsRtn.acqCount == acqCount);
		assert(statisticsRtn.deleteCount == deleteCount);
		assert(statisticsRtn.maxElementCount == maxElementCount);
		assert(statisticsRtn.collisionCount == collisionCount);
		assert(statisticsRtn.minListCount == minListCount);
		assert(statisticsRtn.maxListCount == maxListCount);
		assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		//GetStartPosition(int key)/GetNext
		key = 5;
		pos = mp.GetStartPosition(key);
		assert(mp.GetNext(pos,work2,&len,sizeof(work2)));
		assert(pafee_memcmp(pSetData1,work2,pafee_strlen(pSetData1)) == 0);
		assert(len == (int)pafee_strlen(pSetData1));
		//GetNext(No copy version)
		pGetData2 = (char* )mp.GetNext(pos,&len);
		assert(pGetData2 != NULL);
		assert(pafee_memcmp(pSetData2,pGetData2,pafee_strlen(pSetData2)) == 0);
		assert(len == (int)pafee_strlen(pSetData2));
		//GetNext(No copy version)
		pGetData3 = (char* )mp.GetNext(pos,&len);
		assert(pGetData3 != NULL);
		assert(pafee_memcmp(pSetData3,pGetData3,pafee_strlen(pSetData3)) == 0);
		assert(len == (int)pafee_strlen(pSetData3));
		//GetStartPosition(character key)/GetNext
		cBuffer[0] = (char)(1 + 49);
		//pSkey = &c;
		cBuffer[1] = '\0';
		pos = mp.GetStartPosition(cBuffer);
		assert(mp.GetNext(pos,work2,&len,sizeof(work2)));
		assert(pafee_memcmp(pSetData1,work2,pafee_strlen(pSetData1)) == 0);
		assert(len == (int)pafee_strlen(pSetData1));
		//GetNext(No copy version)
		pGetData2 = (char* )mp.GetNext(pos,&len);
		assert(pGetData2 != NULL);
		assert(pafee_memcmp(pSetData2,pGetData2,pafee_strlen(pSetData2)) == 0);
		assert(len == (int)pafee_strlen(pSetData2));
		//GetNext(No copy version)
		pGetData3 = (char* )mp.GetNext(pos,&len);
		assert(pGetData3 != NULL);
		assert(pafee_memcmp(pSetData3,pGetData3,pafee_strlen(pSetData3)) == 0);
		assert(len == (int)pafee_strlen(pSetData3));
		//GetStartPosition()/GetNext
		size = mp.GetCount();
		assert(size == 36);
		pos = mp.GetStartPosition();
		/*
		for(jj=0;jj<size;jj++)
		{
			assert(mp.GetNext(pos,work2,&len,sizeof(work2)));
			if((jj == 0) || (jj == 1) || (jj == 2) || (jj == 9) || (jj == 10) || (jj == 11) || (jj == 12) || (jj == 21) || (jj == 22) || (jj == 23) || (jj == 30) || (jj == 31))
			{
				assert(pafee_memcmp(pSetData1,work2,pafee_strlen(pSetData1)) == 0);
			}
			else if((jj == 3) || (jj == 4) || (jj == 5) || (jj == 13) || (jj == 14) || (jj == 15) || (jj == 16) || (jj == 24) || (jj == 25) || (jj == 26) || (jj == 32) || (jj == 33))
			{
				assert(pafee_memcmp(pSetData2,work2,pafee_strlen(pSetData2)) == 0);
			}
			else
			{
				assert(pafee_memcmp(pSetData3,work2,pafee_strlen(pSetData3)) == 0);
			}
		}
		*/
#ifdef PAFEE_STATISTICS_ENABLE
		mp.GetStatistics(&statisticsRtn);
		assert(statisticsRtn.addCount == addCount);
		assert(statisticsRtn.acqCount == acqCount);
		assert(statisticsRtn.deleteCount == deleteCount);
		assert(statisticsRtn.maxElementCount == maxElementCount);
		assert(statisticsRtn.collisionCount == collisionCount);
		assert(statisticsRtn.minListCount == minListCount);
		assert(statisticsRtn.maxListCount == maxListCount);
		assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		//Remove
		key=5;
		mp.Remove(key);
		collisionarray[mp.Hash(key)]--;
		collisionarray[mp.Hash(key)]--;
		collisionarray[mp.Hash(key)]--;
		assert(mp.GetCount() == 33);
		assert(!mp.IsEmpty());
		deleteCount+=3;
		presentElementCount-=3;
		maxListCount-=3;
		collisionCount=0;
		minListCount=INT_MAX;
		maxListCount=0;
		for (eh=0;eh<HASHSIZE;eh++)
		{
			if ( collisionarray[eh] >= 1 )
			{
				collisionCount+=collisionarray[eh]-1;
			}
			if ( minListCount > collisionarray[eh] )
			{
				minListCount = collisionarray[eh];
			}
			if ( maxListCount < collisionarray[eh] )
			{
				maxListCount = collisionarray[eh];
			}
		}
#ifdef PAFEE_STATISTICS_ENABLE
		mp.GetStatistics(&statisticsRtn);
		assert(statisticsRtn.addCount == addCount);
		assert(statisticsRtn.acqCount == acqCount);
		assert(statisticsRtn.deleteCount == deleteCount);
		assert(statisticsRtn.maxElementCount == maxElementCount);
		assert(statisticsRtn.collisionCount == collisionCount);
		assert(statisticsRtn.minListCount == minListCount);
		assert(statisticsRtn.maxListCount == maxListCount);
		assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		//Remove
		cBuffer[0] = (char)(2 + 49);
		//pSkey = &c;
		cBuffer[1] = '\0';
		mp.Remove(cBuffer);
		collisionarray[mp.Hash(cBuffer)]--;
		collisionarray[mp.Hash(cBuffer)]--;
		collisionarray[mp.Hash(cBuffer)]--;
		assert(mp.GetCount() == 30);
		assert(!mp.IsEmpty());
		deleteCount+=3;
		minListCount-=3;
		presentElementCount-=3;
		collisionCount=0;
		minListCount=INT_MAX;
		maxListCount=0;
		for (eh=0;eh<HASHSIZE;eh++)
		{
			if ( collisionarray[eh] >= 1 )
			{
				collisionCount+=collisionarray[eh]-1;
			}
			if ( minListCount > collisionarray[eh] )
			{
				minListCount = collisionarray[eh];
			}
			if ( maxListCount < collisionarray[eh] )
			{
				maxListCount = collisionarray[eh];
			}
		}
#ifdef PAFEE_STATISTICS_ENABLE
		mp.GetStatistics(&statisticsRtn);
		assert(statisticsRtn.addCount == addCount);
		assert(statisticsRtn.acqCount == acqCount);
		assert(statisticsRtn.deleteCount == deleteCount);
		assert(statisticsRtn.maxElementCount == maxElementCount);
		assert(statisticsRtn.collisionCount == collisionCount);
		assert(statisticsRtn.minListCount == minListCount);
		assert(statisticsRtn.maxListCount == maxListCount);
		assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		//RemoveAll
		mp.RemoveAll();
		for (jj=0;jj<HASHSIZE;jj++)
		{
			collisionarray[jj]=0;
		}
		assert(!mp.GetCount());
		assert(mp.IsEmpty());
		deleteCount += presentElementCount;
		minListCount = 0;
		maxListCount = 0;
		presentElementCount = 0;
		collisionCount=0;
#ifdef PAFEE_STATISTICS_ENABLE
		mp.GetStatistics(&statisticsRtn);
		assert(statisticsRtn.addCount == addCount);
		assert(statisticsRtn.acqCount == acqCount);
		assert(statisticsRtn.deleteCount == deleteCount);
		assert(statisticsRtn.maxElementCount == maxElementCount);
		assert(statisticsRtn.collisionCount == collisionCount);
		assert(statisticsRtn.minListCount == minListCount);
		assert(statisticsRtn.maxListCount == maxListCount);
		assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		//Set(Number Key)
		for (key=0;key<6;key++)
		{
			assert(mp.Set(key,pSetData1,pafee_strlen(pSetData1)));
			addCount++;
			presentElementCount++;
			collisionarray[mp.Hash(key)]++;
		}
		//Set(Character Key)
		for (key=0;key<6;key++)
		{
			cBuffer[0] = (char)(key + 49);
			//pSkey = &c;
			cBuffer[1] = '\0';
			assert(mp.Set(cBuffer,pSetData1,pafee_strlen(pSetData1)));
			addCount++;
			presentElementCount++;
			collisionarray[mp.Hash(cBuffer)]++;
		}
		//collisionCount+=8;
		//minListCount = 2;
		//maxListCount = 4;
		collisionCount=0;
		minListCount=INT_MAX;
		maxListCount=0;
		for (eh=0;eh<HASHSIZE;eh++)
		{
			if ( collisionarray[eh] >= 1 )
			{
				collisionCount+=collisionarray[eh]-1;
			}
			if ( minListCount > collisionarray[eh] )
			{
				minListCount = collisionarray[eh];
			}
			if ( maxListCount < collisionarray[eh] )
			{
				maxListCount = collisionarray[eh];
			}
		}
#ifdef PAFEE_STATISTICS_ENABLE
		mp.GetStatistics(&statisticsRtn);
		assert(statisticsRtn.addCount == addCount);
		assert(statisticsRtn.acqCount == acqCount);
		assert(statisticsRtn.deleteCount == deleteCount);
		assert(statisticsRtn.maxElementCount == maxElementCount);
		assert(statisticsRtn.collisionCount == collisionCount);
		assert(statisticsRtn.minListCount == minListCount);
		assert(statisticsRtn.maxListCount == maxListCount);
		assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
	}
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
	{
		const int HASHSIZE=4;
		int key;
		int jj;
		int size = 0;
		char* pSetData1 = "XXXXXXX";
		char* pSetData2 = "YYYYYYY";
		char* pSetData3 = "ZZZZZZZ";
		int len;
		char work[5];
		char work2[16];
		char* pGetData2;
		char* pGetData3;
		char cBuffer[2];
		//char *pSkey;
		//STATISTICS
		int addCount=0;
		int acqCount=0;
		int deleteCount=0;
		int maxElementCount=0;
		int collisionCount=0;
		int minListCount=0;
		int maxListCount=0;
		int presentElementCount=0;
		int collisionarray[HASHSIZE];
#ifdef PAFEE_STATISTICS_ENABLE
		MultiMap::STATISTICS statisticsRtn;
#endif
		pafee_memset(work,'\0',sizeof(work));
		pafee_memset(work2,'\0',sizeof(work2));
		//Initialize
		MultiMap mp1(5);
		assert(!mp1.Initialize(10,NULL));
		//test Initialize/GetCount/IsEmpty/GetWriteBuffer/Set(int length)
		MultiMap mp0c;
		mp0c.Initialize(HASHSIZE,NULL);
		len = 7;
		assert(mp0c.GetCount()==0);
		assert(mp0c.IsEmpty());
#ifdef PAFEE_STATISTICS_ENABLE
		mp0c.GetStatistics(&statisticsRtn);
		assert(statisticsRtn.addCount == addCount);
		assert(statisticsRtn.acqCount == acqCount);
		assert(statisticsRtn.deleteCount == deleteCount);
		assert(statisticsRtn.maxElementCount == maxElementCount);
		assert(statisticsRtn.collisionCount == collisionCount);
		assert(statisticsRtn.minListCount == minListCount);
		assert(statisticsRtn.maxListCount == maxListCount);
		assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		for (jj=0;jj<HASHSIZE;jj++)
		{
			collisionarray[jj]=0;
		}
		for (key=0;key<6;key++)
		{
			char* p= (char*)mp0c.GetWriteBuffer(key,len);
			pafee_memcpy(p,"XXXXXXX",len);
			assert(mp0c.Set(len));
			addCount++;
			presentElementCount++;
			maxElementCount++;
			collisionarray[mp0c.Hash(key)]++;
		}
	//	collisionCount = presentElementCount-4;
	//	minListCount = presentElementCount / 4;
	//	maxListCount = (presentElementCount / 4)+1;
		collisionCount=0;
		minListCount=INT_MAX;
		maxListCount=0;
		int eh;
		for (eh=0;eh<HASHSIZE;eh++)
		{
			if ( collisionarray[eh] >= 1 )
			{
				collisionCount+=collisionarray[eh]-1;
			}
			if ( minListCount > collisionarray[eh] )
			{
				minListCount = collisionarray[eh];
			}
			if ( maxListCount < collisionarray[eh] )
			{
				maxListCount = collisionarray[eh];
			}
		}
		assert(mp0c.GetCount() == 6);
		assert(!mp0c.IsEmpty());
#ifdef PAFEE_STATISTICS_ENABLE
		mp0c.GetStatistics(&statisticsRtn);
		assert(statisticsRtn.addCount == addCount);
		assert(statisticsRtn.acqCount == acqCount);
		assert(statisticsRtn.deleteCount == deleteCount);
		assert(statisticsRtn.maxElementCount == maxElementCount);
		assert(statisticsRtn.collisionCount == collisionCount);
		assert(statisticsRtn.minListCount == minListCount);
		assert(statisticsRtn.maxListCount == maxListCount);
		assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		//test Set(int key)
		for (key=0;key<6;key++)
		{
			assert(mp0c.Set(key,pSetData2,pafee_strlen(pSetData2)));
			addCount++;
			presentElementCount++;
			maxElementCount++;
			collisionarray[mp0c.Hash(key)]++;
		}
		for (key=0;key<6;key++)
		{
			assert(mp0c.Set(key,pSetData3,pafee_strlen(pSetData3)));
			addCount++;
			presentElementCount++;
			maxElementCount++;
			collisionarray[mp0c.Hash(key)]++;
		}
		//collisionCount = 14;
		//minListCount = 3;
		//maxListCount = 6;
		collisionCount=0;
		minListCount=INT_MAX;
		maxListCount=0;
		for (eh=0;eh<HASHSIZE;eh++)
		{
			if ( collisionarray[eh] >= 1 )
			{
				collisionCount+=collisionarray[eh]-1;
			}
			if ( minListCount > collisionarray[eh] )
			{
				minListCount = collisionarray[eh];
			}
			if ( maxListCount < collisionarray[eh] )
			{
				maxListCount = collisionarray[eh];
			}
		}
#ifdef PAFEE_STATISTICS_ENABLE
		mp0c.GetStatistics(&statisticsRtn);
		assert(statisticsRtn.addCount == addCount);
		assert(statisticsRtn.acqCount == acqCount);
		assert(statisticsRtn.deleteCount == deleteCount);
		assert(statisticsRtn.maxElementCount == maxElementCount);
		assert(statisticsRtn.collisionCount == collisionCount);
		assert(statisticsRtn.minListCount == minListCount);
		assert(statisticsRtn.maxListCount == maxListCount);
		assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		//MultiMap mp=mp0c;
		MultiMap mp;
		mp=mp0c;
		//GetStartPosition(int key)/GetNext
		key = 5;
		MultiMap::POSITION pos = mp.GetStartPosition(key);
		assert(mp.GetNext(pos,work2,&len,sizeof(work2)));
		assert(pafee_memcmp(pSetData1,work2,pafee_strlen(pSetData1)) == 0);
		assert(len == (int)pafee_strlen(pSetData1));
		//GetNext(No copy version)
		pGetData2 = (char* )mp.GetNext(pos,&len);
		assert(pGetData2 != NULL);
		assert(pafee_memcmp(pSetData2,pGetData2,pafee_strlen(pSetData2)) == 0);
		assert(len == (int)pafee_strlen(pSetData2));
		//GetNext(No copy version)
		pGetData3 = (char* )mp.GetNext(pos,&len);
		assert(pGetData3 != NULL);
		assert(pafee_memcmp(pSetData3,pGetData3,pafee_strlen(pSetData3)) == 0);
		assert(len == (int)pafee_strlen(pSetData3));
		//GetStartPosition()/GetNext
		size = mp.GetCount();
		assert(size == 18);
		pos = mp.GetStartPosition();
	/*
		for(jj=0;jj<size;jj++)
		{
			assert(mp.GetNext(pos,work2,&len,sizeof(work2)));
			if((jj == 0) || (jj == 1) || (jj == 6) || (jj == 7) || (jj == 12) || (jj == 15))
			{
				assert(pafee_memcmp(pSetData1,work2,pafee_strlen(pSetData1)) == 0);//NG
			}
			else if((jj == 2) || (jj == 3) || (jj == 8) || (jj == 9) || (jj == 13) || (jj == 16))
			{
				assert(pafee_memcmp(pSetData2,work2,pafee_strlen(pSetData2)) == 0);
			}
			else
			{
				assert(pafee_memcmp(pSetData3,work2,pafee_strlen(pSetData3)) == 0);
			}
		}
	*/
#ifdef PAFEE_STATISTICS_ENABLE
		mp.GetStatistics(&statisticsRtn);
		assert(statisticsRtn.addCount == addCount);
		assert(statisticsRtn.acqCount == acqCount);
		assert(statisticsRtn.deleteCount == deleteCount);
		assert(statisticsRtn.maxElementCount == maxElementCount);
		assert(statisticsRtn.collisionCount == collisionCount);
		assert(statisticsRtn.minListCount == minListCount);
		assert(statisticsRtn.maxListCount == maxListCount);
		assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		//Remove
		key=5;
		mp.Remove(key);
		collisionarray[mp.Hash(key)]--;
		collisionarray[mp.Hash(key)]--;
		collisionarray[mp.Hash(key)]--;
		assert(mp.GetCount() == 15);
		assert(!mp.IsEmpty());
		deleteCount+=3; 
		presentElementCount-=3;
		collisionCount=0;
		minListCount=INT_MAX;
		maxListCount=0;
		for (eh=0;eh<HASHSIZE;eh++)
		{
			if ( collisionarray[eh] >= 1 )
			{
				collisionCount+=collisionarray[eh]-1;
			}
			if ( minListCount > collisionarray[eh] )
			{
				minListCount = collisionarray[eh];
			}
			if ( maxListCount < collisionarray[eh] )
			{
				maxListCount = collisionarray[eh];
			}
		}
#ifdef PAFEE_STATISTICS_ENABLE
		mp.GetStatistics(&statisticsRtn);
		assert(statisticsRtn.addCount == addCount);
		assert(statisticsRtn.acqCount == acqCount);
		assert(statisticsRtn.deleteCount == deleteCount);
		assert(statisticsRtn.maxElementCount == maxElementCount);
		assert(statisticsRtn.collisionCount == collisionCount);
		assert(statisticsRtn.minListCount == minListCount);
		assert(statisticsRtn.maxListCount == maxListCount);
		assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		key=2;
		mp.Remove(key);
		collisionarray[mp.Hash(key)]--;
		collisionarray[mp.Hash(key)]--;
		collisionarray[mp.Hash(key)]--;
		assert(mp.GetCount() == 12);
		assert(!mp.IsEmpty());
		deleteCount+=3;
		presentElementCount-=3;
		minListCount -= 3;
		collisionCount=0;
		minListCount=INT_MAX;
		maxListCount=0;
		for (eh=0;eh<HASHSIZE;eh++)
		{
			if ( collisionarray[eh] >= 1 )
			{
				collisionCount+=collisionarray[eh]-1;
			}
			if ( minListCount > collisionarray[eh] )
			{
				minListCount = collisionarray[eh];
			}
			if ( maxListCount < collisionarray[eh] )
			{
				maxListCount = collisionarray[eh];
			}
		}
#ifdef PAFEE_STATISTICS_ENABLE
		mp.GetStatistics(&statisticsRtn);
		assert(statisticsRtn.addCount == addCount);
		assert(statisticsRtn.acqCount == acqCount);
		assert(statisticsRtn.deleteCount == deleteCount);
		assert(statisticsRtn.maxElementCount == maxElementCount);
		assert(statisticsRtn.collisionCount == collisionCount);
		assert(statisticsRtn.minListCount == minListCount);
		assert(statisticsRtn.maxListCount == maxListCount);
		assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		key=0;
		mp.Remove(key);
		collisionarray[mp.Hash(key)]--;
		collisionarray[mp.Hash(key)]--;
		collisionarray[mp.Hash(key)]--;
		assert(mp.GetCount() == 9);
		assert(!mp.IsEmpty());
		deleteCount+=3;
		presentElementCount-=3;
		maxListCount-=3;
		collisionCount=0;
		minListCount=INT_MAX;
		maxListCount=0;
		for (eh=0;eh<HASHSIZE;eh++)
		{
			if ( collisionarray[eh] >= 1 )
			{
				collisionCount+=collisionarray[eh]-1;
			}
			if ( minListCount > collisionarray[eh] )
			{
				minListCount = collisionarray[eh];
			}
			if ( maxListCount < collisionarray[eh] )
			{
				maxListCount = collisionarray[eh];
			}
		}
#ifdef PAFEE_STATISTICS_ENABLE
		mp.GetStatistics(&statisticsRtn);
		assert(statisticsRtn.addCount == addCount);
		assert(statisticsRtn.acqCount == acqCount);
		assert(statisticsRtn.deleteCount == deleteCount);
		assert(statisticsRtn.maxElementCount == maxElementCount);
		assert(statisticsRtn.collisionCount == collisionCount);
		assert(statisticsRtn.minListCount == minListCount);
		assert(statisticsRtn.maxListCount == maxListCount);
		assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		//RemoveAll
		mp.RemoveAll();
		for (jj=0;jj<HASHSIZE;jj++)
		{
			collisionarray[jj]=0;
		}
		assert(!mp.GetCount());
		assert(mp.IsEmpty());
		deleteCount += presentElementCount;
		minListCount = 0;
		maxListCount = 0;
		presentElementCount = 0;
		collisionCount=0;
#ifdef PAFEE_STATISTICS_ENABLE
		mp.GetStatistics(&statisticsRtn);
		assert(statisticsRtn.addCount == addCount);
		assert(statisticsRtn.acqCount == acqCount);
		assert(statisticsRtn.deleteCount == deleteCount);
		assert(statisticsRtn.maxElementCount == maxElementCount);
		assert(statisticsRtn.collisionCount == collisionCount);
		assert(statisticsRtn.minListCount == minListCount);
		assert(statisticsRtn.maxListCount == maxListCount);
		assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		//Set/GetCount
		for (key=0;key<6;key++)
		{
			assert(mp.Set(key,pSetData1,pafee_strlen(pSetData1)));
			addCount++;
			presentElementCount++;

			collisionarray[mp.Hash(key)]++;
		}
		//minListCount = presentElementCount / 4;
		//maxListCount = (presentElementCount / 4)+1;
		//collisionCount = collisionCount + (6-4);
		collisionCount=0;
		minListCount=INT_MAX;
		maxListCount=0;
		for (eh=0;eh<HASHSIZE;eh++)
		{
			if ( collisionarray[eh] >= 1 )
			{
				collisionCount+=collisionarray[eh]-1;
			}
			if ( minListCount > collisionarray[eh] )
			{
				minListCount = collisionarray[eh];
			}
			if ( maxListCount < collisionarray[eh] )
			{
				maxListCount = collisionarray[eh];
			}
		}
#ifdef PAFEE_STATISTICS_ENABLE
		mp.GetStatistics(&statisticsRtn);
		assert(statisticsRtn.addCount == addCount);
		assert(statisticsRtn.acqCount == acqCount);
		assert(statisticsRtn.deleteCount == deleteCount);
		assert(statisticsRtn.maxElementCount == maxElementCount);
		assert(statisticsRtn.collisionCount == collisionCount);
		assert(statisticsRtn.minListCount == minListCount);
		assert(statisticsRtn.maxListCount == maxListCount);
		assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		mp.RemoveAll();
		for (jj=0;jj<HASHSIZE;jj++)
		{
			collisionarray[jj]=0;
		}
		deleteCount += presentElementCount;
		minListCount = 0;
		maxListCount = 0;
		presentElementCount = 0;
		//test pattern 2
		//test GetCount/IsEmpty/GetWriteBuffer/Set(int length)
		assert(mp.GetCount()==0);
		assert(mp.IsEmpty());
		for (key=0;key<6;key++)
		{
			cBuffer[0] = (char)(key + 49);
			//pSkey = &c;
			cBuffer[1] = '\0';
			char* p= (char*)mp.GetWriteBuffer(cBuffer,len);
			pafee_memcpy(p,"XXXXXXX",len);
			assert(mp.Set(len));
			addCount++;
			presentElementCount++;
			collisionarray[mp.Hash(cBuffer)]++;
		}
		collisionCount += (6/4)+1;
		minListCount += 6/4;
		maxListCount += (6/4)+1;
		collisionCount=0;
		minListCount=INT_MAX;
		maxListCount=0;
		for (eh=0;eh<HASHSIZE;eh++)
		{
			if ( collisionarray[eh] >= 1 )
			{
				collisionCount+=collisionarray[eh]-1;
			}
			if ( minListCount > collisionarray[eh] )
			{
				minListCount = collisionarray[eh];
			}
			if ( maxListCount < collisionarray[eh] )
			{
				maxListCount = collisionarray[eh];
			}
		}
		assert(mp.GetCount() == 6);
		assert(!mp.IsEmpty());
#ifdef PAFEE_STATISTICS_ENABLE
		mp.GetStatistics(&statisticsRtn);
		assert(statisticsRtn.addCount == addCount);
		assert(statisticsRtn.acqCount == acqCount);
		assert(statisticsRtn.deleteCount == deleteCount);
		assert(statisticsRtn.maxElementCount == maxElementCount);
		assert(statisticsRtn.collisionCount == collisionCount);
		assert(statisticsRtn.minListCount == minListCount);
		assert(statisticsRtn.maxListCount == maxListCount);
		assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		//test Set(character key)
		for (key=0;key<6;key++)
		{
			cBuffer[0] = (char)(key + 49);
			//pSkey = &c;
			cBuffer[1] = '\0';
			assert(mp.Set(cBuffer,pSetData2,pafee_strlen(pSetData2)));
			addCount++;
			presentElementCount++;
			collisionarray[mp.Hash(cBuffer)]++;
		}
		for (key=0;key<6;key++)
		{
			cBuffer[0] = (char)(key + 49);
			//pSkey = &c;
			cBuffer[1] = '\0';
			assert(mp.Set(cBuffer,pSetData3,pafee_strlen(pSetData3)));
			addCount++;
			presentElementCount++;
			collisionarray[mp.Hash(cBuffer)]++;
		}
		//collisionCount += 12;
		//minListCount = 3;
		//maxListCount = 6;
		collisionCount=0;
		minListCount=INT_MAX;
		maxListCount=0;
		for (eh=0;eh<HASHSIZE;eh++)
		{
			if ( collisionarray[eh] >= 1 )
			{
				collisionCount+=collisionarray[eh]-1;
			}
			if ( minListCount > collisionarray[eh] )
			{
				minListCount = collisionarray[eh];
			}
			if ( maxListCount < collisionarray[eh] )
			{
				maxListCount = collisionarray[eh];
			}
		}
#ifdef PAFEE_STATISTICS_ENABLE
		mp.GetStatistics(&statisticsRtn);
		assert(statisticsRtn.addCount == addCount);
		assert(statisticsRtn.acqCount == acqCount);
		assert(statisticsRtn.deleteCount == deleteCount);
		assert(statisticsRtn.maxElementCount == maxElementCount);
		assert(statisticsRtn.collisionCount == collisionCount);
		assert(statisticsRtn.minListCount == minListCount);
		assert(statisticsRtn.maxListCount == maxListCount);
		assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		//GetStartPosition(character key)/GetNext
		cBuffer[0] = (char)(1 + 49);
		//pSkey = &c;
		cBuffer[1] = '\0';
		pos = mp.GetStartPosition(cBuffer);
		assert(mp.GetNext(pos,work2,&len,sizeof(work2)));
		assert(pafee_memcmp(pSetData1,work2,pafee_strlen(pSetData1)) == 0);
		assert(len == (int)pafee_strlen(pSetData1));
		//GetNext(No copy version)
		pGetData2 = (char* )mp.GetNext(pos,&len);
		assert(pGetData2 != NULL);
		assert(pafee_memcmp(pSetData2,pGetData2,pafee_strlen(pSetData2)) == 0);
		assert(len == (int)pafee_strlen(pSetData2));
		//GetNext(No copy version)
		pGetData3 = (char* )mp.GetNext(pos,&len);
		assert(pGetData3 != NULL);
		assert(pafee_memcmp(pSetData3,pGetData3,pafee_strlen(pSetData3)) == 0);
		assert(len == (int)pafee_strlen(pSetData3));
		//GetStartPosition()/GetNext
		size = mp.GetCount();
		assert(size == 18);
		pos = mp.GetStartPosition();
		/*
		for(jj=0;jj<size;jj++)
		{
			assert(mp.GetNext(pos,work2,&len,sizeof(work2)));
			if((jj == 0) || (jj == 3) || (jj == 4) || (jj == 9) || (jj == 10) || (jj == 15))
			{
				assert(pafee_memcmp(pSetData1,work2,pafee_strlen(pSetData1)) == 0);
			}
			else if((jj == 1) || (jj == 5) || (jj == 6) || (jj == 11) || (jj == 12) || (jj == 16))
			{
				assert(pafee_memcmp(pSetData2,work2,pafee_strlen(pSetData2)) == 0);
			}
			else
			{
				assert(pafee_memcmp(pSetData3,work2,pafee_strlen(pSetData3)) == 0);
			}
		}
		*/
#ifdef PAFEE_STATISTICS_ENABLE
		mp.GetStatistics(&statisticsRtn);
		assert(statisticsRtn.addCount == addCount);
		assert(statisticsRtn.acqCount == acqCount);
		assert(statisticsRtn.deleteCount == deleteCount);
		assert(statisticsRtn.maxElementCount == maxElementCount);
		assert(statisticsRtn.collisionCount == collisionCount);
		assert(statisticsRtn.minListCount == minListCount);
		assert(statisticsRtn.maxListCount == maxListCount);
		assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		//Remove
		cBuffer[0] = (char)(1 + 49);
		//pSkey = &c;
		cBuffer[1] = '\0';
		mp.Remove(cBuffer);
		collisionarray[mp.Hash(cBuffer)]--;
		collisionarray[mp.Hash(cBuffer)]--;
		collisionarray[mp.Hash(cBuffer)]--;
		assert(mp.GetCount() == 15);
		assert(!mp.IsEmpty());
		deleteCount+=3;
		presentElementCount-=3;
		collisionCount=0;
		minListCount=INT_MAX;
		maxListCount=0;
		for (eh=0;eh<HASHSIZE;eh++)
		{
			if ( collisionarray[eh] >= 1 )
			{
				collisionCount+=collisionarray[eh]-1;
			}
			if ( minListCount > collisionarray[eh] )
			{
				minListCount = collisionarray[eh];
			}
			if ( maxListCount < collisionarray[eh] )
			{
				maxListCount = collisionarray[eh];
			}
		}
#ifdef PAFEE_STATISTICS_ENABLE
		mp.GetStatistics(&statisticsRtn);
		assert(statisticsRtn.addCount == addCount);
		assert(statisticsRtn.acqCount == acqCount);
		assert(statisticsRtn.deleteCount == deleteCount);
		assert(statisticsRtn.maxElementCount == maxElementCount);
		assert(statisticsRtn.collisionCount == collisionCount);
		assert(statisticsRtn.minListCount == minListCount);
		assert(statisticsRtn.maxListCount == maxListCount);
		assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		cBuffer[0] = (char)(2 + 49);
		//pSkey = &c;
		cBuffer[1] = '\0';
		mp.Remove(cBuffer);
		collisionarray[mp.Hash(cBuffer)]--;
		collisionarray[mp.Hash(cBuffer)]--;
		collisionarray[mp.Hash(cBuffer)]--;
		assert(mp.GetCount() == 12);
		assert(!mp.IsEmpty());
		deleteCount+=3;
		minListCount-=3;
		presentElementCount-=3;
		collisionCount=0;
		minListCount=INT_MAX;
		maxListCount=0;
		for (eh=0;eh<HASHSIZE;eh++)
		{
			if ( collisionarray[eh] >= 1 )
			{
				collisionCount+=collisionarray[eh]-1;
			}
			if ( minListCount > collisionarray[eh] )
			{
				minListCount = collisionarray[eh];
			}
			if ( maxListCount < collisionarray[eh] )
			{
				maxListCount = collisionarray[eh];
			}
		}
#ifdef PAFEE_STATISTICS_ENABLE
		mp.GetStatistics(&statisticsRtn);
		assert(statisticsRtn.addCount == addCount);
		assert(statisticsRtn.acqCount == acqCount);
		assert(statisticsRtn.deleteCount == deleteCount);
		assert(statisticsRtn.maxElementCount == maxElementCount);
		assert(statisticsRtn.collisionCount == collisionCount);
		assert(statisticsRtn.minListCount == minListCount);
		assert(statisticsRtn.maxListCount == maxListCount);
		assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		//RemoveAll
		mp.RemoveAll();
		for (jj=0;jj<HASHSIZE;jj++)
		{
			collisionarray[jj]=0;
		}
		assert(!mp.GetCount());
		assert(mp.IsEmpty());
		deleteCount += presentElementCount;
		minListCount = 0;
		maxListCount = 0;
		presentElementCount = 0;
		collisionCount=0;
#ifdef PAFEE_STATISTICS_ENABLE
		mp.GetStatistics(&statisticsRtn);
		assert(statisticsRtn.addCount == addCount);
		assert(statisticsRtn.acqCount == acqCount);
		assert(statisticsRtn.deleteCount == deleteCount);
		assert(statisticsRtn.maxElementCount == maxElementCount);
		assert(statisticsRtn.collisionCount == collisionCount);
		assert(statisticsRtn.minListCount == minListCount);
		assert(statisticsRtn.maxListCount == maxListCount);
		assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		//test pattern 3
		//Set(Number Key)
		for (key=0;key<6;key++)
		{
			assert(mp.Set(key,pSetData1,pafee_strlen(pSetData1)));
			addCount++;
			presentElementCount++;
			collisionarray[mp.Hash(key)]++;
		}
		//Set(Character Key)
		for (key=0;key<6;key++)
		{
			cBuffer[0] = (char)(key + 49);
			//pSkey = &c;
			cBuffer[1] = '\0';
			assert(mp.Set(cBuffer,pSetData1,pafee_strlen(pSetData1)));
			addCount++;
			presentElementCount++;
			collisionarray[mp.Hash(cBuffer)]++;
		}
		//Set(Number Key)
		for (key=0;key<6;key++)
		{
			assert(mp.Set(key,pSetData2,pafee_strlen(pSetData2)));
			addCount++;
			presentElementCount++;
			collisionarray[mp.Hash(key)]++;
		}
		//Set(Character Key)
		for (key=0;key<6;key++)
		{
			cBuffer[0] = (char)(key + 49);
			//pSkey = &c;
			cBuffer[1] = '\0';
			assert(mp.Set(cBuffer,pSetData2,pafee_strlen(pSetData2)));
			addCount++;
			presentElementCount++;
			collisionarray[mp.Hash(cBuffer)]++;
		}
		//Set(Number Key)
		for (key=0;key<6;key++)
		{
			assert(mp.Set(key,pSetData3,pafee_strlen(pSetData3)));
			addCount++;
			presentElementCount++;
			collisionarray[mp.Hash(key)]++;
		}
		//Set(Character Key)
		for (key=0;key<6;key++)
		{
			cBuffer[0] = (char)(key + 49);
			//pSkey = &c;
			cBuffer[1] = '\0';
			assert(mp.Set(cBuffer,pSetData3,pafee_strlen(pSetData3)));
			addCount++;
			presentElementCount++;
			collisionarray[mp.Hash(cBuffer)]++;
		}
		maxElementCount = 36;
		//collisionCount += 32;
		//minListCount = 6;
		//maxListCount = 12;
		collisionCount=0;
		minListCount=INT_MAX;
		maxListCount=0;
		for (eh=0;eh<HASHSIZE;eh++)
		{
			if ( collisionarray[eh] >= 1 )
			{
				collisionCount+=collisionarray[eh]-1;
			}
			if ( minListCount > collisionarray[eh] )
			{
				minListCount = collisionarray[eh];
			}
			if ( maxListCount < collisionarray[eh] )
			{
				maxListCount = collisionarray[eh];
			}
		}
#ifdef PAFEE_STATISTICS_ENABLE
		mp.GetStatistics(&statisticsRtn);
		assert(statisticsRtn.addCount == addCount);
		assert(statisticsRtn.acqCount == acqCount);
		assert(statisticsRtn.deleteCount == deleteCount);
		assert(statisticsRtn.maxElementCount == maxElementCount);
		assert(statisticsRtn.collisionCount == collisionCount);
		assert(statisticsRtn.minListCount == minListCount);
		assert(statisticsRtn.maxListCount == maxListCount);
		assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		//GetStartPosition(int key)/GetNext
		key = 5;
		pos = mp.GetStartPosition(key);
		assert(mp.GetNext(pos,work2,&len,sizeof(work2)));
		assert(pafee_memcmp(pSetData1,work2,pafee_strlen(pSetData1)) == 0);
		assert(len == (int)pafee_strlen(pSetData1));
		//GetNext(No copy version)
		pGetData2 = (char* )mp.GetNext(pos,&len);
		assert(pGetData2 != NULL);
		assert(pafee_memcmp(pSetData2,pGetData2,pafee_strlen(pSetData2)) == 0);
		assert(len == (int)pafee_strlen(pSetData2));
		//GetNext(No copy version)
		pGetData3 = (char* )mp.GetNext(pos,&len);
		assert(pGetData3 != NULL);
		assert(pafee_memcmp(pSetData3,pGetData3,pafee_strlen(pSetData3)) == 0);
		assert(len == (int)pafee_strlen(pSetData3));
		//GetStartPosition(character key)/GetNext
		cBuffer[0] = (char)(1 + 49);
		//pSkey = &c;
		cBuffer[1] = '\0';
		pos = mp.GetStartPosition(cBuffer);
		assert(mp.GetNext(pos,work2,&len,sizeof(work2)));
		assert(pafee_memcmp(pSetData1,work2,pafee_strlen(pSetData1)) == 0);
		assert(len == (int)pafee_strlen(pSetData1));
		//GetNext(No copy version)
		pGetData2 = (char* )mp.GetNext(pos,&len);
		assert(pGetData2 != NULL);
		assert(pafee_memcmp(pSetData2,pGetData2,pafee_strlen(pSetData2)) == 0);
		assert(len == (int)pafee_strlen(pSetData2));
		//GetNext(No copy version)
		pGetData3 = (char* )mp.GetNext(pos,&len);
		assert(pGetData3 != NULL);
		assert(pafee_memcmp(pSetData3,pGetData3,pafee_strlen(pSetData3)) == 0);
		assert(len == (int)pafee_strlen(pSetData3));
		//GetStartPosition()/GetNext
		size = mp.GetCount();
		assert(size == 36);
		pos = mp.GetStartPosition();
		/*
		for(jj=0;jj<size;jj++)
		{
			assert(mp.GetNext(pos,work2,&len,sizeof(work2)));
			if((jj == 0) || (jj == 1) || (jj == 2) || (jj == 9) || (jj == 10) || (jj == 11) || (jj == 12) || (jj == 21) || (jj == 22) || (jj == 23) || (jj == 30) || (jj == 31))
			{
				assert(pafee_memcmp(pSetData1,work2,pafee_strlen(pSetData1)) == 0);
			}
			else if((jj == 3) || (jj == 4) || (jj == 5) || (jj == 13) || (jj == 14) || (jj == 15) || (jj == 16) || (jj == 24) || (jj == 25) || (jj == 26) || (jj == 32) || (jj == 33))
			{
				assert(pafee_memcmp(pSetData2,work2,pafee_strlen(pSetData2)) == 0);
			}
			else
			{
				assert(pafee_memcmp(pSetData3,work2,pafee_strlen(pSetData3)) == 0);
			}
		}
		*/
#ifdef PAFEE_STATISTICS_ENABLE
		mp.GetStatistics(&statisticsRtn);
		assert(statisticsRtn.addCount == addCount);
		assert(statisticsRtn.acqCount == acqCount);
		assert(statisticsRtn.deleteCount == deleteCount);
		assert(statisticsRtn.maxElementCount == maxElementCount);
		assert(statisticsRtn.collisionCount == collisionCount);
		assert(statisticsRtn.minListCount == minListCount);
		assert(statisticsRtn.maxListCount == maxListCount);
		assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		//Remove
		key=5;
		mp.Remove(key);
		collisionarray[mp.Hash(key)]--;
		collisionarray[mp.Hash(key)]--;
		collisionarray[mp.Hash(key)]--;
		assert(mp.GetCount() == 33);
		assert(!mp.IsEmpty());
		deleteCount+=3;
		presentElementCount-=3;
		maxListCount-=3;
		collisionCount=0;
		minListCount=INT_MAX;
		maxListCount=0;
		for (eh=0;eh<HASHSIZE;eh++)
		{
			if ( collisionarray[eh] >= 1 )
			{
				collisionCount+=collisionarray[eh]-1;
			}
			if ( minListCount > collisionarray[eh] )
			{
				minListCount = collisionarray[eh];
			}
			if ( maxListCount < collisionarray[eh] )
			{
				maxListCount = collisionarray[eh];
			}
		}
#ifdef PAFEE_STATISTICS_ENABLE
		mp.GetStatistics(&statisticsRtn);
		assert(statisticsRtn.addCount == addCount);
		assert(statisticsRtn.acqCount == acqCount);
		assert(statisticsRtn.deleteCount == deleteCount);
		assert(statisticsRtn.maxElementCount == maxElementCount);
		assert(statisticsRtn.collisionCount == collisionCount);
		assert(statisticsRtn.minListCount == minListCount);
		assert(statisticsRtn.maxListCount == maxListCount);
		assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		//Remove
		cBuffer[0] = (char)(2 + 49);
		//pSkey = &c;
		cBuffer[1] = '\0';
		mp.Remove(cBuffer);
		collisionarray[mp.Hash(cBuffer)]--;
		collisionarray[mp.Hash(cBuffer)]--;
		collisionarray[mp.Hash(cBuffer)]--;
		assert(mp.GetCount() == 30);
		assert(!mp.IsEmpty());
		deleteCount+=3;
		minListCount-=3;
		presentElementCount-=3;
		collisionCount=0;
		minListCount=INT_MAX;
		maxListCount=0;
		for (eh=0;eh<HASHSIZE;eh++)
		{
			if ( collisionarray[eh] >= 1 )
			{
				collisionCount+=collisionarray[eh]-1;
			}
			if ( minListCount > collisionarray[eh] )
			{
				minListCount = collisionarray[eh];
			}
			if ( maxListCount < collisionarray[eh] )
			{
				maxListCount = collisionarray[eh];
			}
		}
#ifdef PAFEE_STATISTICS_ENABLE
		mp.GetStatistics(&statisticsRtn);
		assert(statisticsRtn.addCount == addCount);
		assert(statisticsRtn.acqCount == acqCount);
		assert(statisticsRtn.deleteCount == deleteCount);
		assert(statisticsRtn.maxElementCount == maxElementCount);
		assert(statisticsRtn.collisionCount == collisionCount);
		assert(statisticsRtn.minListCount == minListCount);
		assert(statisticsRtn.maxListCount == maxListCount);
		assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		//RemoveAll
		mp.RemoveAll();
		for (jj=0;jj<HASHSIZE;jj++)
		{
			collisionarray[jj]=0;
		}
		assert(!mp.GetCount());
		assert(mp.IsEmpty());
		deleteCount += presentElementCount;
		minListCount = 0;
		maxListCount = 0;
		presentElementCount = 0;
		collisionCount=0;
#ifdef PAFEE_STATISTICS_ENABLE
		mp.GetStatistics(&statisticsRtn);
		assert(statisticsRtn.addCount == addCount);
		assert(statisticsRtn.acqCount == acqCount);
		assert(statisticsRtn.deleteCount == deleteCount);
		assert(statisticsRtn.maxElementCount == maxElementCount);
		assert(statisticsRtn.collisionCount == collisionCount);
		assert(statisticsRtn.minListCount == minListCount);
		assert(statisticsRtn.maxListCount == maxListCount);
		assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		//Set(Number Key)
		for (key=0;key<6;key++)
		{
			assert(mp.Set(key,pSetData1,pafee_strlen(pSetData1)));
			addCount++;
			presentElementCount++;
			collisionarray[mp.Hash(key)]++;
		}
		//Set(Character Key)
		for (key=0;key<6;key++)
		{
			cBuffer[0] = (char)(key + 49);
			//pSkey = &c;
			cBuffer[1] = '\0';
			assert(mp.Set(cBuffer,pSetData1,pafee_strlen(pSetData1)));
			addCount++;
			presentElementCount++;
			collisionarray[mp.Hash(cBuffer)]++;
		}
		//collisionCount+=8;
		//minListCount = 2;
		//maxListCount = 4;
		collisionCount=0;
		minListCount=INT_MAX;
		maxListCount=0;
		for (eh=0;eh<HASHSIZE;eh++)
		{
			if ( collisionarray[eh] >= 1 )
			{
				collisionCount+=collisionarray[eh]-1;
			}
			if ( minListCount > collisionarray[eh] )
			{
				minListCount = collisionarray[eh];
			}
			if ( maxListCount < collisionarray[eh] )
			{
				maxListCount = collisionarray[eh];
			}
		}
#ifdef PAFEE_STATISTICS_ENABLE
		mp.GetStatistics(&statisticsRtn);
		assert(statisticsRtn.addCount == addCount);
		assert(statisticsRtn.acqCount == acqCount);
		assert(statisticsRtn.deleteCount == deleteCount);
		assert(statisticsRtn.maxElementCount == maxElementCount);
		assert(statisticsRtn.collisionCount == collisionCount);
		assert(statisticsRtn.minListCount == minListCount);
		assert(statisticsRtn.maxListCount == maxListCount);
		assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
	}
}
#endif	// PAFEE_UNITTEST

};	// Pafee
