//****************************************************
//
//    project   Pafee
//!   @file:    PafeePriorityQueueT.h
//!   @author:  M.Mori
//!   @brief:   PafeePriorityQueueT Class
//
//*****************************************************

//
// P A F E E   - Pragmatic Application Framework for Embedded Environment- 
//
//
// Copyright (c) 2009 KN Information Systems Corporation,
//                    Hirocon Corporation,
//                    Technocrats Corporation
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//

#ifndef _INCLUDE_PRIORITYQUEUE_H_
#define _INCLUDE_PRIORITYQUEUE_H_
#include "PafeeObject.h"
#include "PafeeAllocator.h"
#include "PafeeListT.h"

namespace Pafee
{

//! PriorityQueueT
template<class T>
class PriorityQueueT : public Object
{
protected:
	inline int AligmentSize(int size) {return (((size-1)|((size-1)^3))+1);};
public:
	PriorityQueueT(
		int priorityRange=0, ///< [in] A range of the priority
		Allocator* pAllocator=NULL ///< [in] Memory allocator
	)
	{
		Initialize(priorityRange, pAllocator);
	}

	PriorityQueueT(
		const PriorityQueueT& pqtuec, ///< [in] initial value of Queue
		Allocator* pAllocator=NULL ///< [in] Memory Allocator
	)
	{
		Initialize(pqtuec.m_priorityRange, pAllocator);
		int i;
		for(i=0;i<pqtuec.m_priorityRange;i++)
		{
			m_pPriorityTable[i] = pqtuec.m_pPriorityTable[i];
		}
	}

	~PriorityQueueT()
	{
		if ( m_pPriorityTable != NULL )
		{
			Deallocate(m_pPriorityTable);
		}
	}

	PriorityQueueT& operator =(
		const PriorityQueueT& pqtuec ///< [in] The object which is copied
	)
	{
		int i;
		RemoveAll();
		for(i=0; i<m_priorityRange; i++)
		{
			m_pPriorityTable[i].~ListT<T>();
		}
		if ( m_pPriorityTable != NULL )
		{
			this->Deallocate(m_pPriorityTable);
			m_pPriorityTable = NULL;
		}
		Initialize(pqtuec.m_priorityRange, pqtuec.m_pAllocator);
		for(i=0;i<pqtuec.m_priorityRange;i++)
		{
			m_pPriorityTable[i] = pqtuec.m_pPriorityTable[i];
		}
		return *this;
	}

	bool Initialize(
		int priorityRange, ///< [in] A range of the priority
		Allocator* pAllocator=NULL ///< [in] Memory Allocator
	)
	{
		m_Count=0;
		m_pPriorityTable=NULL;
		m_priorityRange = priorityRange;
		if ( m_priorityRange <= 0 )
		{
			return false;
		}
		SetAllocator(pAllocator);
		m_pPriorityTable = (ListT<T>*)this->Allocate(m_priorityRange*sizeof(ListT<T>));
		for(int i=0; i<m_priorityRange; i++)
		{
			m_pPriorityTable[i].Initialize(pAllocator);
		}
		return true;
	}

	int GetCount()
	{
		int rtn=0;
		for(int i=0; i<m_priorityRange; i++ )
		{
			int rtn2 = m_pPriorityTable[i].GetCount();
			rtn +=rtn2;
		}
		return rtn;
	}

	bool IsEmpty(void)
	{
		int rtn=0;
		for(int i=0; i<m_priorityRange; i++ )
		{
			int rtn2 = m_pPriorityTable[i].GetCount();
			rtn +=rtn2;
		}
		return rtn==0;
	}

	bool Enqueue (
		int priority, ///< [in] Priority
		const T& Data ///< [in] Data to set
	)
	{
		if ( m_priorityRange < priority )
		{
			return false;
		}
		m_pPriorityTable[priority].AddBack( Data );
		m_Count++;
		return true;
	}

	bool Dequeue(
		int* pPriority, ///< [out] Priority
		T& Buffer ///< [out] The data that was gotten
	)
	{
		for(int i=0; i<m_priorityRange; i++ )
		{
			if ( m_pPriorityTable[i].GetCount() > 0 )
			{
				Buffer = m_pPriorityTable[i].GetFront();
				m_pPriorityTable[i].RemoveFront();
				*pPriority =i;
				m_Count--;
				return true;
			}
		}
		*pPriority =-1;
		return false;
	}

	bool Peek(
		int* pPriority, ///< [out] Priority
		T& Buffer ///< [out] The data that was gotten
	)
	{
		for(int i=0;i<m_priorityRange;i++)
		{
			if ( m_pPriorityTable[i].GetCount() > 0 )
			{
				*pPriority =i;
				Buffer = m_pPriorityTable[i].GetFront();
				return true;
			}
		}
		*pPriority =-1;
		return false;
	}

	void RemoveFront(void)
	{
		for(int i=0; i<m_priorityRange; i++ )
		{
			if (m_pPriorityTable[i].GetCount()!=0)
			{
				m_pPriorityTable[i].RemoveFront();
				m_Count--;
				return;
			}
		}
		return;
	}

	void RemoveAll(void)
	{
		for(int i=0; i<m_priorityRange; i++ )
		{
			int cMax = m_pPriorityTable[i].GetCount();
			for(int j=0; j<cMax; j++ )
			{
				if ( m_pPriorityTable[i].GetCount() == 0 )
				{
					break;
				}
				m_pPriorityTable[i].RemoveFront();
				m_Count--;
			}
		}
		return;
	}

#ifdef PAFEE_STATISTICS_ENABLE
public:
	typedef struct
	{
		int addCount;
		int maxElementCount;
		int presentElementCount;
	} STATISTICS;
	STATISTICS m_Statistics;
	void GetStatistics(STATISTICS* pStatisticsRtn)
	{
		struct Pafee::ListT<T>::STATISTICS ListStatistics;
		m_Statistics.addCount = 0;
		m_Statistics.maxElementCount = 0;
		m_Statistics.presentElementCount = 0;
		ListT<T> *p = m_pPriorityTable;
		for(int i=0;i<m_priorityRange;i++)
		{
			p->GetStatistics(&ListStatistics);
			m_Statistics.addCount            += ListStatistics.addCount;
			m_Statistics.maxElementCount     += ListStatistics.maxElementCount;
			m_Statistics.presentElementCount += ListStatistics.presentElementCount;
			p++;
		}
		pafee_memcpy ( pStatisticsRtn, &m_Statistics, sizeof m_Statistics );
	}
#endif // PAFEE_STATISTICS_ENABLE

private:
	ListT<T> *m_pPriorityTable;
	int m_Count;
	int m_priorityRange;

};

}

#endif

