//****************************************************
//
//    project   Pafee
//!   @file:    PafeeQueue.cpp
//!   @author:  M.Tanaka
//!   @brief:   Queue Class
//
//*****************************************************

//
// P A F E E   - Pragmatic Application Framework for Embedded Environment- 
//
//
// Copyright (c) 2009 KN Information Systems Corporation,
//                    Hirocon Corporation,
//                    Technocrats Corporation
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//

#include "PafeeQueue.h"

using namespace Pafee;

//----------------------------------------
//! Queue
//! @return: void
//----------------------------------------
Queue::Queue(
	Allocator* pAllocator ///< [in] Memory Allocator
)
{
	SetAllocator(pAllocator);
	m_pList = (List*)Allocate(sizeof( List));
	new (m_pList)List(pAllocator);
}

//----------------------------------------
//! Queue
//! @return: void
//----------------------------------------
Queue::Queue(
	const Queue& quec, ///< [in] initial value of Queue
	Allocator* pAllocator ///< [in] Memory Allocator
)
{
	SetAllocator(pAllocator);
	m_pList = (List*)Allocate(sizeof( List));
	new (m_pList)List(pAllocator);
	*m_pList = *quec.m_pList;
}

//----------------------------------------
//! ~Queue
//! @return: void
//----------------------------------------
Queue::~Queue()
{
	m_pList->~List();
	Deallocate(m_pList); // delete m_pList;
}

//-------------------------------------------------------------------------------------
//! operator =
//! @return Result(the address of Queue)
//-------------------------------------------------------------------------------------
Queue& Queue::operator =(
	const Queue& quec ///< [in] initial value of Queue
)
{
	*m_pList = *quec.m_pList;
	return *this;
}

//----------------------------------------
//! Dequeue
//! @return: Result(true:Success false:failure)
//----------------------------------------
bool Queue::Dequeue(
	void* pBuffer, ///< [out] Head Position Element 
	int* pLength, ///< [out] Head Position Element Size
	int bufSize ///< [in] Prepared buffer sizes 
)
{
	if(m_pList->GetFront(pBuffer,pLength,bufSize))
	{
		m_pList->RemoveFront();
		return true;
	}
	return false;
}


#ifdef PAFEE_STATISTICS_ENABLE
//----------------------------------------
//! GetStatistics
//! @return: void
//----------------------------------------
void Queue::GetStatistics(
	Queue::STATISTICS* pStatistics ///< [out] statistics Pointer
)
{
	Pafee::List::STATISTICS statisticsRtn;
	m_pList->GetStatistics(&statisticsRtn);
	m_Statistics.addCount = statisticsRtn.addCount;
	m_Statistics.maxElementCount = statisticsRtn.maxElementCount;
	m_Statistics.maxElementSize = statisticsRtn.maxElementSize;
	m_Statistics.minElementSize = statisticsRtn.minElementSize;
	m_Statistics.presentElementCount = statisticsRtn.presentElementCount;
	pafee_memcpy ( pStatistics, &m_Statistics, sizeof (Queue::STATISTICS) );
}
#endif //PAFEE_STATISTICS_ENABLE


#ifdef PAFEE_UNITTEST
//----------------------------------------------------------------------------
//!	Unit test
//! @return: void
//----------------------------------------------------------------------------
void Queue::UnitTest(void)
{
	{
		const int TEST_QUEUE_RECYCLE_COUNT = 10;
		Queue queue;
		int pData[TEST_QUEUE_RECYCLE_COUNT*2];
		int pOutput[TEST_QUEUE_RECYCLE_COUNT*2];
		int i;
		int size = sizeof(int);
		int bufferSize = sizeof(int);
		int half = TEST_QUEUE_RECYCLE_COUNT / 2;
		//
		//STATISTICS
		int addCount=0;
		int maxElementCount=0;
		int minElementSize=0;
		int maxElementSize=0;
		int presentElementCount=0;
#ifdef PAFEE_STATISTICS_ENABLE
		Queue::STATISTICS statisticsRtn;
#endif
		//
		// initialize test data
		for(i=0; i<TEST_QUEUE_RECYCLE_COUNT*2; i++)
		{
			pData[i] = i;
		}
		//
		// test Initialize
		//
		//	assert(queue.Initialize(TEST_QUEUE_RECYCLE_COUNT));
		//
		// test all functions
		//
		{
			int count = 0;
			char* buf;
			// test GetCount/IsEmpty
			assert(queue.GetCount() == 0);
			assert(queue.IsEmpty());
#ifdef PAFEE_STATISTICS_ENABLE
			queue.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			// Enqueue test using (no copy version)
			for(i = 0; i < TEST_QUEUE_RECYCLE_COUNT; i++)
			{
				buf = (char*)(queue.GetWriteBuffer(size));
				pafee_memcpy(buf, &pData[i], size);
				queue.Enqueue(size);
				count++;
				assert(queue.GetCount() == count);
				addCount++;
				presentElementCount++;
			}
			maxElementCount = TEST_QUEUE_RECYCLE_COUNT;
			minElementSize = size;
			maxElementSize = size;
			assert(!queue.IsEmpty());
#ifdef PAFEE_STATISTICS_ENABLE
			queue.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			// test RemoveAll
			queue.RemoveAll();
			assert(queue.GetCount() == 0);
			assert(queue.IsEmpty());
			count = 0;
			minElementSize = 0;
			maxElementSize = 0;
			presentElementCount =0;
#ifdef PAFEE_STATISTICS_ENABLE
			queue.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			// Enqueue test does not use (no copy version)
			for(i = 0; i < TEST_QUEUE_RECYCLE_COUNT; i++)
			{
				queue.Enqueue(&pData[i], size);
				count++;
				assert(queue.GetCount() == count);
				addCount++;
				presentElementCount++;
			}
			minElementSize = size;
			maxElementSize = size;
			assert(!queue.IsEmpty());
#ifdef PAFEE_STATISTICS_ENABLE
			queue.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			// Dequeue test
			for(i = 0; i < TEST_QUEUE_RECYCLE_COUNT; i++)
			{
				queue.Dequeue(&pOutput[i], &size, bufferSize);
				count--;
				assert(queue.GetCount() == count);
				assert(pOutput[i] == i);
				presentElementCount--;
			}
			minElementSize = 0;
			maxElementSize = 0;
			assert(queue.IsEmpty());
			assert(!queue.Dequeue(&pOutput[0], &size,bufferSize));
#ifdef PAFEE_STATISTICS_ENABLE
			queue.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			// Enqueue and Dequeue test
			for(i = 0; i < TEST_QUEUE_RECYCLE_COUNT; i++)
			{
				queue.Enqueue(&pData[i], size);
				count++;
				assert(queue.GetCount() == count);
				addCount++;
				presentElementCount++;
			}
			minElementSize = size;
			maxElementSize = size;
			assert(!queue.IsEmpty());
#ifdef PAFEE_STATISTICS_ENABLE
			queue.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			for(i = 0; i < half; i++)
			{
				queue.Dequeue(&pOutput[i], &size,bufferSize);
				count--;
				assert(queue.GetCount() == count);
				assert(pOutput[i] == i);
				presentElementCount--;
			}
			assert(!queue.IsEmpty());
#ifdef PAFEE_STATISTICS_ENABLE
			queue.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			for(i = 0; i < half; i++)
			{
				queue.Enqueue(&pData[TEST_QUEUE_RECYCLE_COUNT+i], size);
				count++;
				assert(queue.GetCount() == count);
				addCount++;
				presentElementCount++;
			}
			assert(!queue.IsEmpty());
#ifdef PAFEE_STATISTICS_ENABLE
			queue.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			for(i = 0; i < TEST_QUEUE_RECYCLE_COUNT; i++)
			{
				queue.Dequeue(&pOutput[i], &size,bufferSize);
				count--;
				assert(queue.GetCount() == count);
				assert(pOutput[i] == i+half);
				presentElementCount--;
			}
			minElementSize = 0;
			maxElementSize = 0;
			assert(queue.IsEmpty());
#ifdef PAFEE_STATISTICS_ENABLE
			queue.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			assert(!queue.Dequeue(&pOutput[0], &size,bufferSize));
			// test Peek/PeekLength/RemoveFront/RemoveBack
			for(i = 0; i < TEST_QUEUE_RECYCLE_COUNT; i++)
			{
				queue.Enqueue(&pData[i], size);
				count++;
				assert(queue.GetCount() == count);
				addCount++;
				presentElementCount++;
			}
			minElementSize = size;
			maxElementSize = size;
			assert(!queue.IsEmpty());
#ifdef PAFEE_STATISTICS_ENABLE
			queue.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			int peek;
			// test Peek using (no copy version)
			peek = *((int*)queue.Peek(&size));
			assert(peek == 0);
			// test Peek does not use (no copy version)
			queue.Peek(&peek, &size,bufferSize);
			assert(peek == 0);
			int peekLength;
			// test PeekLength
			peekLength = queue.PeekLength();
			assert(peekLength == 4);
			// test RemoveFront
			queue.RemoveFront();
			queue.Peek(&peek, &size,bufferSize);
			assert(peek == 1);
			presentElementCount--;
#ifdef PAFEE_STATISTICS_ENABLE
			queue.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			queue.RemoveBack();
			count = queue.GetCount();
			assert(count == 8);
			presentElementCount--;
#ifdef PAFEE_STATISTICS_ENABLE
			queue.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		}
	}
/////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////
	{
		const int TEST_QUEUE_RECYCLE_COUNT = 10;
		Queue queue0c;
		int pData[TEST_QUEUE_RECYCLE_COUNT*2];
		int pOutput[TEST_QUEUE_RECYCLE_COUNT*2];
		int i;
		int size = sizeof(int);
		int bufferSize = sizeof(int);
		int half = TEST_QUEUE_RECYCLE_COUNT / 2;
		//
		//STATISTICS
		int addCount=0;
		int maxElementCount=0;
		int minElementSize=0;
		int maxElementSize=0;
		int presentElementCount=0;
#ifdef PAFEE_STATISTICS_ENABLE
		Queue::STATISTICS statisticsRtn;
#endif
		//
		// initialize test data
		for(i=0; i<TEST_QUEUE_RECYCLE_COUNT*2; i++)
		{
			pData[i] = i;
		}
		//
		// test Initialize
		//
		//	assert(queue.Initialize(TEST_QUEUE_RECYCLE_COUNT));
		//
		// test all functions
		//
		{
			int count = 0;
			char* buf;
			// test GetCount/IsEmpty
			assert(queue0c.GetCount() == 0);
			assert(queue0c.IsEmpty());
#ifdef PAFEE_STATISTICS_ENABLE
			queue0c.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			// Enqueue test using (no copy version)
			for(i = 0; i < TEST_QUEUE_RECYCLE_COUNT; i++)
			{
				buf = (char*)(queue0c.GetWriteBuffer(size));
				pafee_memcpy(buf, &pData[i], size);
				queue0c.Enqueue(size);
				count++;
				assert(queue0c.GetCount() == count);
				addCount++;
				presentElementCount++;
			}
			maxElementCount = TEST_QUEUE_RECYCLE_COUNT;
			minElementSize = size;
			maxElementSize = size;
			assert(!queue0c.IsEmpty());
#ifdef PAFEE_STATISTICS_ENABLE
			queue0c.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			// test RemoveAll
			queue0c.RemoveAll();
			assert(queue0c.GetCount() == 0);
			assert(queue0c.IsEmpty());
			count = 0;
			minElementSize = 0;
			maxElementSize = 0;
			presentElementCount =0;
#ifdef PAFEE_STATISTICS_ENABLE
			queue0c.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			// Enqueue test does not use (no copy version)
			for(i = 0; i < TEST_QUEUE_RECYCLE_COUNT; i++)
			{
				queue0c.Enqueue(&pData[i], size);
				count++;
				assert(queue0c.GetCount() == count);
				addCount++;
				presentElementCount++;
			}
			minElementSize = size;
			maxElementSize = size;
			assert(!queue0c.IsEmpty());
#ifdef PAFEE_STATISTICS_ENABLE
			queue0c.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			///
			//Queue queue=queue0c;
			Queue queue;
			queue=queue0c;
			///
			// Dequeue test
			for(i = 0; i < TEST_QUEUE_RECYCLE_COUNT; i++)
			{
				queue.Dequeue(&pOutput[i], &size, bufferSize);
				count--;
				assert(queue.GetCount() == count);
				assert(pOutput[i] == i);
				presentElementCount--;
			}
			minElementSize = 0;
			maxElementSize = 0;
			assert(queue.IsEmpty());
			assert(!queue.Dequeue(&pOutput[0], &size,bufferSize));
#ifdef PAFEE_STATISTICS_ENABLE
			queue.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			// Enqueue and Dequeue test
			for(i = 0; i < TEST_QUEUE_RECYCLE_COUNT; i++)
			{
				queue.Enqueue(&pData[i], size);
				count++;
				assert(queue.GetCount() == count);
				addCount++;
				presentElementCount++;
			}
			minElementSize = size;
			maxElementSize = size;
			assert(!queue.IsEmpty());
#ifdef PAFEE_STATISTICS_ENABLE
			queue.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			for(i = 0; i < half; i++)
			{
				queue.Dequeue(&pOutput[i], &size,bufferSize);
				count--;
				assert(queue.GetCount() == count);
				assert(pOutput[i] == i);
				presentElementCount--;
			}
			assert(!queue.IsEmpty());
#ifdef PAFEE_STATISTICS_ENABLE
			queue.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			for(i = 0; i < half; i++)
			{
				queue.Enqueue(&pData[TEST_QUEUE_RECYCLE_COUNT+i], size);
				count++;
				assert(queue.GetCount() == count);
				addCount++;
				presentElementCount++;
			}
			assert(!queue.IsEmpty());
#ifdef PAFEE_STATISTICS_ENABLE
			queue.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			for(i = 0; i < TEST_QUEUE_RECYCLE_COUNT; i++)
			{
				queue.Dequeue(&pOutput[i], &size,bufferSize);
				count--;
				assert(queue.GetCount() == count);
				assert(pOutput[i] == i+half);
				presentElementCount--;
			}
			minElementSize = 0;
			maxElementSize = 0;
			assert(queue.IsEmpty());
#ifdef PAFEE_STATISTICS_ENABLE
			queue.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			assert(!queue.Dequeue(&pOutput[0], &size,bufferSize));
			// test Peek/PeekLength/RemoveFront/RemoveBack
			for(i = 0; i < TEST_QUEUE_RECYCLE_COUNT; i++)
			{
				queue.Enqueue(&pData[i], size);
				count++;
				assert(queue.GetCount() == count);
				addCount++;
				presentElementCount++;
			}
			minElementSize = size;
			maxElementSize = size;
			assert(!queue.IsEmpty());
#ifdef PAFEE_STATISTICS_ENABLE
			queue.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			int peek;
			// test Peek using (no copy version)
			peek = *((int*)queue.Peek(&size));
			assert(peek == 0);
			// test Peek does not use (no copy version)
			queue.Peek(&peek, &size,bufferSize);
			assert(peek == 0);
			int peekLength;
			// test PeekLength
			peekLength = queue.PeekLength();
			assert(peekLength == 4);
			// test RemoveFront
			queue.RemoveFront();
			queue.Peek(&peek, &size,bufferSize);
			assert(peek == 1);
			presentElementCount--;
#ifdef PAFEE_STATISTICS_ENABLE
			queue.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			queue.RemoveBack();
			count = queue.GetCount();
			assert(count == 8);
			presentElementCount--;
#ifdef PAFEE_STATISTICS_ENABLE
			queue.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		}
	}
}
#endif
