//****************************************************
//
//      project     PAFEE
//!      @file:     PafeeRingBuffer.h
//!      @author:   M.Mori
//!      @brief:    RingBuffer class
//
//*****************************************************

//
// P A F E E   - Pragmatic Application Framework for Embedded Environment- 
//
//
// Copyright (c) 2009 KN Information Systems Corporation,
//                    Hirocon Corporation,
//                    Technocrats Corporation
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//

#include "PafeeObject.h"
#include "PafeeCommon.h"
#include "pafee_string.h"
#include <limits.h>
#ifndef _PAFEE_RING_BUFFER_H_INCLUDED_
#define _PAFEE_RING_BUFFER_H_INCLUDED_

#ifdef PAFEE_UNITTEST
#include <assert.h>
#endif //PAFEE_UNITTEST

namespace Pafee
{

//! RingBuffer
class RingBuffer : public Object
{
public:
	inline int AligmentSize(int size) {return ((((size)-1)|(((size)-1)^3))+1);};
#ifdef PAFEE_STATISTICS_ENABLE
public:
	typedef struct 
	{
		int additionalFrequency;
		int maximumElements;
		int overFlowFrequency;
		int presentNumber;
	} STATISTICS;
	void GetStatistics(RingBuffer::STATISTICS* pStatistics);
protected:
	STATISTICS m_Statistics;
#endif
protected:
	typedef struct t_RingBufferPointer
	{
		char* p;
		t_RingBufferPointer* pNext;
		int Size;
	} RingBufferPointer;
	RingBufferPointer* m_pRingBufferPointer;
	RingBufferPointer* m_pWrite;
	RingBufferPointer* m_pRead;
public:
	RingBuffer(int capacity=0, int elementSize=0, Allocator* pAllocator=NULL);
	RingBuffer(RingBuffer& rbc, Allocator* pAllocator=NULL);
	~ RingBuffer();
	bool Initialize(int capacity, int elementSize, Allocator* pAllocator=NULL);
	RingBuffer& operator = (RingBuffer& rbc);
	int GetCount()
	{
		register int usedCount=0;
		register RingBufferPointer* p=m_pRingBufferPointer;
		for(register int i=0;i<m_capacity;i++)
		{
			usedCount+=(p->Size!=0);
			p++;
		}
		return usedCount;
	}
	bool IsEmpty(void)
	{
		register int usedCount=0;
		register RingBufferPointer* p=m_pRingBufferPointer;
		for(register int i=0;i<m_capacity;i++)
		{
			usedCount+=(p->Size!=0);
			p++;
		}
		return usedCount==0;
	}
	void* GetWriteBuffer();
	bool Enqueue (
		void* pData,  ///< [in] Data
		int length  ///< [in] Data Length
	)
	{
		if ( m_elementSize < length )
		{
			return false;
		}
		if ( m_pWrite->Size != 0 )
		{
#ifdef PAFEE_STATISTICS_ENABLE
			m_Statistics.overFlowFrequency++;
#endif
			return false;
		}
		pafee_memcpy ( m_pWrite->p, pData, length );
		m_pWrite->Size = length;
		m_pWrite=m_pWrite->pNext;
#ifdef PAFEE_STATISTICS_ENABLE
		m_Statistics.additionalFrequency++;
		m_Statistics.maximumElements++;
		if(m_capacity < m_Statistics.maximumElements)
		{
			m_Statistics.maximumElements = m_capacity;
		}
		m_Statistics.presentNumber++;
#endif
		return true;
	}
	////
	bool Enqueue (
		int length  ///< [in] Data Length
	)
	{
		if ( m_pWrite->Size != 0 )
		{
#ifdef PAFEE_STATISTICS_ENABLE
			m_Statistics.overFlowFrequency++;
#endif
			return false;
		}
		m_pWrite->Size = length;
		m_pWrite=m_pWrite->pNext;
#ifdef PAFEE_STATISTICS_ENABLE
		m_Statistics.additionalFrequency++;
		m_Statistics.maximumElements++;
		if(m_capacity < m_Statistics.maximumElements)
		{
			m_Statistics.maximumElements = m_capacity;
		}
		m_Statistics.presentNumber++;
#endif
		return true;
	}
	////
	bool Dequeue(
		void* pBuffer, ///< [out] Prepared Buffer
		int* pLength, ///< [out] Data Length
		int bufSize=INT_MAX ///< [in] Prepared Buffer Length
	)
	{
		if ( m_pRead->Size == 0 )
		{
			return false;
		}
		*pLength = m_pRead->Size;
		if(bufSize < *pLength)
		{
			return false;
		}
		pafee_memcpy ( pBuffer, m_pRead->p, *pLength );
		m_pRead->Size = 0;
		m_pRead=m_pRead->pNext;
#ifdef PAFEE_STATISTICS_ENABLE
		m_Statistics.presentNumber--;
#endif
		return true;
	}
	////
	void* Dequeue(
		int* pLength ///< [out] Data Length
	)
	{
		if ( m_pRead->Size == 0 )
		{
			return NULL;
		}
		*pLength = m_pRead->Size;
		char* pRtn = m_pRead->p;
		m_pRead->Size = 0;
		m_pRead=m_pRead->pNext;
#ifdef PAFEE_STATISTICS_ENABLE
		m_Statistics.presentNumber--;
#endif
		return (void *)pRtn;
	}
	////
	bool Peek(void* pBuffer, int* pLength, int bufSize=INT_MAX);
	void* Peek(int* pLength=NULL); //No copy version
	int PeekLength(void);
	void ClearAll(void);
protected:
	char* m_pBuffer;
	int m_capacity;
	int m_elementSize;
	int m_bufferLength;
	int m_bufferAllocateLength;
#ifdef PAFEE_UNITTEST
public: // unit test function
	static void UnitTest(void);
#endif // PAFEE_UNITTEST

};

}
#endif //_PAFEE_RING_BUFFER_H_INCLUDED_

