//********************************************************************
//		project 	PAFEE
//
//! 	 @file: 	PafeeFastFixedAllocator.cpp
//! 	 @author:	Takatsura Miyawaki
//! 	 @brief:	fast fixed memory allocator
//
//********************************************************************

//
// P A F E E   - Pragmatic Application Framework for Embedded Environment- 
//
//
// Copyright (c) 2009 KN Information Sustems Corporation,
//                    Hirocon Corporation,
//                    Technocrats Corporation
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//
// #### NOTE ####
//
// This FastFixedAllocator is using the allocation algorithm of Loki::SmallObjAllocator by Alexandrescu, Andrei.
//

#include <stdlib.h>
#include "PafeeCommon.h"
#include "PafeeFastFixedAllocator.h"


#ifdef PAFEE_UNITTEST
#include "PafeeStdAllocator.h"
#endif

#ifdef PAFEE_COVERAGE_RECORD
#include "PafeeCoverageRecorder.h"
static Pafee::CoverageRecorder g_coverage;
#define COVERAGE_RECORD(LINE)	g_coverage.Record(LINE)
#else
#define COVERAGE_RECORD(LINE)
#endif

using namespace Pafee;

const unsigned long  FastFixedAllocator::SANITY_CODE	= 0xABCC0000;	///< block sanity check code
const unsigned long  FastFixedAllocator::SANITY_MASK	= 0xFFFC0000;	///< mask for sanity code
const unsigned long  FastFixedAllocator::USEFLAG		= 0x00010000;	///< using flag
const unsigned long  FastFixedAllocator::OWNINDEX_MASK	= 0x0000FFFF;	///< mask for own index

//-------------------------------------------------
//! Constructor
//-------------------------------------------------
FastFixedAllocator::FastFixedAllocator()
{
	m_blockSize = 0;
	m_blockCount = 0;
	m_pageSize = 0;
	m_firstAvailableIndex = 0;
	m_pBlocks = NULL;
	m_pOriginalBuffer = NULL;
	m_pBackendAllocator = NULL;
	COVERAGE_RECORD(__LINE__);
}

//-------------------------------------------------
//! Destructor
//-------------------------------------------------
FastFixedAllocator::~FastFixedAllocator()
{
	if(m_pBackendAllocator)
	{
		m_pBackendAllocator->Deallocate(m_pOriginalBuffer);
		COVERAGE_RECORD(__LINE__);
	}
	else
	{
		free(m_pOriginalBuffer);
		COVERAGE_RECORD(__LINE__);
	}
	m_pOriginalBuffer = NULL;
	m_pBlocks = NULL;
	COVERAGE_RECORD(__LINE__);
}

//-------------------------------------------------
//! Initialize
//! @return: true:success  false:failure
//-------------------------------------------------
bool FastFixedAllocator::Initialize(
									unsigned long blockSize,		///< [in] block size
									unsigned long blockCount,		///< [in] block count (max 65535)
									Allocator* pBackendAllocator	///< [in] back-end allocator
									)
{
	if(blockCount >= 0x10000)
	{
		COVERAGE_RECORD(__LINE__);
		return false;
	}
	blockSize += sizeof(unsigned long);
	m_blockSize  = AlignUp(blockSize);
	m_blockCount = blockCount;
	m_pageSize = blockSize * blockCount;
	m_pBackendAllocator = pBackendAllocator;
	
	if(pBackendAllocator)
	{
		m_pOriginalBuffer = (unsigned char*) pBackendAllocator->Allocate(m_pageSize);
		COVERAGE_RECORD(__LINE__);
	}
	else
	{
		m_pOriginalBuffer = (unsigned char*) malloc(m_pageSize);
		COVERAGE_RECORD(__LINE__);
	}
	if(m_pOriginalBuffer == NULL)
	{
		THROW(std::bad_alloc());
		COVERAGE_RECORD(__LINE__);
		return false;
	}
	m_pBlocks = (unsigned char*) AlignUp((unsigned long) m_pOriginalBuffer);
	
	unsigned char* pBlock = m_pBlocks ;
	for(unsigned long i = 0; i < m_blockCount; i++)
	{
		BLK_HEADER* pHeader = (BLK_HEADER*) pBlock;
		pHeader->sanity_use_index = (SANITY_CODE | i);
		pHeader->nextAvailableIndex = i + 1;
		pBlock += m_blockSize;
		COVERAGE_RECORD(__LINE__);
	}
	m_firstAvailableIndex = 0;
	
#ifdef PAFEE_STATISTICS_ENABLE
		pafee_memset(&m_statistics, 0, sizeof(m_statistics));
		m_statistics.minRequestSize = 0xffffffff;
#endif
	COVERAGE_RECORD(__LINE__);
	return true;
}

//-------------------------------------------------
//! Allocate memory
//!	@return allocated memory pointer
//-------------------------------------------------
void* FastFixedAllocator::Allocate(
									unsigned long size	///< [in] size to allocate
									)
{
	STATISTICS_INC(m_statistics.allocateRequests);
	STATISTICS_MIN(m_statistics.minRequestSize, size);
	STATISTICS_MAX(m_statistics.maxRequestSize, size);
	
	if(size > m_blockSize)
	{
		return NULL;
	}
	
	if(m_firstAvailableIndex < m_blockCount)
	{
		BLK_HEADER* pBlock;
		pBlock = (BLK_HEADER*) (m_pBlocks + (m_blockSize * m_firstAvailableIndex));
		
		if((pBlock->sanity_use_index & SANITY_MASK) == SANITY_CODE)
		{
			pBlock->sanity_use_index |= USEFLAG;
			m_firstAvailableIndex = pBlock->nextAvailableIndex;
			STATISTICS_INC(m_statistics.currentAllocations);
			STATISTICS_MAX(m_statistics.peakAllocations, m_statistics.currentAllocations);
			COVERAGE_RECORD(__LINE__);
			return &(pBlock->nextAvailableIndex);
		}
		STATISTICS_INC(m_statistics.sanityErrors);
		COVERAGE_RECORD(__LINE__);
	}
	STATISTICS_INC(m_statistics.failedAllocateRequests);
	COVERAGE_RECORD(__LINE__);
	return NULL;
}

//-------------------------------------------------
//! Deallocate memory
//-------------------------------------------------
void FastFixedAllocator::Deallocate(
									void* ptr	///< [in] poiner to deallocate
									)
{
	if(ptr == NULL)
	{
		COVERAGE_RECORD(__LINE__);
		return;
	}
	BLK_HEADER* pBlock = (BLK_HEADER*) (((unsigned char*) ptr) - sizeof(unsigned long));
	if((pBlock->sanity_use_index & SANITY_MASK) == SANITY_CODE)
	{
		if(pBlock->sanity_use_index & USEFLAG)
		{
			pBlock->sanity_use_index &= ~USEFLAG;
			pBlock->nextAvailableIndex = m_firstAvailableIndex;
			m_firstAvailableIndex = (pBlock->sanity_use_index & OWNINDEX_MASK);
			COVERAGE_RECORD(__LINE__);
			STATISTICS_INC(m_statistics.deallocateRequests);
			STATISTICS_DEC(m_statistics.currentAllocations);
		}
		else
		{
			COVERAGE_RECORD(__LINE__);
		}
	}
	else
	{
		STATISTICS_INC(m_statistics.sanityErrors);
		COVERAGE_RECORD(__LINE__);
	}
}

//-------------------------------------------------
//! Check given address is in own address range
//-------------------------------------------------
inline bool FastFixedAllocator::IsOwnAddress(
										void* ptr	///< [in] pointer to check
										)
{
	if((m_pBlocks <= ptr) && (ptr < (m_pBlocks+m_pageSize)))
	{
		COVERAGE_RECORD(__LINE__);
		return true;
	}
	COVERAGE_RECORD(__LINE__);
	return false;
}



#ifdef PAFEE_STATISTICS_ENABLE
//-------------------------------------------------
//! Get statistics information
//-------------------------------------------------
void FastFixedAllocator::GetStatistics(FastFixedAllocator::STATISTICS* pStat)
{
	pafee_memcpy(pStat, &m_statistics, sizeof(m_statistics));
	COVERAGE_RECORD(__LINE__);
}

//-------------------------------------------------
//! Get statistics
//-------------------------------------------------
void FastFixedAllocator::GetStatistics(
										StreamOutput& stream				///< [in/out] stream output
										)
{
	stream.Write("==== FixedAllocator statistics ====\n");
	stream.Write("buffer address: %p\n", m_pOriginalBuffer);
	stream.Write("block size: %ld\n", m_blockSize);
	stream.Write("block count: %ld\n", m_blockCount);
	stream.Write("page size: %ld\n", m_pageSize);
	stream.Write("-- statistics --\n");
	stream.Write("allocation count: %ld\n", m_statistics.allocateRequests);
	stream.Write("deallocation count: %ld\n", m_statistics.deallocateRequests);
	stream.Write("allocation failure count: %ld\n", m_statistics.failedAllocateRequests);
	stream.Write("sanity error count: %ld\n", m_statistics.sanityErrors);			
	stream.Write("current allocated block count: %ld\n", m_statistics.currentAllocations);
	stream.Write("peak allocated block count: %ld\n", m_statistics.peakAllocations);
	stream.Write("minimum request size: %ld\n", m_statistics.minRequestSize);
	stream.Write("maximum request size: %ld\n", m_statistics.maxRequestSize);
	
}

#endif
#ifdef PAFEE_UNITTEST
//-------------------------------------------------
//! Unit test
//-------------------------------------------------
void FastFixedAllocator::UnitTest(void)
{
	int i;
	char* pBufs[100];
	STATISTICS stat;
	{
		FastFixedAllocator alloc;
		assert(alloc.Initialize(8, 4));
		assert(alloc.IsOwnAddress(alloc.m_pBlocks));
		assert(alloc.IsOwnAddress(alloc.m_pBlocks+(8+4)*4) == false);

		printf("--- allocate to max ---\n");
		{
			unsigned int size = 4;
			for(i = 0; i < 4; i++)
			{
				pBufs[i] = (char*) alloc.Allocate(size);
				assert(pBufs[i]);
				memset(pBufs[i], 0xEE, size);
				alloc.GetStatistics(&stat);
				assert(stat.allocateRequests == i+1);
				assert(stat.maxRequestSize == size);
				assert(stat.currentAllocations == i+1);
				size++;
			}
			alloc.GetStatistics(&stat);
			assert(stat.allocateRequests == 4);
			assert(stat.deallocateRequests == 0);
			assert(stat.minRequestSize == 4);
			assert(stat.currentAllocations == 4);
			assert(stat.failedAllocateRequests == 0);
			assert(stat.sanityErrors == 0);

			// full
			pBufs[i] = (char*) alloc.Allocate(8);
			assert(pBufs[i] == NULL);
			alloc.GetStatistics(&stat);
			assert(stat.allocateRequests == 5);
			assert(stat.currentAllocations == 4);
			assert(stat.failedAllocateRequests == 1);
		}

		printf("--- deallocate all ---\n");
		{
			for(i = 0; i < 4; i++)
			{
				alloc.Deallocate(pBufs[i] );
			}
			alloc.GetStatistics(&stat);
			assert(stat.allocateRequests == 5);
			assert(stat.deallocateRequests == 4);
			assert(stat.currentAllocations == 0);
			assert(stat.sanityErrors == 0);

			// deallocate twice
			for(i = 0; i < 4; i++)
			{
				alloc.Deallocate(pBufs[i] );
			}
			alloc.GetStatistics(&stat);
			assert(stat.allocateRequests == 5);
			assert(stat.deallocateRequests == 4);
			assert(stat.currentAllocations == 0);
			assert(stat.sanityErrors == 0);
			alloc.Deallocate(NULL);

		}

		printf("--- allocate and deallocate ---\n");
		{
			i = 0;
			pBufs[i] = (char*) alloc.Allocate(8);
			assert(pBufs[i]);
			memset(pBufs[i], 0xCC, 8);

			i = 1;
			pBufs[i] = (char*) alloc.Allocate(8);
			assert(pBufs[i]);
			memset(pBufs[i], 0xCC, 8);

			alloc.Deallocate(pBufs[0] );

			i = 2;
			pBufs[i] = (char*) alloc.Allocate(8);
			assert(pBufs[i]);
			memset(pBufs[i], 0xCC, 8);

			i = 3;
			pBufs[i] = (char*) alloc.Allocate(8);
			assert(pBufs[i]);
			memset(pBufs[i], 0xCC, 8);


			i = 4;
			pBufs[i] = (char*) alloc.Allocate(8);
			assert(pBufs[i]);
			memset(pBufs[i], 0xCC, 8);
			
			alloc.Deallocate(pBufs[2] );
			alloc.Deallocate(pBufs[1] );

			i = 5;
			pBufs[i] = (char*) alloc.Allocate(8);
			assert(pBufs[i]);
			memset(pBufs[i], 0xCC, 8);

			alloc.Deallocate(pBufs[4] );
			alloc.Deallocate(pBufs[5] );

			i = 6;
			pBufs[i] = (char*) alloc.Allocate(8);
			assert(pBufs[i]);
			memset(pBufs[i], 0xCC, 8);

			i = 7;
			pBufs[i] = (char*) alloc.Allocate(8);
			assert(pBufs[i]);
			memset(pBufs[i], 0xCC, 8);

			i = 8;
			pBufs[i] = (char*) alloc.Allocate(8);
			assert(pBufs[i]);
			memset(pBufs[i], 0xCC, 8);

			i = 9;
			pBufs[i] = (char*) alloc.Allocate(8);
			assert(pBufs[i] == NULL);

			alloc.Deallocate(pBufs[6] );
			alloc.Deallocate(pBufs[7] );
			alloc.Deallocate(pBufs[3] );
			alloc.Deallocate(pBufs[8] );

			alloc.GetStatistics(&stat);
			assert(stat.currentAllocations == 0);
			assert(stat.failedAllocateRequests == 2);
		}

		printf("--- sanity error ---\n");
		{
			pBufs[0] = (char*) alloc.Allocate(8);
			*((unsigned long*)(pBufs[0]-4)) = 0xaaaaaaaa;
			alloc.Deallocate(pBufs[0]);
			alloc.GetStatistics(&stat);
			assert(stat.sanityErrors == 1);

			*((unsigned long*)(pBufs[1]-4)) = 0xaaaaaaaa;
			*((unsigned long*)(pBufs[2]-4)) = 0xaaaaaaaa;
			*((unsigned long*)(pBufs[3]-4)) = 0xaaaaaaaa;

			pBufs[0] = (char*) alloc.Allocate(8);
			alloc.GetStatistics(&stat);
			assert(stat.sanityErrors == 2);

		}
	}

	StdAllocator stdAlloc;
	{
		FastFixedAllocator alloc;
		assert(alloc.Initialize(8, 0x10000) == false);
		assert(alloc.Initialize(8, 4, &stdAlloc) == true);
	}
#ifdef PAFEE_COVERAGE_RECORD
	g_coverage.Show();
#endif
}
#endif

