//********************************************************************
//
//! 	 @file: 	PafeeFlexibleAllocator.cpp
//! 	 @author:	T.Miyawaki
//! 	 @brief:	Flexible allocator
//
//********************************************************************

//
// P A F E E   - Pragmatic Application Framework for Embedded Environment- 
//
//
// Copyright (c) 2009 KN Information Sustems Corporation,
//                    Hirocon Corporation,
//                    Technocrats Corporation
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//

#include <stddef.h>
#include <stdlib.h>
#include "pafee_string.h"
#include "PafeeCommon.h"
#include "PafeeFlexibleAllocator.h"



#ifdef PAFEE_COVERAGE_RECORD
#include "PafeeCoverageRecorder.h"
static Pafee::CoverageRecorder g_coverage;
#define COVERAGE_RECORD(LINE)	g_coverage.Record(LINE)
#else
#define COVERAGE_RECORD(LINE)
#endif

#ifdef PAFEE_DEBUG
#define CHECK_LOOP(TARGET)		if((TARGET)->pNext) { assert(((TARGET)->pNext != (TARGET)->pPrev));}
#else
#define CHECK_LOOP(TARGET)
#endif

//#define PAFEE_MEM_DEBUG
#ifdef PAFEE_MEM_DEBUG
#define MEM_TRACE(STR)	printf STR
#else
#define MEM_TRACE(STR)	
#endif

using namespace Pafee;

const unsigned long FlexibleAllocator::SANITY_CODE = 0xACACACAC;	///< Sanity code
const unsigned long FlexibleAllocator::USING_MARK = 0x00000001;		///< using mark

//--------------------------------------------
//! Constractor
//--------------------------------------------
FlexibleAllocator::FlexibleAllocator()
{
	m_heapAddr			= 0;
	m_heapSize		= 0;
	m_maxAllocateSize	= 0;
	m_minBlockSize		= 0;
	m_endAddr			= 0;

	m_pFreeListHead = NULL;
	m_pUsingListHead = NULL;

	m_pAllocator = NULL;
	m_isOwnHeapAllocation = false;

#ifdef PAFEE_STATISTICS_ENABLE
	pafee_memset(&m_stat, 0, sizeof(m_stat));
	m_stat.minFreeBytes = 0xffffffff;
	m_stat.minAllocateSize = 0xffffffff;
#endif
	COVERAGE_RECORD(__LINE__);
}

//--------------------------------------------
//! Destoractor
//--------------------------------------------
FlexibleAllocator::~FlexibleAllocator()
{
	if(m_isOwnHeapAllocation)
	{
		if(m_pAllocator)
		{
			m_pAllocator->Deallocate((char*) m_heapAddr);
		}
		else
		{
			free((char*) m_heapAddr);
		}
	}
	COVERAGE_RECORD(__LINE__);
}


//--------------------------------------------
//! Initialize (Allocate heap by given allocator)
//--------------------------------------------
bool FlexibleAllocator::Initialize(
			unsigned long heapSize,			///< [in] heap size
			unsigned long maxAllocateSize,	///< [in] maximum allocate size
			unsigned long minBlockSize,		///< [in] minimum block size
			Allocator* pAllocator/*=NULL*/	///< [in] Allocator
			)
{
	if(m_heapAddr)
	{
		COVERAGE_RECORD(__LINE__);
		return false;	// can't initialize twice.
	}
	unsigned long heapAddr;
	m_pAllocator = pAllocator;
	if(pAllocator)
	{
		COVERAGE_RECORD(__LINE__);
		heapAddr = (unsigned long) pAllocator->Allocate(heapSize);
	}
	else
	{
		COVERAGE_RECORD(__LINE__);
		heapAddr = (unsigned long) malloc(heapSize);
	}
	if(heapAddr == 0)
	{
#ifdef PAFEE_EXCEPTION
		throw std::bad_alloc();
#endif
		COVERAGE_RECORD(__LINE__);
		return false;
	}
	m_isOwnHeapAllocation= true;
	Initialize(heapAddr, heapSize, maxAllocateSize, minBlockSize);
	COVERAGE_RECORD(__LINE__);
	return true;
}

//--------------------------------------------
//! Initialize  (Use given heap)
//--------------------------------------------
void FlexibleAllocator::Initialize(
			unsigned long heapAddr,			///< [in] start address of heap
			unsigned long heapSize,			///< [in] heap size
			unsigned long maxAllocateSize,	///< [in] maximum allocate size
			unsigned long minBlockSize		///< [in] minimum block size
			)
{
	if(m_heapAddr)
	{
		COVERAGE_RECORD(__LINE__);
		return;	// can't initialize twice.
	}
	m_heapAddr		= AlignUp(heapAddr);
	m_heapSize		= heapSize;
	if(heapAddr & 0x3)
	{
		m_heapSize	-= (4 - (heapAddr & 0x3));
		m_heapSize = AlignDown(m_heapSize);
		COVERAGE_RECORD(__LINE__);
	}
	m_endAddr			= m_heapAddr + m_heapSize;
	m_maxAllocateSize	= AlignDown(maxAllocateSize);
	m_minBlockSize		= AlignDown(minBlockSize);
	m_minBlockSize += sizeof(BLK_HEADER);

	BLK_HEADER* pHeader = (BLK_HEADER*) m_heapAddr;
	pHeader->sanity_u  = SANITY_CODE;
	pHeader->blockSize = m_heapSize - sizeof(BLK_HEADER);
	pHeader->pFrontBlock = NULL;
	pHeader->pBackBlock = NULL;
	pHeader->pPrev = NULL;
	pHeader->pNext = NULL;
	m_pFreeListHead = (BLK_HEADER*)m_heapAddr;

	STATISTICS_SET(m_stat.current.freeBytes, m_heapSize);
	COVERAGE_RECORD(__LINE__);
}

//--------------------------------------------
//! Allocate memory
//!	@return allocated memory pointer
//--------------------------------------------
void* FlexibleAllocator::Allocate(
								unsigned long size	///< [in] size to allocate
								)
{
#ifdef PAFEE_STATISTICS_ENABLE
	unsigned int count = 1;
#endif
	BLK_HEADER* pTarget = m_pFreeListHead;
	unsigned long useSize = AlignUp(size + sizeof(BLK_HEADER));

	MEM_TRACE(("Allocate:%ld ", size));

	if(useSize < m_minBlockSize)
	{
		useSize = m_minBlockSize;
		COVERAGE_RECORD(__LINE__);
	}
	if(size > m_maxAllocateSize)
	{
		COVERAGE_RECORD(__LINE__);
#ifdef PAFEE_EXCEPTION
		throw std::bad_alloc();
#endif
		return NULL;
	}
	
	STATISTICS_INC(m_stat.allocateRequest);
	STATISTICS_MAX(m_stat.maxAllocateSize, size);
	STATISTICS_MIN(m_stat.minAllocateSize, size);

	while(pTarget)
	{
		CHECK_LOOP(pTarget);
		if(pTarget->blockSize >= useSize)
		{
			// Divide block
			if((pTarget->blockSize - useSize) >= (m_minBlockSize))
			{
				//
				// [target.prev]->[target][empty]->[target.next->empty.next]
				//
				BLK_HEADER* pEmpty = NULL;
				pEmpty = (BLK_HEADER*) ((unsigned long)pTarget+useSize);
				pEmpty->sanity_u = SANITY_CODE;
				pEmpty->blockSize = pTarget->blockSize - useSize;
				pEmpty->pFrontBlock = pTarget;
				pEmpty->pBackBlock = pTarget->pBackBlock;
				pEmpty->pPrev = pTarget->pPrev;
				pEmpty->pNext = pTarget->pNext;
				if((unsigned long) pEmpty >= m_endAddr)
				{
					PAFEE_ASSERT(false);
				}
							
				if(pTarget->pBackBlock)
				{
					pTarget->pBackBlock->pFrontBlock = pEmpty;
					COVERAGE_RECORD(__LINE__);
				}
				if(pTarget->pNext)
				{
					pTarget->pNext->pPrev = pEmpty;
					COVERAGE_RECORD(__LINE__);
				}
				if(pTarget->pPrev)
				{
					pTarget->pPrev->pNext = pEmpty;
					COVERAGE_RECORD(__LINE__);
				}
				else
				{
					PAFEE_ASSERT(m_pFreeListHead == pTarget);	
					m_pFreeListHead = pEmpty;
				}
				pTarget->blockSize = useSize;
				pTarget->pBackBlock = pEmpty;
				PAFEE_ASSERT(m_pFreeListHead->pPrev == NULL);
				COVERAGE_RECORD(__LINE__);
			}
			// Take whole block
			else
			{
				// Remove from free-list
				if(pTarget->pPrev)
				{
					pTarget->pPrev->pNext = pTarget->pNext;
					CHECK_LOOP(pTarget->pPrev);
					COVERAGE_RECORD(__LINE__);
				}
				else
				{
					PAFEE_ASSERT(m_pFreeListHead == pTarget);
					m_pFreeListHead = pTarget->pNext;
					COVERAGE_RECORD(__LINE__);
				}
				if(pTarget->pNext)
				{
					pTarget->pNext->pPrev = pTarget->pPrev;
					CHECK_LOOP(pTarget->pNext);
					COVERAGE_RECORD(__LINE__);
				}
				COVERAGE_RECORD(__LINE__);
			}
			
			// Add to using-list's head
			{
				pTarget->sanity_u |= USING_MARK;
				pTarget->pPrev = NULL;
				pTarget->pNext = m_pUsingListHead;
				
				if(m_pUsingListHead)
				{
					m_pUsingListHead->pPrev = pTarget;
					COVERAGE_RECORD(__LINE__);
				}
				pTarget->pPrev = NULL;
				pTarget->pNext = m_pUsingListHead;
				m_pUsingListHead = pTarget;
				COVERAGE_RECORD(__LINE__);
			}

			// return data pointer
			STATISTICS_ADD(m_stat.current.usingBytes, pTarget->blockSize);
			STATISTICS_SUB(m_stat.current.freeBytes, pTarget->blockSize);
			STATISTICS_MAX(m_stat.maxUsedBytes, m_stat.current.usingBytes);
			STATISTICS_MIN(m_stat.minFreeBytes,  m_stat.current.freeBytes);
			STATISTICS_MAX(m_stat.maxAllocateSeeks, count);
			COVERAGE_RECORD(__LINE__);
			MEM_TRACE((" - %p\n", (void*) ((unsigned long)pTarget + sizeof(BLK_HEADER))));
			return (void*) ((unsigned long)pTarget + sizeof(BLK_HEADER));
		}
		pTarget = pTarget->pNext;
		PAFEE_ASSERT((unsigned long) pTarget < m_endAddr);
		COVERAGE_RECORD(__LINE__);
#ifdef PAFEE_STATISTICS_ENABLE
		count++;
#endif
	}
	COVERAGE_RECORD(__LINE__);

	STATISTICS_INC(m_stat.allocateFailure);
	
#ifdef PAFEE_EXCEPTION
		throw std::bad_alloc();
#endif
	MEM_TRACE((" No memory\n"));
	return NULL;
}

//--------------------------------------------
//! Deallocate memory
//--------------------------------------------
void FlexibleAllocator::Deallocate(
								void* ptr	///< [in] poiner to deallocate
								)
{
	MEM_TRACE(("Deallocate %p\n", ptr));
	BLK_HEADER* pTarget = (BLK_HEADER*)((unsigned long)ptr - sizeof(BLK_HEADER));

	PAFEE_ASSERT((m_heapAddr <= (unsigned long)ptr ) && ((unsigned long)ptr < (m_heapAddr+m_heapSize)));

	STATISTICS_INC(m_stat.deallocateRequest);
	if(ptr == NULL)
	{
		COVERAGE_RECORD(__LINE__);
		return;
	}

	// check sanity code
	if((pTarget->sanity_u & 0xfffffffe) != SANITY_CODE)
	{
		STATISTICS_INC(m_stat.sanityErrors);
		COVERAGE_RECORD(__LINE__);
#ifdef PAFEE_EXCEPTION
		throw std::runtime_error("damaged memory");
#endif
		return;
	}
	// already deallocated
	if((pTarget->sanity_u & USING_MARK) == 0)
	{
		COVERAGE_RECORD(__LINE__);
		return;
	}
	// Clear using mark
	pTarget->sanity_u &= ~USING_MARK;

	// Updatet statistitcs
	STATISTICS_SUB(m_stat.current.usingBytes, pTarget->blockSize);
	STATISTICS_ADD(m_stat.current.freeBytes, pTarget->blockSize);
	STATISTICS_MAX(m_stat.maxUsedBytes, m_stat.current.usingBytes);
	STATISTICS_MIN(m_stat.minFreeBytes,  m_stat.current.freeBytes);

	// Remove from using-list
	{
//		PAFEE_TRACE("Deallocate %p %p %p ulh:%p ub:%ld\n", pTarget->pPrev, pTarget, pTarget->pNext, m_pUsingListHead, m_stat.current.usingBytes);
		if(pTarget->pPrev)
		{
			pTarget->pPrev->pNext = pTarget->pNext;
			COVERAGE_RECORD(__LINE__);
		}
		else
		{
			m_pUsingListHead = pTarget->pNext;
			if(m_pUsingListHead)
			{
				m_pUsingListHead->pPrev = NULL;
				COVERAGE_RECORD(__LINE__);
			}
		}
		if(pTarget->pNext)
		{
			pTarget->pNext->pPrev = pTarget->pPrev;
			CHECK_LOOP(pTarget->pNext);
			COVERAGE_RECORD(__LINE__);
		}
		COVERAGE_RECORD(__LINE__);
	}

	// Integration style
	enum {NO_INTEGRATION, FORWARD, BACKWARD, BOTH};
	int integration_style = NO_INTEGRATION;
	{
		if(pTarget->pFrontBlock)
		{
			if((pTarget->pFrontBlock->sanity_u & USING_MARK) == 0)
			{
				integration_style = FORWARD;
				COVERAGE_RECORD(__LINE__);
			}
			COVERAGE_RECORD(__LINE__);
		}
		if(pTarget->pBackBlock)
		{
			if((pTarget->pBackBlock->sanity_u & USING_MARK) == 0)
			{
				integration_style  = (integration_style == FORWARD)? BOTH : BACKWARD;
				COVERAGE_RECORD(__LINE__);
			}
			COVERAGE_RECORD(__LINE__);
		}
		COVERAGE_RECORD(__LINE__);
	}

	// Integrate target into forward neayby-block
	if(integration_style == FORWARD)
	{
		//
		// [nearby->frontBlk][nearby][<=target][target->backBlk]
		//
		BLK_HEADER* pFrontNearby = pTarget->pFrontBlock;
		
		if(pTarget->pBackBlock)
		{
			pTarget->pBackBlock->pFrontBlock = pFrontNearby;
			COVERAGE_RECORD(__LINE__);
		}
		pFrontNearby->pBackBlock = pTarget->pBackBlock;
		pFrontNearby->blockSize += pTarget->blockSize;
	}
	// Integrate target into backward neayby-block
	else if(integration_style == BACKWARD)
	{
		//
		// [target->frontBlk][target][<=nearby][nearby->backBlk]
		//
		BLK_HEADER* pBackNearby = pTarget->pBackBlock;
		if(pBackNearby->pBackBlock)
		{
			pBackNearby->pBackBlock->pFrontBlock = pTarget;
			COVERAGE_RECORD(__LINE__);
		}
		pTarget->pBackBlock = pBackNearby->pBackBlock;
		pTarget->blockSize += pBackNearby->blockSize;
		pTarget->pPrev = pBackNearby->pPrev;
		pTarget->pNext = pBackNearby->pNext;
		CHECK_LOOP(pTarget);

		// backNearby is integrated into target, 
		// So, adjust linkage of free-list
		if(pBackNearby->pPrev)
		{
			pBackNearby->pPrev->pNext = pTarget;
			CHECK_LOOP(pBackNearby->pPrev);
			COVERAGE_RECORD(__LINE__);
		}
		if(pBackNearby->pNext)
		{
			pBackNearby->pNext->pPrev = pTarget;
			CHECK_LOOP(pBackNearby->pNext);
			COVERAGE_RECORD(__LINE__);
		}
		if(m_pFreeListHead == pBackNearby)
		{
			m_pFreeListHead = pTarget;
			COVERAGE_RECORD(__LINE__);
		}
		COVERAGE_RECORD(__LINE__);
	}
	// Integrate target and backword neayby-block into forward nearby-block
	else if(integration_style == BOTH)
	{
		//
		// [frontNearby->frontBlk][frontNearby][<=target][<=backNearby][backNearby->backBlk]
		//
		BLK_HEADER* pFrontNearby = pTarget->pFrontBlock;
		BLK_HEADER* pBackNearby  = pTarget->pBackBlock;
		
		if(pBackNearby->pBackBlock)
		{
			pBackNearby->pBackBlock->pFrontBlock = pFrontNearby;
			COVERAGE_RECORD(__LINE__);
		}
		pFrontNearby->pBackBlock = pBackNearby->pBackBlock;
		pFrontNearby->blockSize += pTarget->blockSize;
		pFrontNearby->blockSize += pBackNearby->blockSize;

		// Remove backNearby from free-list
		if(pBackNearby->pPrev)
		{
			pBackNearby->pPrev->pNext = pBackNearby->pNext;
			if(pBackNearby->pNext)
			{
				pBackNearby->pNext->pPrev = pBackNearby->pPrev;
				CHECK_LOOP(pBackNearby->pNext);
			}
			CHECK_LOOP(pBackNearby->pPrev);
			COVERAGE_RECORD(__LINE__);
		}
		// Deleting backNearby is the head of free-list?
		if(pBackNearby == m_pFreeListHead)
		{
			if(pBackNearby->pNext)
			{
				PAFEE_ASSERT(pBackNearby->pNext->pPrev == pBackNearby);
				pBackNearby->pNext->pPrev = NULL;
				COVERAGE_RECORD(__LINE__);
			}
			m_pFreeListHead = pBackNearby->pNext;
			COVERAGE_RECORD(__LINE__);
		}
		COVERAGE_RECORD(__LINE__);
	}
	// No integration
	else
	{
		// Add to free-list
		if(m_pFreeListHead)
		{
			PAFEE_ASSERT(m_pFreeListHead->pNext != pTarget);
			PAFEE_ASSERT(m_pFreeListHead->pPrev == NULL);
			m_pFreeListHead->pPrev = pTarget;
			CHECK_LOOP(m_pFreeListHead);

			pTarget->pPrev = NULL;
			pTarget->pNext = m_pFreeListHead;
			m_pFreeListHead = pTarget;
			CHECK_LOOP(pTarget);
			COVERAGE_RECORD(__LINE__);
		}
		else
		{
			pTarget->pPrev = NULL;
			pTarget->pNext = NULL;
			m_pFreeListHead = pTarget;
			COVERAGE_RECORD(__LINE__);
		}
		COVERAGE_RECORD(__LINE__);
	}
	COVERAGE_RECORD(__LINE__);
}
#ifdef PAFEE_STATISTICS_ENABLE
//--------------------------------------------
//! Get statistics
//--------------------------------------------
void FlexibleAllocator::GetStatistics(FlexibleAllocator::STATISTICS* pStat)
{
	pafee_memset(&m_stat.current, 0 ,sizeof(m_stat.current));
	m_stat.current.minBlockSizeInFreeList = 0xffffffff;
	m_stat.current.minBlockSizeInUsingList= 0xffffffff;

	// Get free-list statistics
	{
		BLK_HEADER* pTarget = m_pFreeListHead;
		while(pTarget)
		{
			if((pTarget->sanity_u & 0xfffffffe) != SANITY_CODE)
			{
				m_stat.current.sanityErrorInFreeList = (unsigned long) pTarget;
				COVERAGE_RECORD(__LINE__);
				break;
			}			
			m_stat.current.freeBytes += pTarget->blockSize;
			m_stat.current.freeListElements++;
			if(pTarget->blockSize > m_stat.current.maxBlockSizeInFreeList)
			{
				m_stat.current.maxBlockSizeInFreeList = pTarget->blockSize;
				COVERAGE_RECORD(__LINE__);
			}
			if(pTarget->blockSize < m_stat.current.minBlockSizeInFreeList)
			{
				m_stat.current.minBlockSizeInFreeList = pTarget->blockSize;
				COVERAGE_RECORD(__LINE__);
			}
			unsigned long blockSize;
			blockSize = m_maxAllocateSize/1+sizeof(BLK_HEADER);
			m_stat.current.availableMaxSizeBlocks += (pTarget->blockSize/blockSize);
			blockSize = m_maxAllocateSize/2+sizeof(BLK_HEADER);
			if(blockSize < m_minBlockSize)
			{
				blockSize = m_minBlockSize;
			}
			m_stat.current.available1_2SizeBlocks += (pTarget->blockSize/blockSize);
			blockSize = m_maxAllocateSize/4+sizeof(BLK_HEADER);
			if(blockSize < m_minBlockSize)
			{
				blockSize = m_minBlockSize;
			}
			m_stat.current.available1_4SizeBlocks += (pTarget->blockSize/blockSize);
			blockSize = m_maxAllocateSize/8+sizeof(BLK_HEADER);
			if(blockSize < m_minBlockSize)
			{
				blockSize = m_minBlockSize;
			}
			m_stat.current.available1_8SizeBlocks += (pTarget->blockSize/blockSize);
			pTarget = pTarget->pNext;
			COVERAGE_RECORD(__LINE__);
		}
		COVERAGE_RECORD(__LINE__);
	}
	// Get using-list statistics
	{
		BLK_HEADER* pTarget = m_pUsingListHead;
		while(pTarget)
		{
			if((pTarget->sanity_u & 0xfffffffe) != SANITY_CODE)
			{
				m_stat.current.sanityErrorInUsingList = (unsigned long)pTarget;
				COVERAGE_RECORD(__LINE__);
				break;
			}

			m_stat.current.usingBytes += pTarget->blockSize;
			m_stat.current.usingListElements++;
			if(pTarget->blockSize > m_stat.current.maxBlockSizeInUsingList)
			{
				m_stat.current.maxBlockSizeInUsingList = pTarget->blockSize;
				COVERAGE_RECORD(__LINE__);
			}
			if(pTarget->blockSize < m_stat.current.minBlockSizeInUsingList)
			{
				m_stat.current.minBlockSizeInUsingList = pTarget->blockSize;
				COVERAGE_RECORD(__LINE__);
			}
			pTarget = pTarget->pNext;
			COVERAGE_RECORD(__LINE__);
		}
		COVERAGE_RECORD(__LINE__);
	}
	if(m_stat.current.minBlockSizeInFreeList == 0xffffffff)
	{
		m_stat.current.minBlockSizeInFreeList = 0;
	}
	if(m_stat.current.minBlockSizeInUsingList == 0xffffffff)
	{
		m_stat.current.minBlockSizeInUsingList = 0;
	}

	*pStat = m_stat;
	COVERAGE_RECORD(__LINE__);
}

//--------------------------------------------
//! Get statistics
//--------------------------------------------
void FlexibleAllocator::GetStatistics(StreamOutput& stream)
{
	STATISTICS stat;
	GetStatistics(&stat);

	stream.Write("==== FlexibleAllocator statistics ====\n");
	stream.Write("heap address: %p (ownalloc:%d) \n", m_heapAddr, m_isOwnHeapAllocation);
	stream.Write("heap size: %ld\n", m_heapSize);
	stream.Write("max allocation limit: %ld\n", m_maxAllocateSize);
	stream.Write("min allocation limit: %ld\n", m_minBlockSize);
	stream.Write("-- statistics --\n");
	stream.Write("allocation count: %ld\n",		stat.allocateRequest);
	stream.Write("dealocation count: %ld\n",	stat.deallocateRequest);
	stream.Write("allocation failure count: %ld\n",		stat.allocateFailure);
	stream.Write("sanity check error count: %ld\n",		stat.sanityErrors);
	stream.Write("max used memory size: %ld\n",			stat.maxUsedBytes);
	stream.Write("min free memory size: %ld\n",			stat.minFreeBytes);
	stream.Write("max allocation size: %ld\n",	stat.maxAllocateSize);
	stream.Write("min allocation size: %ld\n",	stat.minAllocateSize);
	stream.Write("max seek count on allocation: %ld\n", 	stat.maxAllocateSeeks);
	stream.Write("-- current statistics --\n");
	stream.Write("using memory size: %ld\n", stat.current.usingBytes);	
	stream.Write("free memory size: %ld\n", stat.current.freeBytes);
	stream.Write("free-list element count: %ld\n", stat.current.freeListElements);
	stream.Write("using-list element count: %ld\n", stat.current.usingListElements);
	stream.Write("min size in free-list: %ld\n", stat.current.minBlockSizeInFreeList);
	stream.Write("max size in free-list: %ld\n", stat.current.maxBlockSizeInFreeList);
	stream.Write("max size in using-list: %ld\n", stat.current.minBlockSizeInUsingList);
	stream.Write("min size in using-list: %ld\n", stat.current.maxBlockSizeInUsingList);
	stream.Write("sanity error address in free-list: %ld\n", stat.current.sanityErrorInFreeList);
	stream.Write("sanity error address in using-list: %ld\n", stat.current.sanityErrorInUsingList);
	stream.Write("available max size allocation: %ld\n", stat.current.availableMaxSizeBlocks);
	stream.Write("available 1/2 size allocation: %ld\n", stat.current.available1_2SizeBlocks);
	stream.Write("available 1/4 size allocation: %ld\n", stat.current.available1_4SizeBlocks);
	stream.Write("available 1/8 size allocation: %ld\n", stat.current.available1_8SizeBlocks);
}

#endif

//--------------------------------------------
//! Check given address is in own address range
//! @return true: own pointer   false: not own
//--------------------------------------------
bool FlexibleAllocator::IsOwnAddress(void* ptr)
{
	if(	(m_heapAddr <= (unsigned long)ptr) && 
		((unsigned long)ptr < (m_heapAddr+m_heapSize) ))
	{
		COVERAGE_RECORD(__LINE__);
		return true;
	}
	COVERAGE_RECORD(__LINE__);
	return false;
}

#ifdef PAFEE_UNITTEST
#ifdef PAFEE_STATISTICS_ENABLE
#define HEAP_SIZE	1000
//--------------------------------------------
//! Unit test
//--------------------------------------------
void FlexibleAllocator::UnitTest(void)
{
	unsigned char* pBuf = new unsigned char[HEAP_SIZE];
	FlexibleAllocator alloc;
	STATISTICS stat;
	void* p[200];

	//PAFEE_TRACE("UnitTest pBuf:%p\n", pBuf);
	alloc.Initialize((unsigned long) pBuf, HEAP_SIZE, 64, 16);
	
	// Divide and integrate #1
	{
		printf("====  Divide and integrate #1 ====\n");
		alloc.ShowLists();
		p[0] = alloc.Allocate(16);
		printf("p0:%p(%lx) f:%p u:%p\n", p[0], ((unsigned long)p[0] - sizeof(BLK_HEADER)),alloc.m_pFreeListHead, alloc.m_pUsingListHead);
		pafee_memset(p[0], 0xAA, 16);

		p[1] = alloc.Allocate(32);
		printf("p1:%p(%lx)  f:%p u:%p\n", p[1], ((unsigned long)p[1] - sizeof(BLK_HEADER)),alloc.m_pFreeListHead, alloc.m_pUsingListHead);
		pafee_memset(p[1], 0xBB, 32);
		
		p[2] = alloc.Allocate(64);
		printf("p2:%p(%lx)  f:%p u:%p\n", p[2], ((unsigned long)p[2] - sizeof(BLK_HEADER)),alloc.m_pFreeListHead, alloc.m_pUsingListHead);
		pafee_memset(p[2], 0xCC, 64);

		p[3] = alloc.Allocate(32);
		printf("p3:%p(%lx)  f:%p u:%p\n", p[3], ((unsigned long)p[3] - sizeof(BLK_HEADER)),alloc.m_pFreeListHead, alloc.m_pUsingListHead);
		pafee_memset(p[3], 0xDD, 32);

		p[4] = alloc.Allocate(32);
		printf("p4:%p(%lx)  f:%p u:%p\n", p[4], ((unsigned long)p[4] - sizeof(BLK_HEADER)),alloc.m_pFreeListHead, alloc.m_pUsingListHead);
		pafee_memset(p[4], 0xEE, 32);

		PAFEE_ASSERT(alloc.IsOwnAddress(p[0]) == true);
		PAFEE_ASSERT(alloc.IsOwnAddress(p[4]) == true);
		PAFEE_ASSERT(alloc.IsOwnAddress(pBuf+1001) == false);
		
		alloc.GetStatistics(&stat);

		alloc.ShowLists();
		alloc.Deallocate(p[0]);
		printf("deallocated:%p\n", p[0]);
		alloc.ShowLists();
		alloc.Deallocate(p[1]);
		printf("deallocated:%p\n", p[1]);
		alloc.ShowLists();
		alloc.Deallocate(p[2]);
		printf("deallocated:%p\n", p[2]);
		alloc.ShowLists();
		alloc.Deallocate(p[3]);
		printf("deallocated:%p\n", p[3]);
		alloc.ShowLists();
		alloc.Deallocate(p[4]);
		printf("deallocated:%p\n", p[4]);
		alloc.ShowLists();
	}

	// Divide and integrate #2
	{
		printf("\n====  Divide and integrate #2 ====\n");
		alloc.ShowLists();
		p[0] = alloc.Allocate(16);
		printf("p0:%p(%lx) f:%p u:%p\n", p[0], ((unsigned long)p[0] - sizeof(BLK_HEADER)),alloc.m_pFreeListHead, alloc.m_pUsingListHead);
		pafee_memset(p[0], 0xAA, 16);

		p[1] = alloc.Allocate(32);
		printf("p1:%p(%lx)  f:%p u:%p\n", p[1], ((unsigned long)p[1] - sizeof(BLK_HEADER)),alloc.m_pFreeListHead, alloc.m_pUsingListHead);
		pafee_memset(p[1], 0xBB, 32);
		
		p[2] = alloc.Allocate(64);
		printf("p2:%p(%lx)  f:%p u:%p\n", p[2], ((unsigned long)p[2] - sizeof(BLK_HEADER)),alloc.m_pFreeListHead, alloc.m_pUsingListHead);
		pafee_memset(p[2], 0xCC, 64);

		p[3] = alloc.Allocate(32);
		printf("p3:%p(%lx)  f:%p u:%p\n", p[3], ((unsigned long)p[3] - sizeof(BLK_HEADER)),alloc.m_pFreeListHead, alloc.m_pUsingListHead);
		pafee_memset(p[3], 0xDD, 32);

		p[4] = alloc.Allocate(32);
		printf("p4:%p(%lx)  f:%p u:%p\n", p[4], ((unsigned long)p[4] - sizeof(BLK_HEADER)),alloc.m_pFreeListHead, alloc.m_pUsingListHead);
		pafee_memset(p[4], 0xEE, 32);

		alloc.GetStatistics(&stat);

		alloc.ShowLists();
		alloc.Deallocate(p[4]);
		printf("deallocated:%p\n", p[4]);
		alloc.ShowLists();
		alloc.Deallocate(p[1]);
		printf("deallocated:%p\n", p[1]);
		alloc.ShowLists();
		alloc.Deallocate(p[3]);
		printf("deallocated:%p\n", p[3]);
		alloc.ShowLists();
		alloc.Deallocate(p[2]);
		printf("deallocated:%p\n", p[2]);
		alloc.ShowLists();
		alloc.Deallocate(p[0]);
		printf("deallocated:%p\n", p[0]);
		alloc.ShowLists();
	}
	
	// Divide and integrate #3
	{
		printf("\n====  Divide and integrate #3 ====\n");
		alloc.ShowLists();
		p[0] = alloc.Allocate(16);
		printf("p0:%p(%lx) \n", p[0], ((unsigned long)p[0] - sizeof(BLK_HEADER)));
		pafee_memset(p[0], 0xAA, 16);

		p[1] = alloc.Allocate(16);
		printf("p1:%p(%lx) \n", p[1], ((unsigned long)p[1] - sizeof(BLK_HEADER)));
		pafee_memset(p[1], 0xbb, 16);

		p[2] = alloc.Allocate(48);
		printf("p2:%p(%lx) \n", p[2], ((unsigned long)p[2] - sizeof(BLK_HEADER)));
		pafee_memset(p[2], 0xcc, 48);

		p[3] = alloc.Allocate(48);
		printf("p3:%p(%lx) \n", p[3], ((unsigned long)p[3] - sizeof(BLK_HEADER)));
		pafee_memset(p[3], 0xdd, 48);

		p[4] = alloc.Allocate(64);
		printf("p4:%p(%lx) \n", p[4], ((unsigned long)p[4] - sizeof(BLK_HEADER)));
		pafee_memset(p[4], 0xee, 64);
		alloc.ShowLists();

		alloc.Deallocate(p[1]);
		alloc.Deallocate(p[3]);
		alloc.ShowLists();
		
		p[5] = alloc.Allocate(24);
		printf("p0:%p(%lx) \n", p[5], ((unsigned long)p[5] - sizeof(BLK_HEADER)));
		pafee_memset(p[5], 0xff, 24);
		alloc.ShowLists();
		p[6] = alloc.Allocate(16);
		pafee_memset(p[6], 0xff, 16);
		alloc.ShowLists();

		alloc.Deallocate(p[6]);
		alloc.ShowLists();
		alloc.Deallocate(p[5]);
		alloc.ShowLists();
		alloc.Deallocate(p[0]);
		alloc.ShowLists();
		alloc.Deallocate(p[2]);
		alloc.ShowLists();
		alloc.Deallocate(p[4]);
		alloc.ShowLists();
	}
	// Divide and integrate #4
	{
		printf("\n====  Divide and integrate #4 ====\n");
		alloc.ShowLists();
		p[0] = alloc.Allocate(32);
		pafee_memset(p[0], 0xff, 32);
		p[1] = alloc.Allocate(32);
		pafee_memset(p[1], 0xff, 32);
		p[2] = alloc.Allocate(64);
		pafee_memset(p[2], 0xff, 64);
		p[3] = alloc.Allocate(64);
		pafee_memset(p[3], 0xff, 64);
		
		alloc.Deallocate(p[2]);
		alloc.Deallocate(p[0]);
		alloc.ShowLists();

		p[4] = alloc.Allocate(64);
		pafee_memset(p[4], 0xff, 64);
		alloc.ShowLists();
		alloc.Deallocate(p[4]);
		alloc.Deallocate(p[1]);
		alloc.Deallocate(p[3]);
		alloc.ShowLists();
	}
	// Divide and integrate #5
	{
		printf("\n====  Divide and integrate #5 ====\n");
		alloc.ShowLists();
		p[0] = alloc.Allocate(64);
		pafee_memset(p[0], 0xff, 64);
		p[1] = alloc.Allocate(64);
		pafee_memset(p[1], 0xff, 64);
		p[2] = alloc.Allocate(64);
		pafee_memset(p[2], 0xff, 64);
		alloc.Deallocate(p[1]);
		p[4] = alloc.Allocate(16);
		pafee_memset(p[4], 0xff, 16);
		alloc.Deallocate(p[0]);
		alloc.Deallocate(p[4]);
		alloc.Deallocate(p[2]);
		alloc.ShowLists();
	}

	{
		printf("\n==== Abnormal Deallocation ====\n");
		p[0] = alloc.Allocate(64);
		alloc.Deallocate(p[0]);
		alloc.Deallocate(p[0]);
		alloc.Deallocate(NULL);
	}
	
	{
		printf("\n==== Max Allocation ====\n");
		int i;
		int count;
		for(i = 0; i < 100; i++)
		{
			if((p[i] = alloc.Allocate(40)) == NULL)
			{
				printf("can't allocate at %d times(%dbytes)\n", i, i*40);
				count = i+1;
				break;
			}
		}
		alloc.ShowLists();
		for(i = 0; i < count; i++)
		{
			alloc.Deallocate(p[i]);
		}
		alloc.ShowLists();
		
		void* ptr = alloc.Allocate(64);
		assert(ptr);
		alloc.Deallocate(ptr);

		ptr = alloc.Allocate(64+1);
		assert(ptr == NULL);
	}
	
	{
		printf("\n==== check statistics ====\n");
		int i;
		int index = 0;
		p[index++] = alloc.Allocate(32);
		p[index++] = alloc.Allocate(32);
		p[index++] = alloc.Allocate(64);
		alloc.GetStatistics(&stat);
		for(i = 0; i < 1000000; i++)
		{
			void* ptr = alloc.Allocate(64);
			if(ptr)
			{
				p[index++] = ptr;
			}
			else
			{
				PAFEE_ASSERT((index-3) == (int) stat.current.availableMaxSizeBlocks);
				break;
			}
		}
		for(i = 0; i < index; i++)
		{
			alloc.Deallocate(p[i]);
		}
		index = 0;
		p[index++] = alloc.Allocate(32);
		p[index++] = alloc.Allocate(32);
		p[index++] = alloc.Allocate(64);
		alloc.GetStatistics(&stat);
		for(i = 0; i < 1000000; i++)
		{
			void* pp;
			void* ptr = alloc.Allocate(64/2);
			if(ptr)
			{
				p[index++] = ptr;
				pp = ptr;
			}
			else
			{
				PAFEE_ASSERT(index-3 == (int)stat.current.available1_2SizeBlocks);
				break;
			}
		}
		for(i = 0; i < index; i++)
		{
			alloc.Deallocate(p[i]);
		}
		

		index = 0;
		p[index++] = alloc.Allocate(32);
		p[index++] = alloc.Allocate(32);
		p[index++] = alloc.Allocate(64);
		alloc.GetStatistics(&stat);
		for(i = 0; i < 1000000; i++)
		{
			void* ptr = alloc.Allocate(64/4);
			if(ptr)
			{
				p[index++] = ptr;
			}
			else
			{
				PAFEE_ASSERT(index-3 == (int)stat.current.available1_4SizeBlocks);
				break;
			}
		}
		for(i = 0; i < index; i++)
		{
			alloc.Deallocate(p[i]);
		}

		index = 0;
		p[index++] = alloc.Allocate(32);
		p[index++] = alloc.Allocate(32);
		p[index++] = alloc.Allocate(64);
		alloc.GetStatistics(&stat);
		for(i = 0; i < 1000000; i++)
		{
			void* ptr = alloc.Allocate(64/8);
			if(ptr)
			{
				p[index++] = ptr;
			}
			else
			{
				alloc.ShowLists();
				PAFEE_ASSERT(index-3 == (int)stat.current.available1_8SizeBlocks);
				break;
			}
		}
		for(i = 0; i < index; i++)
		{
			alloc.Deallocate(p[i]);
		}

	}
	{
		printf("\n==== Delete invalid address ====\n");
		p[0] = alloc.Allocate(64);
		p[1] = alloc.Allocate(64);
		p[2] = alloc.Allocate(64);
		alloc.Deallocate(p[1]);

		printf("\n==== detect damaged memory ====\n");
		*(((unsigned char*) p[1]) -24) = 0xee;
		*(((unsigned char*) p[2]) -24) = 0xee;
		alloc.Deallocate(p[2]);
		alloc.GetStatistics(&stat);
		alloc.ShowLists();
	}

	printf("\n==== unalignment address ====\n");
	{
		FlexibleAllocator a;
		char buf[100];
		a.Initialize( (unsigned long) buf+1, 100, 64, 16);
		PAFEE_ASSERT(a.Allocate(100) == false);

	}
#ifdef PAFEE_COVERAGE_RECORD
	g_coverage.Show();
#endif
	delete pBuf;
}

void FlexibleAllocator::ShowLists(void)
{
	{
		printf("free-list:  ");
		BLK_HEADER* pTarget = m_pFreeListHead;
		while(pTarget)
		{
			if((pTarget->sanity_u & 0xfffffffe) == SANITY_CODE)
			{
				printf("[%p:%ld:%X] ", pTarget, pTarget->blockSize, *((unsigned char *) ((unsigned long)pTarget+sizeof(BLK_HEADER))));
				pTarget = pTarget->pNext;
			}
			else
			{
				printf("\nXXXX DAMAGED at %p XXXX\n", pTarget);
				break;
			}
			COVERAGE_RECORD(__LINE__);
		}
		printf("\n");
	}
	{
		printf("using-list: ");
		BLK_HEADER* pTarget = m_pUsingListHead;
		while(pTarget)
		{
			if((pTarget->sanity_u & 0xfffffffe) == SANITY_CODE)
			{
				printf("[%p:%ld:%X] ", pTarget, pTarget->blockSize, *((unsigned char *) ((unsigned long)pTarget+sizeof(BLK_HEADER))));
				pTarget = pTarget->pNext;
			}
			else
			{
				printf("\nXXXX DAMAGED at %p XXXX\n", pTarget);
				break;
			}
			COVERAGE_RECORD(__LINE__);
		}
		printf("\n");
	}
	COVERAGE_RECORD(__LINE__);
}

#endif
#endif

