//********************************************************************
//
//! 	 @file: 	PafeeStatisticsAllocator.cpp
//! 	 @author:	T.Miyawaki
//! 	 @brief:	Statistics memory allocator
//
//********************************************************************

//
// P A F E E   - Pragmatic Application Framework for Embedded Environment- 
//
//
// Copyright (c) 2009 KN Information Sustems Corporation,
//                    Hirocon Corporation,
//                    Technocrats Corporation
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//

#include <stdlib.h>
#include "PafeeCommon.h"
#include "pafee_string.h"
#include "PafeeStatisticsAllocator.h"

using namespace Pafee;

const unsigned int StatisticsAllocator::EMPTY = 0xffffffff;

//----------------------------------------------------------------
//! Constractor
//----------------------------------------------------------------
StatisticsAllocator::StatisticsAllocator(
											Allocator* pAllocator	///< [in] Allocator
											)
{
	Initialize(4, pAllocator);
}

//----------------------------------------------------------------
//! Destoractor
//----------------------------------------------------------------
StatisticsAllocator::~StatisticsAllocator()
{
}

//----------------------------------------------------------------
//! Initialize
//----------------------------------------------------------------
void StatisticsAllocator::Initialize(
									int resolutionBits,				///< [in] record resolution bits
									Allocator* pAllocator			///< [in] allocator
									)
{
	m_pBackendAllocator = pAllocator;
	SetResolution(resolutionBits);
	pafee_memset(&m_stat, 0, sizeof(m_stat));
	for(int i = 0; i < MAX_RECORD_COUNT; i++)
	{
		m_stat.usage[i].allocateSize = EMPTY;
		m_stat.usage[i].allocateCount = 0;
	}
}

//----------------------------------------------------------------
//! Allocate memory
//----------------------------------------------------------------
void* StatisticsAllocator::Allocate(
									unsigned long size		///< [in] size to allocate
									)
{
	m_stat.totalAllocateRequests++;
	UpdateStatistics(size);
	if(m_pBackendAllocator)
	{
		return m_pBackendAllocator->Allocate(size);
	}
	return malloc(size);
}

//----------------------------------------------------------------
//! Deallocate memory
//----------------------------------------------------------------
void StatisticsAllocator::Deallocate(
									void* ptr				///< [in] pointer to deallocate
									)
{
	m_stat.totalDeallocateRequests++;
	if(m_pBackendAllocator)
	{
		m_pBackendAllocator->Deallocate(ptr);
	}
	else
	{
		free(ptr);
	}
}


//----------------------------------------------------------------
//! Check given address is in own address range
//----------------------------------------------------------------
bool StatisticsAllocator::IsOwnAddress(
									void* ptr				///< [in] pointer to check
									)
{
	if(m_pBackendAllocator)
	{
		return m_pBackendAllocator->IsOwnAddress(ptr);
	}
	return true;
}
//----------------------------------------------------------------
//! Get statistics
//----------------------------------------------------------------
void StatisticsAllocator::GetStatistics(
									StatisticsAllocator::STATISTICS* pStat	///< [in] statistics information
									)
{
	pafee_memcpy(pStat, &m_stat, sizeof(m_stat));
}

#ifdef PAFEE_STATISTICS_ENABLE
//----------------------------------------------------------------
//! Get statistics
//----------------------------------------------------------------
void StatisticsAllocator::GetStatistics(StreamOutput& stream)
{
	int elementCount = 0;

	// Get element count of usage table
	{
		for(elementCount = 0; elementCount < MAX_RECORD_COUNT; elementCount++)
		{
			if(m_stat.usage[elementCount].allocateSize == EMPTY)
			{
				break;
			}
		}
	}
	
	// Bubble sort, it's slow, but no need for extra memory.
	{
		for(int i = 0; i < elementCount-1; i++)
		{
			for(int j = elementCount-1; j > i ; j--)
			{
				if(m_stat.usage[j].allocateSize < m_stat.usage[j-1].allocateSize)
				{
					Swap(m_stat.usage[j].allocateSize, m_stat.usage[j-1].allocateSize);
					Swap(m_stat.usage[j].allocateCount, m_stat.usage[j-1].allocateCount);
				}
			}
		}
	}
	
	// Print
	{
		unsigned int extra = 0;
		unsigned int bit = 0x00000001;
		for(unsigned int index = 0; index < m_resolution; index++)
		{
			extra |= bit << index;
		}
		
		stream.Write("--- Allocation statistics ---\n");
		stream.Write("resolution unit:%d bits\n", m_resolution);
		stream.Write("total allocation requests:%d\n", m_stat.totalAllocateRequests);
		stream.Write("total deallocation requests:%d\n", m_stat.totalDeallocateRequests);
		stream.Write("allocation size Min-Max: count\n");
		for(int i = 0; i < elementCount; i++)
		{
			if(m_stat.usage[i].allocateSize != EMPTY)
			{
				stream.Write("%d-%d: %d\n",	(m_stat.usage[i].allocateSize << m_resolution),
											(m_stat.usage[i].allocateSize << m_resolution)+extra,
											m_stat.usage[i].allocateCount);
			}
			else
			{
				break;
			}
		}
		if(m_stat.overflow)
		{
			stream.Write("=== WARNING ===\n");
			stream.Write("Record table is full, %d allocation requests are ignored.\n", m_stat.overflow);
		}
	}
}
#endif

//----------------------------------------------------------------
//! SetAllocator
//----------------------------------------------------------------
void StatisticsAllocator::SetAllocator(
									Allocator* pAllocator			///< [in] allocator
									)
{
	m_pBackendAllocator = pAllocator;
}

//----------------------------------------------------------------
//! Set record resolution
//----------------------------------------------------------------
void StatisticsAllocator::SetResolution(
									int resolutionBits				///< [in] record resolution bits
									)
{
	m_resolution = 0;
	while(resolutionBits > 1)
	{
		m_resolution++;
		resolutionBits = resolutionBits >> 1;
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////
// protected members

//----------------------------------------------------------------
//! Update statistitcs
//----------------------------------------------------------------
void StatisticsAllocator::UpdateStatistics(size_t allocateSize)
{
	unsigned int size = allocateSize >> m_resolution;
	for(int i = 0; i < MAX_RECORD_COUNT; i++)
	{
		if(m_stat.usage[i].allocateSize == size)
		{
			m_stat.usage[i].allocateCount++;
			return;
		}
		else if(m_stat.usage[i].allocateSize == EMPTY)
		{
			m_stat.usage[i].allocateSize = size;
			m_stat.usage[i].allocateCount++;
			return;
		}
	}
	m_stat.overflow++;
};

//----------------------------------------------------------------
//! Swap data
//----------------------------------------------------------------
void StatisticsAllocator::Swap(unsigned int& dataA, unsigned int& dataB)
{
	int temp;
	temp = dataA;
	dataA = dataB;
	dataB = temp;
}


