//********************************************************************
//		project 	PAFEE
//
//! 	 @file: 	PafeeSemaphore.h
//! 	 @author:	Technocrats
//! 	 @brief:	PafeeSemaphore class header
//
//********************************************************************
//
// P A F E E   - Pragmatic Application Framework for Embedded Environment- 
//
//
// Copyright (c) 2009 KN Information Sustems Corporation,
//                    Hirocon Corporation,
//                    Technocrats Corporation
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//

#ifndef _PAFEESEMAPHORE_H_INCLUDED_
#define _PAFEESEMAPHORE_H_INCLUDED_

// if no OS definition, PafeeSemaphore class is unavailable
#if defined(PAFEE_OS_WINDOWS)||defined(PAFEE_OS_LINUX)||defined(PAFEE_OS_ITRON)

#ifdef PAFEE_OS_WINDOWS
#include <windows.h>
#endif

#ifdef PAFEE_OS_LINUX
#include <sys/errno.h>
#include <time.h>
#include <semaphore.h>
#endif

#ifdef PAFEE_OS_ITRON
#include <kernel.h>
#include <itron.h>
#endif

#ifdef PAFEE_DEBUG
#include <stdio.h>
#endif

#ifdef __cplusplus
namespace Pafee
{
#endif

#ifdef PAFEE_DEBUG
#if defined(PAFEE_OS_WINDOWS)||defined(PAFEE_OS_LINUX)
#define SEM_DEBUG(STRING)   printf STRING
#elif defined(PAFEE_OS_ITRON)
#define SEM_DEBUG(STRING)   SemaphoreSyslog STRING
#endif
#else
#define SEM_DEBUG(STRING)
#endif

class PafeeSemaphore
{

public:
	/**
	 * @brief  Semaphore error enum
	 */
	typedef enum tagSEM_ERROR
	{
		ERROR_NONE, /**< Error none */
		ERROR_NOHANDLE, /**< function is called without handle */
		ERROR_HAS_HANDLE, /**< Allready has handle */
		ERROR_CANNOT_CREATE, /**< Can not create */
		ERROR_CANNOT_DELETE, /**< Can not delete */
		ERROR_CANNOT_GET, /**< Can not get */
		ERROR_TIME_OUT, /**< Can not get because of time out */
		ERROR_CANNOT_RELEASE, /**< Can not release */
		ERROR_MAX /**< MAX */
	} SEM_ERROR;

private:
#ifdef PAFEE_OS_WINDOWS
	HANDLE m_handle; /**< Windows semaphore handle */
#endif // PAFEE_OS_WINDOWS

#ifdef PAFEE_OS_LINUX
	sem_t m_handle; /**< Linux semaphore handle */
	bool m_isCreated; /**< semaphore created flag */
#endif // PAFEE_OS_LINUX

#ifdef PAFEE_OS_ITRON
	ID m_handle; /**< Itron semaphore handle */
#endif // PAFEE_OS_ITRON

	SEM_ERROR m_error;


public:
	/**
	 * @brief  PafeeSemaphore class constructor
	 * @param  (none)
	 * @return (none)
	 */
	PafeeSemaphore();

	/**
	 * @brief  PafeeSemaphore class destructor
	 * @param  (none)
	 * @return (none)
	 */
	~PafeeSemaphore();

#if defined(PAFEE_OS_WINDOWS) || defined(PAFEE_OS_LINUX)
	/**
	 * @brief  Make new semaphore
	 * @param  int : Maximum count of semaphore
	 * @return Create success or not
	 * @retval true:create succeeded
	 * @retval false:create failed
	 */
	bool Create(
		int maxCount ///< [in] Max count of semaphore
	);
#endif

#ifdef PAFEE_OS_ITRON
	/**
	 * @brief  Attach static semaphore id
	 * @return Attach success or not
	 * @retval true:attach succeeded
	 * @retval false:attach failed
	 */
	bool Attach(
		ID id ///< [in] Semaphore ID
	);
#endif

	/**
	 * @brief  Delete mutex or Free static mutex id
	 * @param  (none)
	 * @return Delete success or not
	 * @retval true:Delete succeeded
	 * @retval false:Delete failed
	 */
	bool Delete();

	/**
	 * @brief  Take semaphore. If can not take, wait until get take
	 * @param  (none)
	 * @return Take success or not
	 * @retval true:Take succeeded
	 * @retval false:Take failed
	 */
	bool Take();

	/**
	 * @brief  Asynchronized take. If can not take, return false
	 * @param  (none)
	 * @return Take success or not
	 * @retval true:Take succeeded
	 * @retval false:Take failed
	 */
	bool AsyncTake();

	/**
	 * @brief  Timed take. If can not take, wait until time is up
	 * @return Take success or not
	 * @retval true:Take succeeded
	 * @retval false:Take failed
	 */
	bool TimedTake(
		int time ///< [in] timeout time
	);

	/**
	 * @brief  Release semaphore. If can not release, return false
	 * @param  (none)
	 * @return Release success or not
	 * @retval true:Release succeeded
	 * @retval false:Release failed
	 */
	bool Release();

	/**
	 * @brief  Get error code. 
	 * @param  (none)
	 * @return Mutex error code
	 * @retval MutexError
	 */
	SEM_ERROR GetLastError();
}; // class PafeeSemaphore

#ifdef __cplusplus
}; // namespace pafee
#endif

#endif // PAFEE_OS_WINDOWS || PAFEE_OS_LINUX || PAFEE_OS_ITRON

#endif // _PAFEESEMAPHORE_H_INCLUDED_
