//********************************************************************
//		project 	PAFEE
//
//! 	 @file: 	PafeeTask.h
//! 	 @author:	Technocrats
//! 	 @brief:	PafeeTask class header
//
//********************************************************************
//
// P A F E E   - Pragmatic Application Framework for Embedded Environment- 
//
//
// Copyright (c) 2009 KN Information Sustems Corporation,
//                    Hirocon Corporation,
//                    Technocrats Corporation
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//

#ifndef _PAFEETASK_H_INCLUDED_
#define _PAFEETASK_H_INCLUDED_

// if no OS definition, PafeeTask class is unavailable
#if defined(PAFEE_OS_WINDOWS)||defined(PAFEE_OS_LINUX)||defined(PAFEE_OS_ITRON)

#ifdef PAFEE_OS_WINDOWS
#include <windows.h>
#endif

#ifdef PAFEE_OS_LINUX
#include <pthread.h>
#include <sched.h>
//#include <sys/time.h>
#endif

#ifdef PAFEE_OS_ITRON
#include <kernel.h> 
#include <itron.h>
#define MAX_TASK 256  // temporary value
#endif

#ifdef PAFEE_DEBUG
#include <stdio.h>
#endif

#ifdef __cplusplus
namespace Pafee
{
#endif

// Task ID typedef
#ifdef PAFEE_OS_WINDOWS
	typedef HANDLE TASK_ID;
#endif
#ifdef PAFEE_OS_LINUX
	typedef pthread_t TASK_ID;
#endif
#ifdef PAFEE_OS_ITRON
	typedef ID TASK_ID;
#endif

#ifdef PAFEE_DEBUG
#if defined(PAFEE_OS_WINDOWS)||defined(PAFEE_OS_LINUX)
#define TASK_DEBUG(STRING)   printf STRING
#elif defined(PAFEE_OS_ITRON)
void TaskSyslog(const char *format, ...);
#define TASK_DEBUG(STRING)   TaskSyslog STRING
#endif
#else
#define TASK_DEBUG(STRING)
#endif

class PafeeTask
{

public:
	/**
	 * @brief  Task error enum
	 */
	typedef enum tagTASK_ERROR
	{
		ERROR_NONE, /**< Error none */
		ERROR_NOHANDLE, /**< Function is called without handle */
		ERROR_HAS_HANDLE, /**< Allready has handle */
		ERROR_CANNOT_CREATE, /**< Can not create */
		ERROR_CANNOT_DELETE, /**< Can not delete */
		ERROR_CANNOT_START, /**< Can not start */
		ERROR_STILL_SUSPENDED, /**< Stopped more times than Started */
		ERROR_TIME_OUT, /**< Can not get because of time out */
		ERROR_CANNOT_STOP, /**< Can not stop */
		ERROR_CANNOT_RELEASE, /**< Can not release */
		ERROR_MAX /**< MAX */
	} TASK_ERROR;

	// Task priority typedef
	// - Windows : use defined value :
	//             THREAD_PRIORITY_ABOVE_NORMAL
	//             THREAD_PRIORITY_BELOW_NORMAL
	//             THREAD_PRIORITY_HIGHEST
	//             THREAD_PRIORITY_LOWEST
	//             THREAD_PRIORITY_IDLE
	//             THREAD_PRIORITY_NORMAL
	//             THREAD_PRIORITY_TIME_CRITICAL
	// - Linux : use int value (1 to 99)
	// - ITRON : not used
	typedef int TASK_PRIORITY;

	// Task scheduling policy typedef
	// - for Linux : use defined value :
	//   SCHED_OTHER, SCHED_RR, SCHED_FIFO
	typedef int TASK_SCHEDULING;

	typedef struct tagTHREAD_PARAM
	{
		PafeeTask* instance;
		void* pParam;
	} THREAD_PARAM;

private:
	TASK_ID m_handle;
	TASK_ERROR m_error;
	bool m_isCreated;
	THREAD_PARAM m_threadParam;
#if 0
//#ifdef PAFEE_WINDOWS_VC6
	HMODULE m_libraryHandle;
#endif

public:
	/**
	 * @brief  PafeeTask class constructor
	 * @param  (none)
	 * @return (none)
	 */
	PafeeTask();

	/**
	 * @brief  PafeeTask class destructor
	 * @param  (none)
	 * @return (none)
	 */
	virtual ~PafeeTask();

#if defined(PAFEE_OS_WINDOWS) || defined(PAFEE_OS_LINUX)
	/**
	 * @brief  Create new task
	 * @param  TASK_ID* : Task ID
	 * @param  TASK_PRIORITY : Task priority
	 * @param  TASK_SCHDUKLING : Task scheduling
	 * @param  int : Stack size
	 * @param  void* : argument of TaskProcedure
	 * @param  bool : Start task or not on creating
	 * @return Create success or not
	 * @retval true:create succeeded
	 * @retval false:create failed
	 */
	bool Create(
		TASK_ID* pTaskID,
		TASK_PRIORITY priority,
		TASK_SCHEDULING schedule = 0, // 0 : SCHED_OTHER(Linux)
		int stackSize = 0,
		void* pParam = NULL,
		bool isStartImmediately = true
	);
#endif
#ifdef PAFEE_OS_ITRON
	/**
	 * @brief  Attach static task id
	 * @return Attach success or not
	 * @retval true:attach succeeded
	 * @retval false:attach failed
	 */
	bool Attach(
		TASK_ID id ///< [in] Task ID
	);
#endif

#if defined(PAFEE_OS_WINDOWS) || defined(PAFEE_OS_ITRON)
	/**
	 * @brief  Start task
	 * @param  (none)
	 * @return Start success or not
	 * @retval true:Start succeeded
	 * @retval false:Start failed
	 */
	bool Start();

	/**
	 * @brief  Stop task
	 * @param  (none)
	 * @return Stop success or not
	 * @retval true:Stop succeeded
	 * @retval false:Stop failed
	 */
	bool Stop();
#endif

#if defined(PAFEE_OS_WINDOWS) || defined(PAFEE_OS_LINUX)
	/**
	 * @brief  Delete task
	 * @param  (none)
	 * @return Delete success or not
	 * @retval true:Delete succeeded
	 * @retval false:Delete failed
	 */
	bool Delete();
#endif

	/**
	 * @brief  Task procedure function
	 * @param  parameter
	 * @return return value
	 */
	virtual void* TaskProcedure(
		void* pParam
	)=0;

	/**
	 * @brief  Get error code. 
	 * @param  (none)
	 * @return Task error code
	 * @retval TaskError
	 */
	TASK_ERROR GetLastError();

private:
	/**
	 * @brief  static function to use function pointer
	 * @param  void* : Task procedure information structure
	 * @return void pointer
	 */
#ifdef PAFEE_OS_WINDOWS
	static ULONG WINAPI TaskLauncher(LPVOID arg)
	{
		THREAD_PARAM* pointer;
		pointer = reinterpret_cast<THREAD_PARAM*>(arg);
		pointer->instance->TaskProcedure(pointer->pParam);
		return 0;
	}
#endif
#ifdef PAFEE_OS_LINUX
	static void* TaskLauncher(void* arg)
	{
		THREAD_PARAM* pointer;
		pointer = (THREAD_PARAM*)arg;
		pointer->instance->TaskProcedure(pointer->pParam);
		return 0;
	}
#endif

}; // class PafeeTask

#ifdef __cplusplus
}; // namespace pafee
#endif

#endif // PAFEE_OS_WINDOWS || PAFEE_OS_LINUX || PAFEE_OS_ITRON

#endif // _PAFEETASK_H_INCLUDED_
