//****************************************************
//
//  project     PAFEE
//! @file:      PafeeString.cpp
//! @author:    Xie Jian
//! @brief:     string
//
//*****************************************************

//
// P A F E E   - Pragmatic Application Framework for Embedded Environment- 
//
//
// Copyright (c) 2009 KN Information Systems Corporation,
//                    Hirocon Corporation,
//                    Technocrats Corporation
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//

#ifdef ENV_WIN32
#pragma warning(disable:4996)
#endif
#include "PafeeString.h"

namespace Pafee
{

//----------------------------------------
//! default constructor
//----------------------------------------
String::String(
	Allocator *pAllocator ///< [in] memory allocator
)
{
	m_localBuffer[0] = '\0';
	m_pData = m_localBuffer;
	m_length = 0;
	m_allocateSize = LOCAL_BUFFER_SIZE;
	SetAllocator(pAllocator);
}

//----------------------------------------
//! copy constructor
//----------------------------------------
String::String(
	const String& src, ///< [in] initial value of string
	Allocator* pAllocator ///< [in] memory allocator
)
{
	SetAllocator(pAllocator);
	m_allocateSize = src.m_allocateSize;
	m_length = src.m_length;
	if(m_allocateSize == LOCAL_BUFFER_SIZE)
	{
		m_pData = m_localBuffer;
	}
	else
	{
		m_pData = (char*)Allocate(m_allocateSize);
	}
	pafee_memcpy(m_pData,src.m_pData,m_length);
	m_pData[m_length] = '\0';
}

//----------------------------------------
//! constructor
//----------------------------------------
String::String(
	const char* szSrc, ///< [in] initial value of string
	Allocator* pAllocator  ///< [in] memory allocator
)
{
	SetAllocator(pAllocator);
	m_length = pafee_strlen(szSrc);
	int allocateSize = AligmentSize(m_length + 1);
	if(allocateSize <= LOCAL_BUFFER_SIZE)
	{
		m_allocateSize = LOCAL_BUFFER_SIZE;
		m_pData = m_localBuffer;
	}
	else
	{
		m_allocateSize = allocateSize;
		m_pData = (char*)Allocate(m_allocateSize);
	}
	pafee_memcpy(m_pData, szSrc, m_length);
	m_pData[m_length] = '\0';
}

//----------------------------------------
//! destructor
//----------------------------------------
String::~String()
{
	if(m_allocateSize > LOCAL_BUFFER_SIZE)
	{
		Deallocate(m_pData);
		m_pData = NULL;
	}
}

//----------------------------------------
//! Initialize
//! @return void
//----------------------------------------
void String::Initialize(
	Allocator* pAllocator  ///< [in] memory allocator
)
{
	m_localBuffer[0] = '\0';
	m_pData = m_localBuffer;
	m_length = 0;
	m_allocateSize = LOCAL_BUFFER_SIZE;
	SetAllocator(pAllocator);
}

//-------------------------------------------------------------------------------------
//! operator =  characters
//! @return Result(the address of string)
//-------------------------------------------------------------------------------------
String& String::operator =(
	const char* szSrc  ///< [in] the characters for new data
)
{
	m_length = pafee_strlen(szSrc);
	int allocateSize = AligmentSize(m_length + 1);
	if(allocateSize > m_allocateSize)
	{
		if(m_allocateSize > LOCAL_BUFFER_SIZE)//(m_pData != m_localBuffer)
		{
			Deallocate(m_pData);
		}
		m_allocateSize = allocateSize;
		m_pData = (char*)Allocate(allocateSize);
	}
	pafee_memcpy(m_pData, szSrc, m_length);
	m_pData[m_length] = '\0';
	return *this;
}

//-------------------------------------------------------------------------------------
//! operator =. character
//! @return Result(the address of string)
//-------------------------------------------------------------------------------------
String& String::operator = (
	char src ///< [in] the character for new data
)
{
	m_length = 1;
	*m_pData = src;
	m_pData[m_length] = '\0';
	return *this;
}

//-------------------------------------------------------------------------------------
//! operator +=. characters
//! @return Result(the address of string)
//-------------------------------------------------------------------------------------
String& String::operator += (
	const char* szSrc ///< [in] the characters for new data
)
{
	int length = pafee_strlen(szSrc);
	int allocateSize = AligmentSize(m_length + length + 1);	
	if(allocateSize > m_allocateSize)
	{
		char* newData = (char*)Allocate(allocateSize);
		pafee_memcpy(newData,m_pData,m_length);
		pafee_memcpy((newData + m_length),szSrc,length);
		if(m_allocateSize > LOCAL_BUFFER_SIZE)//(m_pData != m_localBuffer)
		{
			Deallocate(m_pData);
		}
		m_allocateSize = allocateSize;
		m_pData = newData;
	}
	else
	{
		pafee_memcpy((m_pData + m_length),szSrc,length);
	}
	m_length += length;
	m_pData[m_length] = '\0';
	return *this;
}

//-------------------------------------------------------------------------------
//! Find.Find the characters that equals "szTarget" from a string
//! @return Result(The index of the begin position at string:Success  -1:Failure)
//-------------------------------------------------------------------------------
int String::Find(
	const char* szTarget, ///< [in] the characters that is wanted to be find 
	int from              ///< [in] begin position(for find)
)
{
	char* p= pafee_strstr((m_pData + from), szTarget);
	if(p != NULL)
	{
		return p - m_pData; 
	}
	return -1;
}

//----------------------------------------------------------------------------------------------
//! ReverseFind.Find the characters that equals "szTarget" from the last position of a string
//! @return Result(The index of the begin position at string:Success  -1:Failure)
//----------------------------------------------------------------------------------------------
int String::ReverseFind(
	const char* szTarget ///< [in] the characters that is wanted to be find
)
{
	int flag = -1;
	int index = 0;
	int length = pafee_strlen(szTarget);
	for(int i = 0; i < (m_length - length + 1); i++)
	{
		index = Find(szTarget,index);
		if(index == -1)
		{
			break;
		}
		flag = index;
		index++;
	}
	return flag;
}

//--------------------------------------------------------------
//! Private function : Allocate memory for data by Formated 
//! @return Result(void)
//--------------------------------------------------------------
void String::GetAllocateSize(
	int count ///< [in] Allocate Size
)
{
	int allocateSize = AligmentSize(count + 1);	
	if(allocateSize > m_allocateSize)
	{
		if(m_allocateSize > LOCAL_BUFFER_SIZE)
		{
			Deallocate(m_pData);
		}
		m_allocateSize = allocateSize;
		m_pData = (char*)Allocate(allocateSize);
	}
}

//--------------------------------------------------------------
//! Private function : Get the amount of length of all the data 
//! @return Result(The amount of length of all the data)
//--------------------------------------------------------------
int String::GetNewLen(
	const char *format, ///< [in] format
	va_list ap ///< [in] Argument list
)
{
	int count;
	char buffer[20];
	int length = 0;
	char* p;
	char* pfind;
	char* ptemp;
	int min;
	int lenAtoi;
	char* start;
	//
	int size = strlen(format);
	char* workBuffer = (char*)Allocate(size+1);
	memcpy(workBuffer,format,size);
	workBuffer[size] = '\0';
	p = workBuffer;
	while(*p != '\0')
	{
		pfind = pafee_strchr(p, '%');
		if(pfind == NULL)
		{
			break;
		}
		p = pfind;
		*p = '\0';
		p++;
	}
	start= workBuffer;
	p = workBuffer;
	p++;
	while((p-start) < size)
	{
		if(p == NULL)
		{
			break;
		}
		min = 0;
		pfind = p;
		count = 0;
		ptemp = pafee_strchr(p, 'c');
		if(ptemp != NULL)
		{
			min = ptemp - p;
			pfind = ptemp;
			count = 1;
			va_arg(ap,int);
		}
		ptemp = pafee_strchr(p, 'd');
		if(ptemp != NULL)
		{
			min = ptemp - p;
			pfind = ptemp;
			count = pafee_sprintf(buffer,"%d",va_arg(ap,int));
		}
		ptemp = pafee_strchr(p, 'p');
		if(ptemp != NULL)
		{
			min = ptemp - p;
			pfind = ptemp;
			count = 8;
			va_arg(ap,char*);
		}
		ptemp = pafee_strchr(p, 's');
		if(ptemp != NULL)
		{
			min = ptemp - p;
			pfind = ptemp;
			count = pafee_strlen(va_arg(ap,char*));
		}
		ptemp = pafee_strchr(p, 'x');
		if(ptemp != NULL)
		{
			min = ptemp - p;
			pfind = ptemp;
			count = pafee_sprintf(buffer,"%x",va_arg(ap,int));
		}
		ptemp = pafee_strchr(p, 'X');
		if(ptemp != NULL)
		{
			min = ptemp - p;
			pfind = ptemp;
			count = pafee_sprintf(buffer,"%X",va_arg(ap,int));
		}
		if(min > 0)
		{
			ptemp = pfind;
			*ptemp = '\0';
			
			ptemp = p;
			lenAtoi = pafee_atoi(ptemp);
			if(lenAtoi > count)
			{
				count = lenAtoi;
			}
		}
		length += count;
		p = pfind;
		p++;
		p++;
	}
	Deallocate((char*)workBuffer);
	return length;
}

//--------------------------------------------------------------
//! Format: Make the data by a specified format. 
//! @return Result(void)
//--------------------------------------------------------------
void String::Format(
	const char* format,   ///< [in] format
	...                   ///< [in] datas
)
{
	va_list ap;
    va_start(ap, format);
	m_length = this->GetNewLen(format,ap);
	this->GetAllocateSize(m_length);//for m_pData
	vsprintf(m_pData,format,ap);
	va_end(ap);
}

//-------------------------------------------------------------------------------------
//! Insert. Insert characters(szSrc) into a string from index(the position)
//! @return Result(void)
//-------------------------------------------------------------------------------------
void String::Insert(
	int index,         ///< [in] begin position
	const char* szSrc  ///< [in] data buffer for insert
)
{
	int length = pafee_strlen(szSrc);
	if((index < 0) || (index > m_length))
	{
		return;
	}
	int allocateSize = AligmentSize(m_length + length + 1);	
	if(allocateSize > m_allocateSize)
	{
		char *newData = (char*)Allocate(allocateSize);
		pafee_memcpy(newData,m_pData,index);//index
		if(m_length > index)
		{	
			pafee_memcpy((newData + index + length),(m_pData + index),(m_length - index));
		}
		if(m_allocateSize > LOCAL_BUFFER_SIZE)
		{
			Deallocate(m_pData);
		}
		m_allocateSize = allocateSize;
		m_pData = newData;
	}
	else
	{
		if(m_length > index)
		{	
			pafee_memmove((m_pData + index + length),(m_pData + index),(m_length - index));
		}
	}
	pafee_memcpy((m_pData + index),szSrc,length);
	m_length += length;
	m_pData[m_length] = '\0';
}

//-------------------------------------------------------------------------------------
//! Delete. Delete characters (amount of count) from right
//! @return Result(void)
//-------------------------------------------------------------------------------------
void String::DeleteRight(
	int count ///< [in] Delete size
)
{
	int index = m_length - count;
	if(index < 0)
	{
		index = 0;
	}
	Delete(index, count);
}

//-------------------------------------------------------------------------------------
//! Delete. Delete characters (amount of count) from index(the position)
//! @return Result(void)
//-------------------------------------------------------------------------------------
void String::Delete(
	int index, ///< [in] begin position
	int count  ///< [in] the length you want to delete
)
{
	int length = m_length;
	if((index < 0) || (index >= length))
	{
		return;
	}
	if((index + count) > length)
	{
		count = length - index;
	}
	m_length = length - count;
	pafee_memcpy((m_pData + index),(m_pData + index + count),(length - index - count));
	m_pData[m_length] = '\0';
}

//-------------------------------------------------------------------------------------
//! Substring. Get substring (amount of "count") from "from"(the position)
//! @return Result(void)
//-------------------------------------------------------------------------------------
String String::Substring(
	int from,  ///< [in] begin position
	int count  ///< [in] the length of substring you want to get
)
{
	String s;
	int length = m_length;
	if((from < 0) || (from >= length) || (count < 0))
	{
		return s;
	}
	if((from + count) > m_length)
	{
		count = m_length - from;
	}
	s.m_length = count;
	int allocateSize = AligmentSize(s.m_length + 1);
	if(allocateSize > LOCAL_BUFFER_SIZE)
	{
		s.m_allocateSize = allocateSize;
		s.m_pData = (char*)Allocate(allocateSize);
	}
	pafee_memcpy(s.m_pData,(m_pData + from), count);
	s.m_pData[s.m_length] = '\0';
	return s;
}

//-------------------------------------------------------------------------------------
//! Right. Get substring (amount of "count") from "Right"(the position)
//! @return Result(void)
//-------------------------------------------------------------------------------------
String String::Right(
	int count ///< [in] the length of substring you want to get 
)
{
	int index = m_length - count;
	if(index < 0)
	{
		index = 0;
	}
	return Substring(index,count);
}

//---------------------------------------------------------------------
//! Trim.Delete the character that equals "target" from left
//! @return Result(the address of this string)
//---------------------------------------------------------------------
String& String::TrimLeft(
	char target ///< [in] the characters that is wanted to be trim
)
{
	char *p=m_pData;
	int i;
	for(i = 0; i < m_length; i++)
	{
		if(*p != target)
		{
			break;
		}
		p++;
	}
	pafee_memcpy(m_pData, p, (m_length - i));
	m_length = m_length - i;
	m_pData[m_length] = '\0';
	return *this;
}

//---------------------------------------------------------------------
//! Trim.Delete the character that equals "target" from right
//! @return Result(the address of this string)
//---------------------------------------------------------------------
String& String::TrimRight(
	char target ///< [in] the characters that is wanted to be trim
)
{
	char *p = m_pData + m_length - 1;
	for (;m_length > 0; m_length--)
	{
		if(*p != target )
		{
			break;
		}
		p--;
	}
	m_pData[m_length] = '\0';
	return *this;
}

//-----------------------------------------------------------------------
//! Trim.Delete the characters that equals "szTarget" from left 
//! @return Result(the address of this string)
//-----------------------------------------------------------------------
String& String::TrimLeft(
	const char* szTarget ///< [in] the characters that is wanted to be trim
)
{
	char *p = m_pData;
	int i;
	int targetLength = pafee_strlen(szTarget);
	int hitCount = 0;
	for(i = 0; i < m_length; i++)
	{
		hitCount = 0;
		for(int j = 0; j < targetLength; j++)
		{
			if(*p != szTarget[j])
			{
				hitCount++;
			}
		}
		if ( hitCount >= targetLength )
		{
				break;
		}
		p++;
	}
	pafee_memcpy(m_pData, p, (m_length - i));	
	m_length = m_length - i;
	m_pData[m_length] = '\0';
	return *this;
}

//-----------------------------------------------------------------------
//! Trim.Delete the characters that equals "szTarget" from right
//! @return Result(the address of this string)
//-----------------------------------------------------------------------
String& String::TrimRight(
	const char* szTarget ///< [in] the characters that is wanted to be trim
)
{
	char *p = m_pData + m_length - 1;
	int targetLength = pafee_strlen(szTarget);
	int hitCount = 0;
	for (;m_length > 0; m_length--)
	{
		hitCount = 0;
		for(int j = 0; j < targetLength; j++)
		{
			if(*p != szTarget[j])
			{
				hitCount++;
			}
		}
		if ( hitCount >= targetLength )
		{
			break;
		}
		p--;
	}
	m_pData[m_length] = '\0';
	return *this;
}

//--------------------------------------------------------------
//! ToLower:made all the characters to lower characters
//! @return Result(the address of this string)
//--------------------------------------------------------------
String& String::ToLower()
{
	for(char* p = m_pData; *p != '\0'; p++)
	{
		if ( *p >= 0x41 && *p <= 0x5a )
		{
			*p+=32;
		}
	}
	return *this;
}

//--------------------------------------------------------------
//! ToUpper:made all the characters to upper characters
//! @return Result(the address of this string)
//--------------------------------------------------------------
String& String::ToUpper()
{
	for(char* p = m_pData; *p != '\0'; p++)
	{
		if ( *p >= 0x61 && *p <= 0x7a )
		{
			*p-=32;
		}
	}
	return *this;
}

#ifdef PAFEE_UNITTEST
//----------------------------------------------------------------------------
//!	Unit test
//! @return: void
//----------------------------------------------------------------------------
void String::UnitTest(void)
{
	String str1("a");
	String str2("a");
	char* pChar = "a";
	char c = 'a';
	//test GetBuffer
	assert(str1.GetBuffer() == str1);
	//test operator/Compare/CompareNoCase/ToLower/ToUpper
	{
		//(operator ==)/Compare/CompareNoCase
		assert(str1 == str2);
		assert(str1 == pChar);
		assert(str1 == c);
		assert(Compare(str1,str2) == 0);
		assert(Compare(str1,pChar) == 0);
		assert(Compare(str1,c) == 0);
		assert(CompareNoCase(str1,str2) == 0);
		//(operator =)/(operator !=)
		str1.ToUpper();//"A"
		assert(str1 != str2);
		assert(str1 != pChar);
		assert(str1 != c);
		assert(Compare(str1,str2) != 0);
		assert(Compare(str1,pChar) != 0);
		assert(Compare(str1,c) != 0);
		assert(CompareNoCase(str1,str2) == 0);
		//(operator +)
		str1 = "AA";
		str1.ToLower();//"aa"
		str2 = str2 + "A";
		assert(str1 != str2);
		assert(Compare(str1,str2) != 0);
		assert(CompareNoCase(str1,str2) == 0);
		str2 = pChar + str2;
		assert(str2 == "aaA");
		str2 = c + str2;
		assert(str2 == "aaaA");
		str2 = str2 + c;
		assert(str2 == "aaaAa");
		str2 = str1 + str2;
		assert(str2 == "aaaaaAa");
		//(operator +=)
		str1 = "aa";
		str2 = "A";
		str2 += "a";
		assert(str1 != str2);
		assert(Compare(str1,str2) != 0);
		assert(CompareNoCase(str1,str2) == 0);
		//(operator <)/(operator <=)/(operator >)/(operator >=)
		str1 = "a";//97
		str2 = "A";//65
		pChar = "aa";
		c = 'b';
		assert((str1 < str2) == 0);
		assert((str1 <= str2) == 0);
		assert((str1 < pChar) == 1);
		assert((str1 <= pChar) == 1);
		assert((str1 < c) == 1);
		assert((str1 <= c) == 1);
		assert((str1 > str2) == 1);
		assert((str1 >= str2) == 1);
		assert((str1 > pChar) == 0);
		assert((str1 >= pChar) == 0);
		assert((str1 > c) == 0);
		assert((str1 >= c) == 0);
		//operator [] 
		{
			//
		}
	}
	//test Find/ReverseFind
	{
		str1 = "AaBcj";
		str2 = "aB";
		pChar = "j";
		c = 'x';
		assert(str1.Find(str2) == 1);
		assert(str1.Find(pChar) == 4);
		assert(str1.Find(c) == -1);
		//ReverseFind()
		str1 = "01234a6a89";//str1.m_length = 10
		str2 = "6a89";		
		pChar = "a";
		c = 'x';
		assert(str1.ReverseFind(str2) == 6);
		assert(str1.ReverseFind(pChar) == 7);
		assert(str1.ReverseFind(c) == -1);
	}
	//test IsEmpty/Insert/Delete/DeleteLeft/DeleteRight/GetLength/GetLeft/GetRight/GetSubstring
	//GetLeft/GetRight/GetSubstring-->//Left/Right/SubString
	{
		str1 = "";
		str2 = "aa";
		pChar = "b";
		c = 'c';
		assert(str1.IsEmpty());
		str1.Insert(0,str2);//aa
		assert(str1.Left(1) == str1.Right(1));
		str1.Insert(0,pChar);//baa
		assert(str1.Substring(0,1) == pChar);
		assert(str1.Left(1) == pChar);
		assert(str1.Right(2) == str2);
		assert(str1.GetLength() == 3);
		assert(str1.Left(str1.GetLength()) == str1.Substring(0));
		str1.Insert(2,c);//baca
		assert(str1.Left(1) == pChar);
		assert(str1.Right(2) == str1.Substring(2,2));
		assert(str1.Left(4) == str1.Right(str1.GetLength()));
		str1 = "0123456789";
		str1.Delete(10,1);
		assert(str1 == "0123456789");//str1 = "";?
		str1.Delete(4);
		assert(str1 == "012356789");
		str1.DeleteLeft(2);
		assert(str1 == "2356789");
		str1.DeleteRight(5);
		assert(str1 == "23");
	}
	//test Trim/TrimLeft/TrimRight
	{
		str1 = "1ab11ab5611ab";
		str1.TrimLeft('a');
		assert(str1 == "1ab11ab5611ab");
		str1.TrimLeft('1');
		assert(str1 == "ab11ab5611ab");
		str1.TrimRight('b');
		assert(str1 == "ab11ab5611a");
		str1.Trim('a');
		assert(str1 == "b11ab5611");
		str1.Trim('b');
		assert(str1 == "11ab5611");
		str1 = "b11ab5611";
		str1.Trim('1');
		assert(str1 == "b11ab56");
		str1 = "1ab11ab5611ab";
		str1.TrimLeft("a");
		assert(str1 == "1ab11ab5611ab");
		str1.TrimLeft("1");
		assert(str1 == "ab11ab5611ab");
		str1.TrimRight("b");
		assert(str1 == "ab11ab5611a");
		str1.Trim("a");
		assert(str1 == "b11ab5611");
		str1 = "b11ab5611";
		str1.Trim("1");
		assert(str1 == "b11ab56");
		str1 = "b11ab5611";
		str1.Trim("b");
		assert(str1 == "11ab5611");
		str1 = "11ab5611";
		str1.Trim("11a");
		assert(str1 == "b56");
		str1 = "11ab5611";
		str1.Trim("5611");
		assert(str1 == "ab");
		str1 = "11ab5611";
		str1.Trim("xxx");
		assert(str1 == "11ab5611");
		str1 = "11ab5611";
		str1.Trim("11");
		assert(str1 == "ab56");
		str1 = "   0	";
		str1.Trim();
		assert(str1.GetLength() == 1);
	}
}
#endif //PAFEE_UNITTEST

}//namespace Pafee

