//****************************************************
//
//      project     PAFEE
//!      @file:     PafeeTokenizer.cpp
//!      @author:   M.Mori
//!      @brief:    Tokenizer class
//
//*****************************************************

//
// P A F E E   - Pragmatic Application Framework for Embedded Environment- 
//
//
// Copyright (c) 2009 KN Information Systems Corporation,
//                    Hirocon Corporation,
//                    Technocrats Corporation
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//

#ifdef ENV_WIN32
#pragma warning(disable:4996)
#endif
#include "PafeeTokenizer.h"
#include "pafee_string.h"
namespace Pafee
{

//----------------------------------------------------------------------------
//!  Constructor
//----------------------------------------------------------------------------
Tokenizer::Tokenizer(
	Allocator* pAllocator ///< [in] Memory Allocator
)
{
	SetAllocator(pAllocator);
	m_pData = NULL;
	m_ppData = NULL;
	m_indexSize =0;
}

//----------------------------------------------------------------------------
//!  Constructor
//----------------------------------------------------------------------------
Tokenizer::Tokenizer(
	const char* szTarget, ///< [in] Character string to handle
	char token,           ///< [in] Delimiter
	char quote,           ///< [in] Quote
	char* szEndMark ,     ///< [in] The last sign
	Allocator* pAllocator ///< [in] Memory Allocator
)
{
	SetAllocator(pAllocator);
	m_pData = NULL;
	m_ppData = NULL;
	m_indexSize =0;
	Split(szTarget, token, quote, szEndMark);
}

//----------------------------------------------------------------------------
//!  Destructor
//----------------------------------------------------------------------------
Tokenizer::~Tokenizer()
{
	if ( m_ppData != NULL)
	{
		Deallocate(m_ppData);
	}
	if ( m_pData != NULL)
	{
		Deallocate(m_pData);
	}
}

//----------------------------------------------------------------------------
//!  Split 
//!  @return: Result (Split Length)
//----------------------------------------------------------------------------
int Tokenizer::Split(
	const char* szTarget, ///< [in] Character string to handle
	char token,           ///< [in] Delimiter
	char quote,           ///< [in] Quote
	char* szEndMark       ///< [in] The last sign
)
{
	int length;
	int allocateSize;
	int rtn;
	char *p;
	char* pEndMark= TokenFind(szTarget, szEndMark, quote);
	if ( pEndMark == NULL )
	{
		length = pafee_strlen(szTarget);
		rtn = length;
	}
	else
	{
		length = pEndMark-szTarget;
		rtn = length + pafee_strlen(szEndMark);
	}
	int IndexSize=0;
	if ( length <= PAFEE_STATIC_MAX_BYTE-1 )
	{
		m_length = length;
		pafee_strncpy ( m_szData, szTarget, m_length  );
		m_szData[m_length] = '\0';
		m_pStart=m_szData;
		m_pData = NULL;
		m_allocateSize =0;
	}
	else
	{
		allocateSize = AligmentSize(length+1);
		if ( m_pData != NULL )
		{
			if ( allocateSize > m_allocateSize )
			{
				Deallocate(m_pData);
			}
		}
		m_allocateSize = allocateSize;
		m_pData = (char*)Allocate(m_allocateSize);
		m_length = length;
		pafee_strncpy ( m_pData, szTarget, m_length  );
		m_pData[m_length]='\0';
		m_pStart=m_pData;
	}
	char *pStart = m_pStart;
	while(1)
	{
		p = TokenFind(pStart, token, quote);
		if ( p == NULL )
		{
			IndexSize++;
			break;
		}
		pStart = p + 1;
		IndexSize++;
	}
	if ( m_ppData != NULL )
	{
		if ( IndexSize > m_indexSize )
		{
			Deallocate(m_pData);
		}
	}
	m_indexSize = IndexSize;
	m_ppData = (char**)Allocate(m_indexSize*sizeof(char*));
	pStart = m_pStart;
	char** pp = m_ppData;
	for (int i=0;i<m_indexSize;i++)
	{
		p = TokenFind(pStart, token, quote);
		if ( p == NULL )
		{
			*pp = pStart;
			pp++;
			break;
		}
		*p = '\0';
		*pp = pStart;
		pStart = p+1;
		pp++;
	}
	return rtn;
}

//----------------------------------------------------------------------------
//!  Get
//!  @return: Result (true:Success false:failure)
//----------------------------------------------------------------------------
bool Tokenizer::Get(
	int index,      ///< [in] index
	char* szBuffer, ///< [in] Buffer
	int bufferSize  ///< [in] bufferSize
)
{
	if ( index > m_indexSize )
	{
		return false;
	}
	else
	{
		pafee_strncpy ( szBuffer, m_ppData[index], bufferSize-1 );
		return true;
	}
}

//----------------------------------------------------------------------------
//!  Get
//!  @return: Result (DataBuffer:Success NULL:failure)
//----------------------------------------------------------------------------
char* Tokenizer::Get(
	int index      ///< [in] index
)
{
	if ( index > m_indexSize )
	{
		return NULL;
	}
	else
	{
		return m_ppData[index];
	}
}

//----------------------------------------------------------------------------
//!  GetLength 
//!  @return: Result (Length:Success 0:failure)
//----------------------------------------------------------------------------
int Tokenizer::GetLength(
	int index ///< [in] index
)
{
	if ( index > m_indexSize )
	{
		return 0;
	}
	else
	{
		if ( m_ppData == NULL )
		{
			return 0;
		}
		return pafee_strlen(m_ppData[index]);
	}
}

//----------------------------------------------------------------------------
//!  TokenFind 
//!  @return: Result (Find Position:Success NULL:failure)
//----------------------------------------------------------------------------
char* Tokenizer::TokenFind(
	const char* str, ///< [in] str
	char* findStr,   ///< [in] Character string to find out
	char quote       ///< [in] quote
)
{
	char* pFind;
	char* pQuoteStart;
	char* pQuoteEnd;
	int i=0;
	while(1)
	{
		pFind = pafee_strstr((char*)&str[i], findStr);
		pQuoteStart = pafee_strchr((char*)&str[i], quote);
		if ( pQuoteStart != NULL )
		{
			pQuoteEnd = pafee_strchr(pQuoteStart+1, quote);
			if ( pQuoteEnd != NULL )
			{
				if ( (pFind > pQuoteStart) && (pFind < pQuoteEnd) )
				{
					pFind = pafee_strstr(pQuoteEnd+1, findStr);
					if ( pFind != NULL )
					{
						return pFind;
					}
				}
				return pFind;
			}
			char* pFind2 = pafee_strstr(pQuoteStart+1, findStr);
			if ( pFind2 != NULL )
			{
				return pFind2;
			}
			i = pQuoteEnd + pafee_strlen(findStr) - &str[0];
		}
		return pFind;
	}
	return NULL;
}

//----------------------------------------------------------------------------
//!  TokenFind 
//!  @return: Result (Find Position:Success NULL:failure)
//----------------------------------------------------------------------------
char* Tokenizer::TokenFind(
	const char* str, ///< [in] str
	char findChr,    ///< [in] Character to find out
	char quote       ///< [in] quote
)
{
	char* pFind;
	char* pQuoteStart;
	char* pQuoteEnd;
	int i=0;
	while(1)
	{
		pFind = pafee_strchr((char*)&str[i], findChr);
		pQuoteStart = pafee_strchr((char*)&str[i], quote);
		if ( pQuoteStart != NULL )
		{
			pQuoteEnd = pafee_strchr(pQuoteStart+1, quote);
			if ( pQuoteEnd != NULL )
			{
				if ( (pFind > pQuoteStart) && (pFind < pQuoteEnd) )
				{
					pFind = pafee_strchr(pQuoteEnd+1, findChr);
					if ( pFind != NULL )
					{
						return pFind;
					}
				}
				return pFind;
			}
			char* pFind2 = pafee_strchr(pQuoteStart+1, findChr);
			if ( pFind2 != NULL )
			{
				return pFind2;
			}
			i = pQuoteEnd + 1 - &str[0];
		}
		return pFind;
	}
	return NULL;
}

#ifdef PAFEE_UNITTEST
//----------------------------------------------------------------------------
//! Unit test
//! @return: void
//----------------------------------------------------------------------------
void Tokenizer::UnitTest(void)
{
	struct _strData
	{
		char name[16];
		char Data[64];
	} strData[4]
	=
	{
		{"ipad1","192.168.1.1"},
		{"user1","\"Oba======ma\""},
		{"ipad2","192.168.1.3"},
		{"user2","\"Amer=====ica\""}
	};
	char buf[128];
	pafee_memset ( buf, '\0', sizeof (buf) );
	int i;
	for (i=0;i<4;i++)
	{
		pafee_strcat (buf, strData[i].name );
		pafee_strcat (buf, "=");
		pafee_strcat (buf, strData[i].Data );
		pafee_strcat (buf, "\r\n");
	}
	//
	Tokenizer tok;
	int index;
	int length = pafee_strlen(buf);
	//
	assert(tok.GetCount() == 0);
	assert(tok.IsEmpty());
	//
	int Count=0;
	char outBuf[256];
	index = tok.Split(buf, '=', '\"', "\r\n");
	Count=tok.GetCount();
	for ( i=0;i<Count;i++)
	{
		tok.Get(i,outBuf,sizeof(outBuf));

		assert (tok.GetLength(i) == (int)pafee_strlen(outBuf));
		assert (pafee_strcmp(tok.Get(i),outBuf)==0);
	}
	assert(tok.GetInt(1,10) == 192);
	assert (index == (pafee_strstr(buf,"\r\n")+pafee_strlen("\r\n")-buf));
	while(index < length)
	{
		index += tok.Split(buf+index, '=', '\"', "\r\n");
		Count=tok.GetCount();
		for ( i=0;i<Count;i++)
		{
			tok.Get(i,outBuf,sizeof(outBuf));
			assert (tok.GetLength(i) == (int)pafee_strlen(outBuf));
			assert (pafee_strcmp(tok.Get(i),outBuf)==0);
		}
	}
}

#endif	// PAFEE_UNITTEST

}

