//****************************************************
//
//  project     PAFEE
//! @file:      PafeeFile.cpp
//! @author:    Xie Jian
//! @brief:     Supply file access interface
//
//*****************************************************

//
// P A F E E   - Pragmatic Application Framework for Embedded Environment- 
//
//
// Copyright (c) 2009 KN Information Systems Corporation,
//                    Hirocon Corporation,
//                    Technocrats Corporation
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//

#include "PafeeFile.h"
//
namespace Pafee
{

//--------------------------------------
//!	default constructor
//--------------------------------------
File::File()
{
}

//--------------------------------------
//!	destructor
//--------------------------------------
File::~File()
{
}

//--------------------------------------------------------------------------------------------------------
//!	Open a file.
//!	@return Result(1:Open Success  0:Open Failure)
//--------------------------------------------------------------------------------------------------------
bool File::Open(
	const char* filePath, ///< [in/out]file path
	const OPENMODE mode ///< [in/out]File::OPENMODE
)
{
	m_localOpenMode = mode;
	m_pFile = fopen(filePath,MODENAME[mode]);
	if(m_pFile == NULL)
	{
		return false;
	}
	else
	{
		return true;
	}
}

//-------------------------------------------------------------------------
//!	Close the file(close the stream). 
//-------------------------------------------------------------------------
void File::Close(void)
{
	fclose(m_pFile);
}

//-------------------------------------------------------------------------
//!	Get the length of the file. 
//!	@return Result(length: the length of the file)
//-------------------------------------------------------------------------
long File::GetLength(void)
{
	long length = 0;
	if(m_pFile != NULL)
	{
		unsigned long currentPosition;
		currentPosition = ftell(m_pFile);
		fseek(m_pFile,0L,SEEK_END);
		length = ftell(m_pFile);
		fseek(m_pFile,currentPosition,SEEK_SET);
	}
	return length;
}

//--------------------------------------------------------------------------------------
//!	move the pointer to the appointed position
//!	@return Result(1:Success   0:failure)
//--------------------------------------------------------------------------------------
bool File::Seek(
	int offset, ///< [in] the offset from origin
	int origin  ///< [in] File::ORIGIN
)
{
	if(origin == TOP)
	{
		File::ORIGIN = TOP;
		origin = SEEK_SET;
	}
	else if(origin == CURRENT)
	{
		File::ORIGIN = CURRENT;
		origin = SEEK_CUR;
	}
	else if(origin == END)
	{
		File::ORIGIN = END;
		origin = SEEK_END;
	}
	else
	{
		return false;
	}
	return (fseek(m_pFile,(long)offset,origin) == 0);
}

#ifdef PAFEE_UNITTEST
//----------------------------------------------------------------------------
//!	Unit test
//! @return: void
//----------------------------------------------------------------------------
void File::UnitTest(void)
{
	const int TEST_FILE_DATA_SIZE = 10;
	const int TEST_FILE_DATA_ADD_SIZE = 5;
	char wbuf[TEST_FILE_DATA_SIZE];
	char rbuf[TEST_FILE_DATA_SIZE];
	char waddbuf[TEST_FILE_DATA_ADD_SIZE];
	File file,file2;
	char* fileName = "C:\\FileTest.txt";
	char* fileName2 = "C:\\FileTest2.txt";
	char* fileNameBin = "C:\\FileTest.xls";
	OPENMODE openMode,openMode2;
	int i;
	//
	for(i = 0; i < TEST_FILE_DATA_SIZE/2; i++)
	{
		wbuf[i] = 'A';
	}
	for(i = TEST_FILE_DATA_SIZE/2; i < TEST_FILE_DATA_SIZE; i++)
	{
		wbuf[i] = 'B';
	}
	for(i = 0; i < TEST_FILE_DATA_ADD_SIZE; i++)
	{
		waddbuf[i] = 'C';
	}
	//test MODE_READ
	{
		openMode = File::MODE_READ;
		assert(file.Open(fileName,openMode) == false);
	}
	//test MODE_WRITE
	{
		openMode = File::MODE_WRITE;
		assert(file.Open(fileName,openMode) == true);
		assert(file.GetLength() == 0);
		assert(file.Write(wbuf,TEST_FILE_DATA_SIZE) >= 0);
		file.Flush();
		assert(file.Seek(0,TOP));
		assert(file.Read(rbuf,TEST_FILE_DATA_SIZE) == 0);
		file.Close();
	}
	//test MODE_READ
	{
		openMode = File::MODE_READ;
		assert(file.Open(fileName,openMode) == true);
		assert(file.GetLength() >= 0);
		assert(file.Write(wbuf,TEST_FILE_DATA_SIZE) == 0);
		file.Flush();
		assert(file.Seek(0,TOP));
		assert(file.Read(rbuf,TEST_FILE_DATA_SIZE) >= 0);
		file.Close();
	}
	//test MODE_APPEND
	{
		openMode = File::MODE_APPEND;
		assert(file.Open(fileName,openMode) == true);
		assert(file.GetLength() >= 0);
		assert(file.Write(waddbuf,TEST_FILE_DATA_ADD_SIZE) >= 0);
		file.Flush();
		assert(file.Seek(0,TOP));
		assert(file.Read(rbuf,TEST_FILE_DATA_SIZE) == 0);
		file.Close();
	}
	//test MODE_READPLUS
	{
		openMode = File::MODE_READPLUS;
		assert(file.Open(fileName,openMode) == true);
		assert(file.GetLength() >= 0);
		assert(file.Write(waddbuf,TEST_FILE_DATA_ADD_SIZE/2) >= 0);
		file.Flush();
		assert(file.Seek(0,TOP));
		assert(file.Seek(10L,TOP));
		assert(file.Read(rbuf,TEST_FILE_DATA_SIZE) >= 0);
		file.Close();
	}
	//test MODE_WRITEPLUS
	{
		openMode = File::MODE_WRITEPLUS;
		assert(file.Open(fileName,openMode) == true);
		assert(file.GetLength() == 0);
		assert(file.Write(wbuf,TEST_FILE_DATA_SIZE) >= 0);
		file.Flush();
		assert(file.Seek(0L,TOP));
		assert(file.Seek(-5L,END));
		assert(file.Read(rbuf,TEST_FILE_DATA_SIZE) >= 0);
		file.Close();
	}
	//test MODE_APPENDPLUS
	{
		openMode = File::MODE_APPENDPLUS;
		assert(file.Open(fileName,openMode) == true);
		assert(file.GetLength() >= 0);
		assert(file.Write(waddbuf,TEST_FILE_DATA_ADD_SIZE) >= 0);
		file.Flush();
		assert(file.Seek(0,TOP));
		assert(file.Seek(-10,END));
		assert(file.Read(rbuf,TEST_FILE_DATA_SIZE) >= 0);
		file.Close();
	}
	//test MODE_WRITE_BINARY
	{
		openMode = File::MODE_WRITE_BINARY;
		assert(file.Open(fileNameBin,openMode) == true);
		assert(file.GetLength() == 0);
		assert(file.Write(wbuf,TEST_FILE_DATA_SIZE) >= 0);
		file.Flush();
		assert(file.Seek(0,TOP));
		assert(file.Read(rbuf,TEST_FILE_DATA_SIZE) == 0);
		file.Close();
	}
	//test MODE_READ_BINARY
	{
		openMode = File::MODE_READ_BINARY;
		assert(file.Open(fileNameBin,openMode) == true);
		assert(file.GetLength() >= 0);
		assert(file.Write(wbuf,TEST_FILE_DATA_SIZE) == 0);
		file.Flush();
		assert(file.Seek(0,TOP));
		assert(file.Read(rbuf,TEST_FILE_DATA_SIZE) >= 0);
		file.Close();
	}
	//test MODE_APPEND_BINARY
	{
		openMode = File::MODE_APPEND_BINARY;
		assert(file.Open(fileNameBin,openMode) == true);
		assert(file.GetLength() >= 0);
		assert(file.Write(waddbuf,TEST_FILE_DATA_ADD_SIZE) >= 0);
		file.Flush();
		assert(file.Seek(0,TOP));
		assert(file.Read(rbuf,TEST_FILE_DATA_SIZE) == 0);
		file.Close();
	}
	//test MODE_READPLUS_BINARY
	{
		openMode = File::MODE_READPLUS_BINARY;
		assert(file.Open(fileNameBin,openMode) == true);
		assert(file.GetLength() >= 0);
		assert(file.Write(wbuf,TEST_FILE_DATA_SIZE) >= 0);
		file.Flush();
		assert(file.Seek(0,TOP));
		assert(file.Read(rbuf,TEST_FILE_DATA_SIZE) >= 0);
		file.Close();
	}
	//test MODE_WRITEPLUS_BINARY
	{
		openMode = File::MODE_WRITEPLUS_BINARY;
		assert(file.Open(fileNameBin,openMode) == true);
		assert(file.GetLength() == 0);
		assert(file.Write(wbuf,TEST_FILE_DATA_SIZE) >= 0);
		file.Flush();
		assert(file.Seek(0,TOP));
		assert(file.Read(rbuf,TEST_FILE_DATA_SIZE) >= 0);
		file.Close();
	}
	//test MODE_APPENDPLUS_BINARY
	{
		openMode = File::MODE_APPENDPLUS_BINARY;
		assert(file.Open(fileNameBin,openMode) == true);
		assert(file.GetLength() >= 0);
		assert(file.Write(waddbuf,TEST_FILE_DATA_ADD_SIZE) >= 0);
		file.Flush();
		assert(file.Seek(0,TOP));
		assert(file.Read(rbuf,TEST_FILE_DATA_SIZE) >= 0);
		file.Close();
	}
	//test pattern 2
	//test MODE_WRITE
	{
		memset(rbuf,'\0',TEST_FILE_DATA_SIZE);
		openMode = File::MODE_READ;
		assert(file.Open(fileName,openMode) == true);
		openMode2 = File::MODE_WRITE;
		assert(file2.Open(fileName2,openMode2) == true);
		assert(file.Seek(0,TOP));
		int rSize = file.Read(rbuf,TEST_FILE_DATA_SIZE/2);
		assert( rSize >= 0);
		assert(file2.Write(rbuf,rSize) >= 0);
		file2.Flush();
		file.Close();
		file2.Close();
	}
	//test MODE_READ
	{
		memset(rbuf,'\0',strlen(rbuf));
		openMode = File::MODE_READ;
		assert(file.Open(fileName,openMode) == true);
		openMode2 = File::MODE_APPEND;
		assert(file2.Open(fileName2,openMode2) == true);
		assert(file.Seek(5L,TOP));
		int rSize = file.Read(rbuf,TEST_FILE_DATA_SIZE/2);
		assert( rSize >= 0);
		assert(file2.Write(rbuf,rSize) >= 0);
		file2.Flush();
		file.Close();
		file2.Close();
	}
	remove(fileName);
	remove(fileName2);
	remove(fileNameBin);
}
#endif // PAFEE_UNITTEST

}
